/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * projection.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 *
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "projection.h"

//WGS-84 projection

double acos_safe(double d)
{
  if (d>1)
    return 0; 
  else if (d<-1) 
    return pi;
  else 
    return acos(d);
}

int wgs84_latradius(long lat)
{
  long r = earthRadius * cos(lat * pi / 180);
  return r;
}

long wgs84_latangle(long distance)
{
  double latangle = 180.0 * (double)distance / (pi * earthRadius);
  return (long)(latangle * 1000000);
}

long wgs84_lonangle(long lat, long distance)
{
  double r = earthRadius * cos(lat * pi / 180000000.0);
  double lonangle = 180.0 * (double)distance / (pi * r);
  return (long)(lonangle * 1000000);
}

int wgs84_compare(long lat1, long lon1, long lat2, long lon2)
{
  if ((abs(lat1-lat2)<=2) && (abs(lon1-lon2)<=2))
    return 0;
  else
    return 1;
}

double wgs84_distance_f(double lat1, double lon1, double lat2, double lon2)
{
  double r = earthRadius * cos(lat2 * pi / 180);
  double yy = - (lat1 - lat2) * pi / 180 * earthRadius;
  double xx = (lon1 - lon2) * pi / 180 * r;

  double c = sqrt(xx*xx+yy*yy);
  return c;
}

int wgs84_distance(long lat1, long lon1, long lat2, long lon2)
{
  int d = (int)(wgs84_distance_f((double)lat1/1000000.0, (double)lon1/1000000.0, (double)lat2/1000000.0, (double)lon2/1000000.0)+0.5);
  return d;
}

int wgs84_distance_mm(long lat1, long lon1, long lat2, long lon2)
{
  int d = (int)(wgs84_distance_f((double)lat1/1000000.0, (double)lon1/1000000.0, (double)lat2/1000000.0, (double)lon2/1000000.0)*1000+0.5);
  return d;
}

double wgs84_angle_f(double lat1, double lon1, double lat2, double lon2)
{
  double r = cos(lat2 * pi / 180);
  //double x = (lon1 - lon2) * r;
  //double y = lat1 - lat2;
  double x = (lon2 - lon1) * r;
  double y = lat2 - lat1;

  double angle = 90 - atan2(y,x) * 180 / pi;
  if (angle>180)
    angle -= 360;

  //printf("%f,%f - %f,%f -> %f, %f, %f , %f\n", lat1, lon1, lat2, lon2, r, x, y, angle);

  return angle;
}

int wgs84_angle(long lat1, long lon1, long lat2, long lon2)
{
  int angle = (int)wgs84_angle_f((double)lat1/1000000.0, (double)lon1/1000000.0, (double)lat2/1000000.0, (double)lon2/1000000.0);
  return angle;
}

int wgs84_waydistance(long lat, long lon, long lat1, long lon1, long lat2, long lon2)
{
  //compute the distance to the way defined by lat1,lon1->lat2,lon2
  int d;
  double latf = ((double)lat)/1000000.0;
  double lonf = ((double)lon)/1000000.0;
  double lat1f = ((double)lat1)/1000000.0;
  double lon1f = ((double)lon1)/1000000.0;
  double lat2f = ((double)lat2)/1000000.0;
  double lon2f = ((double)lon2)/1000000.0;
  double d1 = wgs84_distance_f(lat1f, lon1f, latf, lonf);   //lenght to point1
  double d2 = wgs84_distance_f(lat2f, lon2f, latf, lonf);   //lenght to point2
  double d3 = wgs84_distance_f(lat1f, lon1f, lat2f, lon2f); //lenght p1 to p2
  double a1 = acos_safe((d2*d2+d3*d3-d1*d1)/(2.0*d2*d3));
  double a2 = acos_safe((d3*d3+d1*d1-d2*d2)/(2.0*d1*d3));
  //double a3 = acos_safe((d2*d2+d1*d1-d3*d3)/(2*d1*d2));
  if ((a1<pi/2)&&(a2<pi/2))
  {
    d = (int)(d1 * sin(a2));
  }
  else if (d1<d2)
  {
    d = (int)d1;
  }
  else
  {
    d = (int)d2;
  }
  return d;
}

int wgs84_waysection(long lat, long lon, long lat1, long lon1, long lat2, long lon2)
{
  //compute the intersection square distance of the way defined by lat1,lon1->lat2,lon2
  int wdist;
  double latf = (double)lat/1000000.0;
  double lonf = (double)lon/1000000.0;
  double lat1f = (double)lat1/1000000.0;
  double lon1f = (double)lon1/1000000.0;
  double lat2f = (double)lat2/1000000.0;
  double lon2f = (double)lon2/1000000.0;
  double d1 = wgs84_distance_f(lat1f, lon1f, latf, lonf);   //lenght to point1
  double d2 = wgs84_distance_f(lat2f, lon2f, latf, lonf);   //lenght to point2
  double d3 = wgs84_distance_f(lat1f, lon1f, lat2f, lon2f); //lenght p1 to p2
  double a1 = acos_safe((d2*d2+d3*d3-d1*d1)/(2*d2*d3));
  double a2 = acos_safe((d3*d3+d1*d1-d2*d2)/(2*d1*d3));
  //double a3 = acos_safe((d2*d2+d1*d1-d3*d3)/(2*d1*d2));
  if ((a1<pi/2)&&(a2<pi/2))
  {
    wdist = (int)(d1 * cos(a2));
  }
  else if (d1<d2)
  {
    wdist = 0;
  }
  else
  {
    wdist = (int)d3;
  }
  return wdist;
}

//return distance and sections in meters
int wgs84_waysnap(long lat, long lon, long lat1, long lon1, long lat2, long lon2, long *lats, long *lons, int *distance, int *section1, int *section2)
{
  //compute the intersection square distance of the way defined by lat1,lon1->lat2,lon2
  double latf = (double)lat/1000000.0;
  double lonf = (double)lon/1000000.0;
  double lat1f = (double)lat1/1000000.0;
  double lon1f = (double)lon1/1000000.0;
  double lat2f = (double)lat2/1000000.0;
  double lon2f = (double)lon2/1000000.0;
  double d1 = wgs84_distance_f(lat1f, lon1f, latf, lonf);   //lenght to point1
  double d2 = wgs84_distance_f(lat2f, lon2f, latf, lonf);   //lenght to point2
  double d3 = wgs84_distance_f(lat1f, lon1f, lat2f, lon2f); //lenght p1 to p2
  double a1 = acos_safe((d2*d2+d3*d3-d1*d1)/(2*d2*d3));
  double a2 = acos_safe((d3*d3+d1*d1-d2*d2)/(2*d1*d3));
  //double a3 = acos_safe((d2*d2+d1*d1-d3*d3)/(2*d1*d2));
  if ((a1<pi/2)&&(a2<pi/2))
  {
    double f = d1 * cos(a2);
    if (lats == NULL)
      return -1;
    *lats = lat1 + (int)(1000000.0 * f * (lat2f - lat1f) / d3);
    *lons = lon1 + (int)(1000000.0 * f * (lon2f - lon1f) / d3);
    *distance = (int)(d1 * sin(a2));
    *section1 = (int)(d1 * cos(a2));
    *section2 = (int)(d2 * cos(a1));
  }
  else if (d1<d2)
  {
    *lats = lat1;
    *lons = lon1;
    *distance = (int)d1;
    *section1 = 0;
    *section2 = (int)d3;
  }
  else
  {
    *lats = lat2;
    *lons = lon2;
    *distance = (int)d2;
    *section1 = (int)d3;
    *section2 = 0;
  }
  return 0;
}

//compute crossing point of abscisse defined by (1-2) and (3-4)
//if do not cross, return 0
int between(long a, long a1, long a2)
{ 
  if (a1<a2)
  {
    if ((a>=a1) && (a<=a2))
      return 1;
  }
  else
  {
    if ((a>=a2) && (a<=a1))
      return 1;
  }
  return 0;
}

int wgs84_cross(long lat1, long lon1, long lat2, long lon2, 
                long lat3, long lon3, long lat4, long lon4, 
                long *lats, long *lons)
{
  double k, s;
  double a1 = lat2 - lat1;
  double b1 = lon2 - lon1;
  double a2 = lat4 - lat3;
  double b2 = lon4 - lon3;
  if (lat1==lat2)
  {
    if (lat3==lat4)
      return -1;
    s = (lat1 - lat3) / a2;
    *lats = lat1;
    *lons = lon3 + s * b2;
    if (between(*lons, lon1, lon2) && between(*lons, lon3, lon4) && between(*lats, lat3, lat4))
      return 1;
  }
  else if (lat3==lat4)
  {
    k = (lat3 - lat1) / a1;
    *lats = lat3;
    *lons = lon1 + k * b1;
    if (between(*lons, lon1, lon2) && between(*lons, lon3, lon4) && between(*lats, lat1, lat2))
      return 1;
  }
  else if (lon1==lon2)
  {
    if (lon3==lon4)
      return -1;
    s = (lon1 - lon3) / b2;
    *lats = lat3 + s * a2;
    *lons = lon1;
    if (between(*lats, lat1, lat2) && between(*lats, lat3, lat4) && between(*lons, lon3, lon4))
      return 1;
  }
  else if (lon3==lon4)
  {
    k = (lon3 - lon1) / b1;
    *lats = lat1 + k * a1 ;
    *lons = lon3;
    if (between(*lats, lat1, lat2) && between(*lats, lat3, lat4) && between(*lons, lon1, lon2))
      return 1;
  }
  else
  {
    double s = (lon1 * a1 + b1 * lat3 - b1 * lat1 - a1 * lon3) / (a1 * b2 - b1 * a2);
    double k = (lat3 + s * a2 - lat1) / a1;
    *lats = lat1 + k * a1;
    *lons = lon1 + k * b1;
    if ((k>=0) && (k<=1) && (s>=0) && (s<=1))
    {
      return 1;
    }
  }
  return 0;
}

//x,y in floating meters
void wgs84_pos2screen_f(double lat, double lon, double zero_lat, double zero_lon, double zero_angle, double *x, double *y)
{
  double dlat = (zero_lat - lat) * pi / 180;
  double dlon = (zero_lon - lon) * pi / 180;

  double c = sqrt(dlat*dlat+dlon*dlon);
  double angle = atan2(dlat,dlon) - zero_angle * pi / 180;

  double t1 = -c * cos(angle);
  double t2 = c * sin(angle);

  *x = t1 * earthRadius;
  *y = t2 * earthRadius;
}

void wgs84_screen2pos_f(double x, double y, double zero_lat, double zero_lon, double zero_angle, double *lat, double *lon)
{
  double c = sqrt(x*x+y*y);
  double angle = atan2(x,y) - zero_angle * pi / 180;

  double xx = c * sin(angle);
  double yy = c * cos(angle);

  *lat = zero_lat + yy  * 180 / (pi * earthRadius);
  *lon = zero_lon + xx  * 180 / (pi * earthRadius);
}

//"my" center projection
void cent_pos2screen_f(double lat, double lon, double zero_lat, double zero_lon, double zero_angle, double *x, double *y)
{
  double r = earthRadius * cos(zero_lat * pi / 180);
  double yy = - (lat - zero_lat) * pi / 180 * earthRadius;
  double xx = (lon - zero_lon) * pi / 180 * r;

  double c = sqrt(xx*xx+yy*yy);
  double angle = atan2(yy, xx) + zero_angle * pi / 180;

  *x = c * cos(angle);
  *y = c * sin(angle);
}

void cent_screen2pos_f(double x, double y, double zero_lat, double zero_lon, double zero_angle, double *lat, double *lon)
{
  double r = earthRadius * cos(zero_lat * pi / 180);

  double c = sqrt(x*x+y*y);
  double angle = atan2(y,x) + zero_angle * pi / 180;

  double xx = c * cos(angle);
  double yy = c * sin(angle);

  *lat = zero_lat + yy  * 180 / (pi * earthRadius);
  *lon = zero_lon + xx  * 180 / (pi * r);
}

//mercator projection

void merc_to_f(double lat, double lon, double *x, double *y)
{
  *x = lon * earthRadius * pi / 180;
  *y = log(tan(pi/4 + lat * pi/360)) * earthRadius;
}

void merc_from_f(double x, double y, double *lat, double *lon)
{
  *lat = atan(exp(y/earthRadius)) / pi * 360 - 90;
  *lon = x / earthRadius / pi * 180;
}

//TODO optimize the code
void merc_pos2screen_f(double lat, double lon, double zero_lat, double zero_lon, double zero_angle, double *x, double *y)
{
  double xx = (zero_lon-lon) * earthRadius * pi / 180;
  double yy = earthRadius * log(tan(pi/4 + zero_lat * pi/360)/tan(pi/4 + lat * pi/360));

  double c = sqrt(xx*xx+yy*yy);
  double angle = atan2(yy,xx) - zero_angle * pi / 180;

  *x = -c * cos(angle);
  *y = c * sin(angle);
}

void merc_screen2pos_f(double x, double y, double zero_lat, double zero_lon, double zero_angle, double *lat, double *lon)
{
  //double x2 = zero_lon * earthRadius * pi / 180;
  //double y2 = earthRadius * log(tan(pi/4 + zero_lat * pi/360));

  double c = sqrt(x*x+y*y);
  double angle = atan2(x,y) - zero_angle * pi / 180;

  double xx = c * sin(angle);
  double yy = c * cos(angle);

  *lat = zero_lat + atan(exp(yy/earthRadius)) / pi * 360 - 90;
  *lon = zero_lon + xx / earthRadius / pi * 180;
}

//x,y in integer milimeters
void wgs84_pos2screen(long lat, long lon, long zero_lat, long zero_lon, int zero_angle, long *x, long *y)
{
  double fx, fy;
  cent_pos2screen_f((double)lat/1000000.0, (double)lon/1000000.0, (double)zero_lat/1000000.0, (double)zero_lon/1000000.0, (double)zero_angle, &fx, &fy);
  *x = (int)(fx*1000);
  *y = (int)(fy*1000);
}

void wgs84_screen2pos(long x, long y, long zero_lat, long zero_lon, int zero_angle, long *lat, long *lon)
{
  double flat, flon;
  cent_screen2pos_f((double)x/1000.0, (double)y/1000.0, (double)zero_lat/1000000.0, (double)zero_lon/1000000.0, (double)zero_angle, &flat, &flon);
  *lat = (int)(flat*1000000);
  *lon = (int)(flon*1000000);
}
