/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * map_country.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#if !defined(WIN32)
#include <unistd.h>
#endif
#include <expat.h>
#include "mcthread.h"
#include "mcconf.h"
#include "projection.h"
#include "gpx.h"
#include "safefile.h"

#define BUFF_SIZE 1000

gpx_track *gpxtrack = NULL;
gpx_track *gpxlog = NULL;
mcmutex gpxmutex;

//common
int gpx_lock()
{
  return mcmutex_lock(&gpxmutex);
}

int gpx_unlock()
{
  return mcmutex_unlock(&gpxmutex);
}

int gpx_free(gpx_track *gpx)
{
  gpx_points *p = gpx->first;
  while (p)
  {
    gpx_points *p2 = p->next;
    free(p);
    p = p2;
  }
  gpx->first = NULL;
  gpx->last = NULL;
  gpx->count = 0;
  return 0;
}

int gpx_header(FILE *f)
{
  fprintf(f, "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\n");
  fprintf(f, "<gpx version=\"1.1\" creator=\"MC Navi\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.topografix.com/GPX/1/1\" xsi:schemaLocation=\"http://www.topografix.com/GPX/1/1\">\n");
  fprintf(f, "<trk>\n");
  fprintf(f, "  <name>Noname</name>\n");
  return 0;
}

int gpx_footer(FILE *f)
{
  fprintf(f, "</trk>\n");
  fprintf(f, "</gpx>\n");
  return 0;
}

int gpx_segment_start(FILE *f)
{
  fprintf(f, "  <trkseg>\n");
  return 0;
}

int gpx_segment_stop(FILE *f)
{
  fprintf(f, "  </trkseg>\n");
  return 0;
}

int gpx_write(FILE *f, gpx_points *p)
{
  int i;
  char s[100];
  for (i=0; i<p->count; i++)
  {
    snprintf(s, sizeof(s), "    <trkpt lat=\"%1.7f\" lon=\"%1.7f\">\n", p->points[i].lat, p->points[i].lon);
    fputs(s, f);
    snprintf(s, sizeof(s), "      <time>%1.1f</time>\n", p->points[i].altitude);
    fprintf(f, "      <time>");
    time_t t;
    t = (time_t)p->points[i].time;
    struct tm lt;
    gmtime_r (&t, &lt);
    int milisec = (p->points[i].time - trunc(p->points[i].time)) * 100;
    strftime(s, sizeof(s), "%Y-%m-%dT%H:%M:%S", &lt);
    fputs(s, f);
    snprintf(s, sizeof(s), ".%02dZ", milisec);
    fputs(s, f);
    fprintf(f, "</time>\n");
    snprintf(s, sizeof(s), "      <ele>%1.1f</ele>\n", p->points[i].altitude);
    fputs(s, f);
    snprintf(s, sizeof(s), "      <speed>%1.1f</speed>\n", p->points[i].speed);
    fputs(s, f);
    snprintf(s, sizeof(s), "      <magvar>%1.1f</magvar>\n", p->points[i].course);
    fputs(s, f);
    snprintf(s, sizeof(s), "      <hdop>%1.1f</hdop>\n", p->points[i].eph);
    fputs(s, f);
    snprintf(s, sizeof(s), "      <vdop>%1.1f</vdop>\n", p->points[i].epv);
    fputs(s, f);
    fprintf(f, "    </trkpt>\n");
  }
  return 0;
}

int gpx_init()
{
  mcmutex_init(&gpxmutex); 
  return 0;
}

int gpx_deinit()
{
  gpx_log_stop();
  mcmutex_destroy(&gpxmutex); 
  if (gpxtrack)
  {
    gpx_free(gpxtrack);
    gpxtrack = NULL;
  }
  if (gpxlog)
  {
    gpx_free(gpxlog);
    gpxlog = NULL;
  }
  return 0;
}

//track
typedef struct
{
  int depth;
  int gpx;
  int gpx_trk;
  int gpx_trkseg;
} gpx_load_data;

char *attr_str(const char **attr, const char *name)
{
  int i;
  for (i = 0; attr[i]; i += 2)
  {
    //printf("%s = %s\n", attr[i], attr[i+1]);
    if (strcmp(attr[i], name) == 0)
      return (char*)attr[i+1];
  }
  return NULL;
}

int attr_long(const char **attr, const char *name, long *val)
{
  char *an;
  if ((an = attr_str(attr, name)) != NULL)
  {
    *val = (long)atol(an);
    return 0;
  }
  return -1;
}

int attr_double(const char **attr, const char *name, double *val)
{
  char *an;
  if ((an = attr_str(attr, name)) != NULL)
  {
    *val = (double)atof(an);
    return 0;
  }
  return -1;
}

static void XMLCALL startElement(void *userData, const char *name, const char **attr)
{
  gpx_load_data *data = (gpx_load_data *)userData;

  if ((data->depth == 0) && (strcmp(name, "gpx") == 0))
  {
    data->gpx = 1;
  }
  else if ((data->depth == 1) && (strcmp(name, "trk") == 0) && (data->gpx == 1))
  {
    data->gpx_trk = 1;
  }
  else if ((data->depth == 2) && (strcmp(name, "trkseg") == 0) && (data->gpx_trk == 1))
  {
    data->gpx_trkseg = 1;
    gpx_points *p = (gpx_points*)malloc(sizeof(gpx_points));
    p->count = 0;;
    p->split = 1;
    p->next = NULL;
    p->prev = gpxtrack->last;
    if (gpxtrack->first == NULL)
      gpxtrack->first = p;
    if (gpxtrack->last)
      gpxtrack->last->next = p;
    gpxtrack->last = p;
  }
  else if ((data->depth == 3) && (strcmp(name, "trkpt") == 0) && (data->gpx_trkseg == 1))
  {
    double lat, lon;
    if ((attr_double(attr, "lat", &lat) == 0) && (attr_double(attr, "lon", &lon) == 0))
    {
      if (gpxtrack->last->count == GPXPOINTSLEN)
      {
        gpx_points *p = (gpx_points*)malloc(sizeof(gpx_points));
        p->count = 0;;
        p->split = 0;
        p->next = NULL;
        p->prev = gpxtrack->last;
        gpxtrack->last->next = p;
        gpxtrack->last = p;
      }
      gpxtrack->last->points[gpxtrack->last->count].lat = lat;
      gpxtrack->last->points[gpxtrack->last->count].lon = lon;
      gpxtrack->last->count++;
      gpxtrack->count++;
      //printf("%d: %f,%f\n", gpxtrack->count, lat, lon);
    }
    else if ((data->depth == 4) && (strcmp(name, "ele") == 0) && (data->gpx_trkseg == 1))
    {
      
    }
  }
  data->depth++;
}

static void XMLCALL endElement(void *userData, const char *name)
{
  gpx_load_data *data = (gpx_load_data *)userData;
  data->depth--;

  if ((data->depth == 0) && (strcmp(name, "gpx") == 0))
  {
    data->gpx = 0;
  }
  else if ((data->depth == 1) && (strcmp(name, "trk") == 0))
  {
    data->gpx_trk = 0;
  }
  else if ((data->depth == 2) && (strcmp(name, "trkseg") == 0))
  {
    data->gpx_trkseg = 0;
  }
}

int gpx_load(char *filename)
{
  printf("Load %s\n", filename);
  gpx_lock();

  if (gpxtrack)
  {
    gpx_free(gpxtrack);
    gpxtrack = NULL;
  }

  FILE *f;
  if ((f = fopen(filename, "r")) == NULL)
  {
    printf("Error opening gpx file!\n");
    gpx_unlock();
    return -1;
  }
  
  //open XML parser
  int depth = 0;
  gpx_load_data data;
  memset(&data, 0, sizeof(data));
  printf("Import data from %s\n", filename);
  XML_Parser parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser, &data);
  XML_SetElementHandler(parser, startElement, endElement);

  //crate track
  gpxtrack = (gpx_track*)malloc(sizeof(gpx_track));
  memset(gpxtrack, 0, sizeof(gpx_track));
  gpxtrack->filename = (char*)malloc(strlen(filename)+1);
  strcpy(gpxtrack->filename, filename);

  //iterate items
  int done;
  do
  {
    void *buf = XML_GetBuffer(parser, BUFF_SIZE);
    if (buf == NULL)
    {
      printf("Error alloc buffer\n");
      gpx_unlock();
      fclose(f);
      return -1;
    }
    int len = (int)fread(buf, 1, BUFF_SIZE, f);
    //printf("LEN: %d\n", len);
    done = (int)(len < BUFF_SIZE);
    if (! XML_ParseBuffer(parser, len, done))
    {
      fprintf(stderr,
              "%s at line %lu\n",
              XML_ErrorString(XML_GetErrorCode(parser)),
              XML_GetCurrentLineNumber(parser));
      gpx_unlock();
      fclose(f);
      return -1;
    }
  } while (!done);
  fclose(f);
  gpx_unlock();
  if (gpxtrack)
    printf("Points: %d\n", gpxtrack->count);
  return 0;
}

int gpx_close()
{

  return 0;
}

//log
int gpx_log(char *filename)
{
  gpx_lock();
  if (gpxlog)
  {
    gpx_unlock();
    return -1;
  }
  gpxlog = (gpx_track*)malloc(sizeof(gpx_track));
  memset(gpxlog, 0, sizeof(gpx_track));
  gpxlog->filename = (char*)malloc(strlen(filename)+1);
  strcpy(gpxlog->filename, filename);
  gpxlog->f = fsafeopen(filename, "w");
  gpx_header(gpxlog->f);
  gpx_segment_start(gpxlog->f);
  gpx_unlock();
  return 0;
}

void gpx_addpoint(gpx_track *gpxlog, gpx_point *point)
{
  if (gpxlog->last && (gpxlog->last->count == GPXPOINTSLEN))
  {
    //write to file
    gpx_write(gpxlog->f, gpxlog->last);
  }
  if ((gpxlog->last==NULL) || (gpxlog->last->count == GPXPOINTSLEN))
  {
    //create new structure
    gpx_points *p = (gpx_points*)malloc(sizeof(gpx_points));
    p->count = 0;
    p->next = NULL;
    p->prev = gpxlog->last;
    if (gpxlog->last)
      gpxlog->last->next = p;
    if (gpxlog->first == NULL)
    {
      gpxlog->first = p;
      p->split = 1;
    }
    gpxlog->last = p;
  }
  gpxlog->last->points[gpxlog->last->count] = *point;
  gpxlog->last->count++;
  gpxlog->count++;
}

void gpx_replacepoint(gpx_track *gpxlog, gpx_point *point)
{
  gpxlog->last->points[gpxlog->last->count-1] = *point;
}

/*
FILE *ftest = NULL;
int itest = 0;

void testpoint(gpx_point *point)
{
  if (ftest==NULL)
  {
    ftest = fopen("aaa.gpm", "r");
  }

  char s[100];
  fgets(s, sizeof(s), ftest);
  point->lat = atof(s);
  fgets(s, sizeof(s), ftest);
  point->lon = atof(s);
  point->altitude = 231;
  itest++;
}
*/

int gpx_logpoint(gpx_point *point)
{
  gpx_lock();
  if (gpxlog==NULL)
  {
    gpx_unlock();
    return -1;
  }

  //testpoint(point);

  //detect duplicated points
  if (gpxlog->last && (gpxlog->last->count>0))
  {
    gpx_point *p = &gpxlog->last->points[gpxlog->last->count-1];
    if ((p->lat==point->lat) && (p->lon==point->lon) && (p->altitude==point->altitude) && ((p->time-point->time)<=1))
    {
      gpx_unlock();
      return -1;
    }
  }

  /*if (itest>=8)
    printf("%f %f\n", point->lat, point->lon);*/

  //when decimating, write point only if difference greater then limit
  //we compute the difference as distance and the angle from last saved point
  /*printf("%d %d\n", config->gpx_decimate, gpxlog->count);
  if (gpxlog->last)
    printf("X %d %d\n", gpxlog->last->split, gpxlog->last->count);*/
  if ((config->gpx_decimate == 0) || (gpxlog->count==0) || (gpxlog->last->split && gpxlog->last->count==0))
  {
    gpxlog->lat = point->lat;
    gpxlog->lon = point->lon;
    gpxlog->altitude = point->altitude;
    gpxlog->aja = 0;
    gpx_addpoint(gpxlog, point);
  }
  else
  {
    //distance from previous saved point
    double dd = wgs84_distance_f(gpxlog->lat, gpxlog->lon, point->lat, point->lon);
    int d = (int)(dd*1000);
    //printf("d=%d\n", d);
    if (d<config->gpx_distance)
    {
      if (gpxlog->aja == 0)
      {
        gpx_addpoint(gpxlog, point);
        gpxlog->aja = 1;
        //printf("angle=%f rise=%f\n", gpxlog->angle, gpxlog->rise);
      }
      else
      {
        gpx_replacepoint(gpxlog, point);
      }
    }
    else
    {
      if (gpxlog->aja < 2)
      {
        if (gpxlog->aja == 1)
        {
          gpxlog->lat = gpxlog->last->points[gpxlog->last->count-1].lat;
          gpxlog->lon = gpxlog->last->points[gpxlog->last->count-1].lon;
          point->altitude = gpxlog->last->points[gpxlog->last->count-1].altitude;
        }
        gpxlog->angle =  wgs84_angle_f(gpxlog->lat, gpxlog->lon, point->lat, point->lon);
        gpxlog->rise = 180 * atan2(dd, point->altitude - gpxlog->altitude) / pi;
        gpxlog->aja = 2;
        gpx_addpoint(gpxlog, point);
        //printf("angle=%f rise=%f\n", gpxlog->angle, gpxlog->rise);
      }
      else
      {
        //printf("  lat=%f lon=%f alt=%f\n", gpxlog->lat, gpxlog->lon, gpxlog->altitude);
        //printf("    lat=%f lon=%f alt=%f\n", point->lat, point->lon, point->altitude);
        double angle = wgs84_angle_f(gpxlog->last->points[gpxlog->last->count-1].lat, gpxlog->last->points[gpxlog->last->count-1].lon, point->lat, point->lon);
        double rise = 180 * atan2(dd, gpxlog->last->points[gpxlog->last->count-1].altitude - gpxlog->altitude) / pi;
        //printf("angle=%f rise=%f %f %f %f\n", angle, rise, point->altitude, gpxlog->altitude, dd);
        int tangle = fabs(gpxlog->angle - angle);
        if (tangle>180)
          tangle = 360-tangle;
        int trise = fabs(gpxlog->rise - rise);
        
        if (((tangle*1000)>=config->gpx_angle) || ((trise*1000)>=config->gpx_rise))
        {
          gpx_addpoint(gpxlog, point);
          gpxlog->lat = point->lat;
          gpxlog->lon = point->lon;
          gpxlog->altitude = point->altitude;
          gpxlog->angle = angle;
          gpxlog->rise = rise;
          gpxlog->aja = 2;
        }
        else
        {
          gpx_replacepoint(gpxlog, point);
        }
      }
    }
  }

  gpx_unlock();
  return 0;
}

int gpx_log_split()
{
  gpx_lock();
  if ((gpxlog==NULL) || (gpxlog->last==NULL) || (gpxlog->last->count==0))
  {
    gpx_unlock();
    return -1;
  }
  
  //write to file
  gpx_write(gpxlog->f, gpxlog->last);

  //create new segment
  gpx_segment_stop(gpxlog->f);
  gpx_segment_start(gpxlog->f);

  //create new structure
  gpx_points *p = (gpx_points*)malloc(sizeof(gpx_points));
  p->count = 0;
  p->split = 1;
  p->next = NULL;
  p->prev = gpxlog->last;
  gpxlog->last->next = p;
  gpxlog->last = p;
  gpxlog->aja = 0;
  
  gpx_unlock();
  return 0;
}

int gpx_log_stop()
{
  gpx_lock();
  if (gpxlog==NULL)
  {
    gpx_unlock();
    return -1;
  }
  if (gpxlog->last)
  {
    //write to file
    gpx_write(gpxlog->f, gpxlog->last);
  }
  gpx_segment_stop(gpxlog->f);
  gpx_footer(gpxlog->f);
  fclose(gpxlog->f);
  gpxlog->f = NULL;
  if (gpxtrack == NULL)
    gpxtrack = gpxlog;
  else
    gpx_free(gpxlog);
  gpxlog = NULL;
  gpx_unlock();
  return 0;
}
