/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * map_render.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "draw_imlib2.h"
#include "functions.h"
#include "projection.h"
#include "map_render.h"
#include "map_read.h"
#include "navigation.h"
#include "gpx.h"
#include "mctime.h"
#include "mcconf.h"

#define MAX_ZOOM_CYCLE 13
#define MAX_ZOOM_FOOT  12
#define MIN_DRAWNAME_WIDTH 10

int background = 0xe9e7dc;
//int background = 0xd2d0c6;
#define gpxtrackcolor 0x22ff00
#define gpxlogcolor 0xff2200

typedef struct
{
  drw_image icon;
  int width;
  int height;
} mapicon;

/*long tourist_colors[WF_COLOR_COUNT] = 
{ 0xff0000, 0x00ff00, 0x0000ff, 0xffff00, 0x964b00, 0xff5400, 0xff54ff, 0xffa854, 0x54e0ff, 0x887b36 };*/

mapicon pointer;
mapicon poi_icon[cntpoitypes];
mapicon turn_icon[cntroutedir];
drw_font fonts[mapfonts];

void *mr_thread(void *t);


// ************************************
// create and destroy
// ************************************

int mr_create(mr_image *img)
{
  int i;

  memset(poi_icon, 0, sizeof(poi_icon));
  memset(turn_icon, 0, sizeof(turn_icon));
  memset(fonts, 0, sizeof(fonts));
  memset(&pointer, 0, sizeof(pointer));

  draw_init();

  //load fonts
  if (img->style)
  {
    for (i=0; i<mapfonts; i++)
    {
      if (img->style->font[i])
      {
        //printf("%s\n",img->style->zoom[img->zoom].ways[maptype].font);
        fonts[i] = draw_loadfont(img->style->font[i]->name);
        if (fonts[i] == NULL)
        {
          printf("Cannot load font\n");
        }
      }
    }

    //load turn arrows
    for (i=0; i<cntroutedir; i++)
    {
      if (img->style->turn[i])
      {
        //printf("%2d %s\n", i, img->style->turn[i]);
        turn_icon[i].icon = draw_loadimage(img->style->turn[i]);
        if (turn_icon[i].icon)
        {
          draw_getsize(turn_icon[i].icon, &turn_icon[i].width, &turn_icon[i].height);
        }
        else
        {
          printf("Cannot load arrow icon %s\n", img->style->turn[i]);
        }
      }
    }
  }

  //create rendering thread
  mcsema_init(&img->sema);
  mcmutex_init(&img->mutex);
  mcthread_create(&img->thread, mr_thread, img, 1);
  
  return 0;
}

int mr_free(mr_image *img)
{
  int i;

  //destroy the thread
  img->terminate = 1;
  mcsema_signal(&img->sema);
  mcthread_join(&img->thread);
  mcsema_destroy(&img->sema);
  mcmutex_destroy(&img->mutex);

  //destroy icons and fonts
  for (i=0; i<cntpoitypes; i++)
  {
    if (poi_icon[i].icon)
      draw_destroy(poi_icon[i].icon);
  }
  memset(poi_icon, 0, sizeof(poi_icon));
  for (i=0; i<cntroutedir; i++)
  {
    if (turn_icon[i].icon)
      draw_destroy(turn_icon[i].icon);
  }
  memset(turn_icon, 0, sizeof(turn_icon));
  for (i=0; i<mapfonts; i++)
  {
    if (fonts[i])
      draw_freefont(fonts[i]);
  }
  memset(fonts, 0, sizeof(fonts));
  if (pointer.icon)
  {
    draw_destroy(pointer.icon);
    pointer.icon = NULL;
  }

  draw_deinit();
  printf("Render gone\n");
  return 0;
}

// render track

void render_gpx(mr_image *img, gpx_track *track, int color)
{
  gpx_points *points = track->first;
  int first = 1;
  void *image = img->select ? img->image1 : img->image2;
  long size = 1 << img->zoom;
  draw_setcolor(color);
  while (points)
  {
    long i, x1, y1, x2, y2;
    long lat, lon;
    if (points->split)
      first = 1;
    for (i=0; i<points->count; i++)
    {
      lat = (long)(points->points[i].lat * 1000000);
      lon = (long)(points->points[i].lon * 1000000);
      wgs84_pos2screen(lat, lon, img->lat, img->lon, img->angle, &x2, &y2);
      x2 = img->x + x2 / size;
      y2 = img->y + y2 / size;
      if (first == 0)
      {
        if (is_rectinrect(x1, y1, x2, y2, 0, 0, img->width, img->height))
        {
          draw_line(image, x1, y1, x2, y2, 3);
        }
      }
      x1 = x2;
      y1 = y2;
      first = 0;
    }
    points = points->next;
  }
}

// ************************************
// render the map
// ************************************

int do_render(mr_image *img)
{
  long size;
  long lat1, lon1, lat2, lon2, lat3, lon3, lat4, lon4;
  long lata1, lona1, lata2, lona2;
  int i;
  maptypes maptype;
  void *image = img->select ? img->image1 : img->image2;

  if (img->style==NULL)
    return -1;

  struct timeval timer;
  timer_start(&timer);
  //printf("Render start\n");
  //printf("%ld,%ld - zoom %d angle %d w %d h %d x %d y %d\n", img->lat, img->lon, img->zoom, img->angle, img->width, img->height, img->x, img->y);

  draw_setcolor(background);
  draw_rectangle(image, 0, 0, img->width, img->height, 1);

  size = 1 << img->zoom;

  //TODO fix for x,y
  wgs84_screen2pos(-img->x*size,             img->y*size,              img->lat, img->lon, img->angle, &lat1, &lon1);
  wgs84_screen2pos((img->width-img->x)*size, img->y*size,              img->lat, img->lon, img->angle, &lat2, &lon2);
  wgs84_screen2pos(-img->x*size,             -(img->height-img->y)*size, img->lat, img->lon, img->angle, &lat3, &lon3);
  wgs84_screen2pos((img->width-img->x)*size, -(img->height-img->y)*size, img->lat, img->lon, img->angle, &lat4, &lon4);

  lata1 = min4(lat1, lat2, lat3, lat4);
  lona1 = min4(lon1, lon2, lon3, lon4);
  lata2 = max4(lat1, lat2, lat3, lat4);
  lona2 = max4(lon1, lon2, lon3, lon4);
  //printf("%ld,%ld - %ld,%ld\n", lata1, lona1, lata2, lona2);

  //draw planes first
  maptype = (maptypes)(maxmaptypes+1);
  mapplanesearch *ps = map_planes_start(lata1, lona1, lata2, lona2);
  while (maptype>firstplanetype)
  {
    mplane *plane;
    maptype = (maptypes)(maptype-1);
    
    if (img->style->zoom[img->zoom].ways[maptype].enable == 0)
      continue;

    long color = img->style->zoom[img->zoom].ways[maptype].color;
    //printf("Zoom %d, maptype %d color %lx\n", img->zoom, maptype, color);
    map_planes_type(ps, maptype);
    while ((plane = map_planes_get(ps)) != NULL)
    {
      int k = 0;
      while (k < plane->mp->count)
      {
        drw_pos *poly = (drw_pos *)malloc(sizeof(drw_pos)*plane->mp->polygon[k]->count);
        if (poly)
        {
          drw_pos *poly2 = poly;
          drw_pos *poly3 = NULL;
          int pointcount = 0;
          for (i=0; i<plane->mp->polygon[k]->count; i++)
          {
            long x1, y1;
            wgs84_pos2screen(plane->mp->polygon[k]->points[i].lat, plane->mp->polygon[k]->points[i].lon, img->lat, img->lon, img->angle, &x1, &y1);
            poly2->x = img->x + x1 / size;
            poly2->y = img->y + y1 / size;
            //make a little filtering to discard very small segments
            if ((poly3 == NULL) || (abs(poly3->x-poly2->x)>1) || (abs(poly3->y-poly2->y)>1))
            {
              poly3 = poly2;
              poly2++;
              pointcount++;
            }
          }
          if (k==0)
            draw_setcolor(color);
          else
            draw_setcolor(background);
          draw_polygon(image, poly, pointcount, 1);
        }
        free(poly);
        k++;
      }
    }
  }
  map_planes_stop(ps);
  //printf("Planes: ");  timer_print(&timer);  timer_start(&timer);

  //draw the route
  nav_lock();
  if (route && route->route)
  {
    long x1, y1, x2, y2;
    map_position pos;
    route_node *node = route->route;

    draw_setcolor(img->style->zoom[img->zoom].nav.color);

    pos = node->pos;   //TODO sometimes crashes here when rendering and exiting application !!!!
    wgs84_pos2screen(pos.lat, pos.lon, img->lat, img->lon, img->angle, &x1, &y1);
    x1 = img->x + x1 / size;
    y1 = img->y + y1 / size;
    node = node->next_node;
    while (node)
    {
      wgs84_pos2screen(node->pos.lat, node->pos.lon, img->lat, img->lon, img->angle, &x2, &y2);
      x2 = img->x + x2 / size;
      y2 = img->y + y2 / size;
      if (is_rectinrect(pos.lat, pos.lon, node->pos.lat, node->pos.lon, lata1, lona1, lata2, lona2))
      {
        int width = img->style->zoom[img->zoom].ways[node->waytype].width + img->style->routewidth;
        //int width = img->style->zoom[img->zoom].ways[node->waytype].width * 4;
        //int width = img->style->zoom[img->zoom].nav.width;
        draw_line(image, x1, y1, x2, y2, width);
      }
      x1 = x2;
      y1 = y2;
      pos = node->pos;
      node = node->next_node;
    }
  }
  nav_unlock();

  //draw cycle and tourist ways
  //TODO optimize, now it scans for segments many times, this is CPU hungry, scan only once and cache for later usage
  /*long wi, we, wm;
  if ((config->route_vehicle == RV_FOOT) && (img->zoom <= MAX_ZOOM_FOOT))
  {
    wi = WF_COLOR_END;
    we = WF_COLOR_START;
    wm = WF_COLOR_MASK_FOOT;
  }
  else if ((config->route_vehicle == RV_MTB) && (img->zoom <= MAX_ZOOM_FOOT))
  {
    wi = WF_COLOR_CYCLE;
    we = WF_COLOR_START;
    wm = WF_COLOR_MASK_MTB;
  }
  else if ((config->route_vehicle == RV_CYCLE) && (img->zoom <= MAX_ZOOM_CYCLE))
  {
    wi = WF_COLOR_CYCLE;
    we = WF_COLOR_CYCLE;
    wm = WF_COLOR_MASK_CYCLE;
  }
  else
  {
    wi = 0;
    we = 1;
  }

  while (wi>=we)
  {
    long wf = 1<<wi;
    maptype = maxwaytypes+1;
    while (maptype>0)
    {
      mapwaysegsearch *cs;
      map_wayseg *way;
      maptype--;

      if (img->style->zoom[img->zoom].ways[maptype].enable == 0)
        continue;
      
      cs = map_wayseg_start(lata1, lona1, lata2, lona2, maptype);
      while ((way = map_wayseg_get(cs)) != NULL)
      {
        if ((wf & way->flags) != 0)
        {
          long x1, y1, x2, y2;
          wgs84_pos2screen(way->start.lat, way->start.lon, img->lat, img->lon, img->angle, &x1, &y1);
          wgs84_pos2screen(way->end.lat, way->end.lon, img->lat, img->lon, img->angle, &x2, &y2);
          x1 = img->x + x1 / size;
          y1 = img->y + y1 / size;
          x2 = img->x + x2 / size;
          y2 = img->y + y2 / size;
          //printf("  %ld [%ld,%ld] [%ld,%ld] -> [%ld,%ld] [%ld,%ld]\n", way->id, way->start.lat, way->start.lon, way->start.lat, way->start.lon, x1, y1, x2, y2);
          int ww = upper_bits(way->flags & wm, wi);
          int wt = scan_bits(way->flags & wm);
          //int width = img->style->zoom[img->zoom].ways[maptype].width * 3 - img->style->zoom[img->zoom].ways[maptype].width * ww;
          int width = img->style->zoom[img->zoom].ways[maptype].width + img->style->touristwidth * (wt-ww);
          long color = tourist_colors[wi-WF_COLOR_START];
          draw_setcolor(color);
          draw_line(image, x1, y1, x2, y2, width);
        }
      }
      map_wayseg_stop(cs);
    }
    wi--;
  }*/
  //printf("Tourist: ");
  //timer_print(&timer);

  //draw from lowest ways to major
  mapsegmentsearch *cs = map_segment_start(lata1, lona1, lata2, lona2);
  maptype = (maptypes)(maxwaytype+1);
  while (maptype>0)
  {
    map_segment *way;
    maptype = (maptypes)(maptype-1);

    //printf("Zoom %d, maptype %d\n", img->zoom, maptype);
    if (img->style->zoom[img->zoom].ways[maptype].enable == 0)
      continue;

    int width = img->style->zoom[img->zoom].ways[maptype].width;
    int fontindex = img->style->zoom[img->zoom].ways[maptype].font;
    int names = (fontindex>0) && (width>=img->style->font[fontindex-1]->size) ? 1 : 0;
    map_segment_type(cs, maptype, names);
    while ((way = map_segment_get(cs)) != NULL)
    {
      //draw way segment
      long x1, y1, x2, y2;
      wgs84_pos2screen(way->start.lat, way->start.lon, img->lat, img->lon, img->angle, &x1, &y1);
      wgs84_pos2screen(way->end.lat, way->end.lon, img->lat, img->lon, img->angle, &x2, &y2);
      x1 = img->x + x1 / size;
      y1 = img->y + y1 / size;
      x2 = img->x + x2 / size;
      y2 = img->y + y2 / size;
      //printf("    [%ld,%ld] [%ld,%ld] -> [%ld,%ld] [%ld,%ld]\n", way->start.lat, way->start.lon, way->start.lat, way->start.lon, x1, y1, x2, y2);
      unsigned long color = img->style->zoom[img->zoom].ways[maptype].color;
      if (way->flags & WF_TUNNEL)
      {
        color += ((0xff - ((color >> 16) & 0xff)) / 2) << 16;
        color += ((0xff - ((color >> 8) & 0xff)) / 2) << 8;
        color += (0xff - (color & 0xff)) / 2;
      }
      draw_setcolor(color);
      draw_line(image, x1, y1, x2, y2, width);

      //arrow for oneway roads
      if (((way->flags & WF_ONEWAY) != 0) && (width>8))
      {
        int xl = x2-x1;
        int yl = y2-y1;
        int seglen = sqrt(xl*xl+yl*yl);
        if (seglen > (3*width))
        {
          //main line
          int ax1 = x1 + xl * width / seglen;
          int ay1 = y1 + yl * width / seglen;
          int ax2 = x1 + xl * 2 * width / seglen;
          int ay2 = y1 + yl * 2 * width / seglen;
          draw_setcolor(0x0000ff);
          draw_line(image, ax1, ay1, ax2, ay2, 1);
          //point for arrow
          ax1 = x1 + xl * 2 * 6 * width / seglen / 7;
          ay1 = y1 + yl * 2 * 6 * width / seglen / 7;
          //first side line
          int xl1 = -yl;
          int yl1 = xl;
          int bx1 = ax1 + xl1 * width / seglen / 5;
          int by1 = ay1 + yl1 * width / seglen / 5;
          draw_line(image, bx1, by1, ax2, ay2, 1);
          //second side line
          xl1 = yl;
          yl1 = -xl;
          bx1 = ax1 + xl1 * width / seglen / 5;
          by1 = ay1 + yl1 * width / seglen / 5;
          draw_line(image, bx1, by1, ax2, ay2, 1);
        }
      }

      //draw way name
      if (names && (way->name) && (fontindex>0) && (fonts[fontindex-1]))
      {
        int xl = x2-x1;
        int yl = y2-y1;
        int seglen = sqrt(xl*xl+yl*yl);
        draw_setfont(fonts[fontindex-1]);
        int textlen = draw_text_width(image, way->name) * 3 / 2;
        if (seglen>textlen)
        {
          double angle = atan2(yl, xl);
          if (angle>pi/2)
            angle -= pi;
          if (angle<-pi/2)
            angle += pi;
          draw_setcolor(img->style->zoom[img->zoom].ways[maptype].textcolor);
          xl = (x1+x2)/2;
          yl = (y1+y2)/2;
          draw_text(image, xl, yl, 0, 0, angle, way->name);
        }
      }
    }
  }
  map_segment_stop(cs);
  //printf("Ways: ");  timer_print(&timer);  timer_start(&timer);

  //TODO sectors ?

  //POIs
  if (img->zoom<12)
  {
    int poiinit = 0;
    maptype = (maptypes)(maxpoitype+1);
    mappoisearch *pos = NULL;
    while (maptype>firstpoitype)
    {
      mpoi *poi;
      long x, y;
      int fontindex = 0;

      maptype = (maptypes)(maptype-1);
      /*printf ("Zoom=%d type=%d enable=%d text=%d\n", 
              img->zoom, maptype,
              img->style->zoom[img->zoom].ways[maptype].enable, img->style->zoom[img->zoom].ways[maptype].text);*/

      if ((img->style->zoom[img->zoom].ways[maptype].enable == 0) && (img->style->zoom[img->zoom].ways[maptype].text == 0))
        continue;

      if (poiinit == 0)
      {
        pos = map_pois_start(lata1, lona1, lata2, lona2);
        poiinit = 1;
      }

      int poiindex = maptype-firstpoitype;
      if ((poi_icon[poiindex].icon == NULL) && (img->style->zoom[img->zoom].ways[maptype].icon))
      {
        //printf("%s\n", img->style->zoom[img->zoom].ways[maptype].icon);
        poi_icon[poiindex].icon = draw_loadimage(img->style->zoom[img->zoom].ways[maptype].icon);
        if (poi_icon[poiindex].icon)
        {
          draw_getsize(poi_icon[poiindex].icon, &poi_icon[poiindex].width, &poi_icon[poiindex].height);
        }
        else
        {
          printf("%s not found\n", img->style->zoom[img->zoom].ways[maptype].icon);
        }
      }

      //load font if needed
      int names = img->style->zoom[img->zoom].ways[maptype].text ? 1 : 0;
      if (names)
      {
        fontindex = img->style->zoom[img->zoom].ways[maptype].font;
        if (fonts[fontindex-1])
        {
          draw_setfont(fonts[fontindex-1]);
        }
        else
        {
          fontindex = 0;
        }
      }

      //if (maptype==POI_ADDRESS)
      //printf("Maptype %d\n", maptype);
      map_pois_type(pos, maptype, names);
      while ((poi = map_pois_get(pos)) != NULL)
      {
        if (img->style->zoom[img->zoom].ways[maptype].enable)
        {
          wgs84_pos2screen(poi->pos.lat, poi->pos.lon, img->lat, img->lon, img->angle, &x, &y);
          x = img->x + x / size;
          y = img->y + y / size;
          if (poi_icon[poiindex].icon)
          {
            //printf("POI %ld,%ld to %ld,%ld\n", poi->pos.lat, poi->pos.lon, x, y);
            draw_blend(image, poi_icon[poiindex].icon, 
                       x-poi_icon[poiindex].width/2, 
                       y-poi_icon[poiindex].height/2, 
                       poi_icon[poiindex].width, poi_icon[poiindex].height);
            y += poi_icon[poiindex].height;
          }
          if (fontindex && names && poi->name)
          {
            draw_setcolor(img->style->zoom[img->zoom].ways[maptype].textcolor);
            draw_text(image, x, y, 0, 0, 0, poi->name);
            //printf("%s\n",poi->name);
          }
        }
      }
    }
    if (poiinit)
    {
      map_pois_stop(pos);
    }
    //printf("POIs: ");  timer_print(&timer);  timer_start(&timer);
  }

  //Locations
  int locinit = 0;
  maptype = (maptypes)(maxloctype+1);
  maplocsearch *los = NULL;
  while (maptype>firstloctype)
  {
    mloc *loc;
    long x, y;
    int fontindex = 0;

    maptype = (maptypes)(maptype-1);

    if ((img->style->zoom[img->zoom].ways[maptype].enable == 0) && (img->style->zoom[img->zoom].ways[maptype].text == 0))
      continue;

    if (locinit == 0)
    {
      los = map_locs_start(lata1, lona1, lata2, lona2);
      locinit = 1;
    }

    //load font if needed
    int names = img->style->zoom[img->zoom].ways[maptype].text ? 1 : 0;
    if (names)
    {
      fontindex = img->style->zoom[img->zoom].ways[maptype].font;
      if (fonts[fontindex-1])
      {
        draw_setfont(fonts[fontindex-1]);
      }
      else
      {
        fontindex = 0;
      }
    }

    //printf("Loctype %d\n", maptype);
    map_locs_type(los, maptype);
    while ((loc = map_locs_get(los)) != NULL)
    {
      if (img->style->zoom[img->zoom].ways[maptype].enable)
      {
        wgs84_pos2screen(loc->pos.lat, loc->pos.lon, img->lat, img->lon, img->angle, &x, &y);
        x = img->x + x / size;
        y = img->y + y / size;
        if (fontindex && names && loc->name)
        {
          draw_setcolor(img->style->zoom[img->zoom].ways[maptype].textcolor);
          draw_text(image, x, y, 0, 0, 0, loc->name);
          //printf("%s\n",loc->name);
        }
      }
    }
  }
  if (locinit)
  {
    map_locs_stop(los);
  }
  //printf("Locs: ");  timer_print(&timer);  timer_start(&timer);

  //draw the route marks
  //TODO sometimes crashes at app exit
  nav_lock();
  if (route)
  {
    if (route->route)
    {
      route_node *node = route->route;

      while (node)
      {
        if (node->direction)
        {
          if (is_inrect(node->pos.lat, node->pos.lon, lata1, lona1, lata2, lona2))
          {
            long x, y;
            wgs84_pos2screen(node->pos.lat, node->pos.lon, img->lat, img->lon, img->angle, &x, &y);
            x = img->x + x / size;
            y = img->y + y / size;
            if (turn_icon[node->direction].icon)
            {
              draw_blend(image, turn_icon[node->direction].icon, 
                         x-turn_icon[node->direction].width/2, 
                         y-turn_icon[node->direction].height/2, 
                         turn_icon[node->direction].width, turn_icon[node->direction].height);
            }
          }
        }
        node = node->next_node;
      }
    }
    if (route->itinerary)
    {
      route_itinerary *ri = route->itinerary;
      while (ri)
      {
        if (ri->nodetype == nt_direction)
        {
          if (is_inrect(ri->pos.lat, ri->pos.lon, lata1, lona1, lata2, lona2))
          {
            long x, y;
            wgs84_pos2screen(ri->pos.lat, ri->pos.lon, img->lat, img->lon, img->angle, &x, &y);
            x = img->x + x / size;
            y = img->y + y / size;
            if (turn_icon[dir_direction].icon)
            {
              draw_blend(image, turn_icon[dir_direction].icon, 
                         x-turn_icon[dir_direction].width/2, 
                         y-turn_icon[dir_direction].height/2, 
                         turn_icon[dir_direction].width, turn_icon[dir_direction].height);
            }
          }
        }
        ri = ri->next;
      }
    }
  }
  nav_unlock();

  //draw track
  gpx_lock();
  if (gpxtrack)
  {
    render_gpx(img, gpxtrack, gpxtrackcolor);
  }
  if (gpxlog)
  {
    render_gpx(img, gpxlog, gpxlogcolor);
  }
  gpx_unlock();

  //draw the anchor
  if (img->anchor)
  {
    if ((pointer.icon==NULL) && img->style->pointer)
    {
      pointer.icon = draw_loadimage(img->style->pointer);
      if (pointer.icon)
      {
        draw_getsize(pointer.icon, &pointer.width, &pointer.height);
      }
    }
    if (pointer.icon)
    {
      //printf("POI %ld,%ld to %ld,%ld\n", poi->pos.lat, poi->pos.lon, x, y);
      draw_blend(image, pointer.icon, 
                 img->x-pointer.width/2, 
                 img->y-pointer.height/2, 
                 pointer.width, pointer.height);
    }
  }
  else
  {
    draw_setcolor(0);
    draw_line(image, img->x-10, img->y, img->x+10, img->y, 1);
    draw_line(image, img->x, img->y-10, img->x, img->y+10, 1);
  }
  printf("Render: ");  timer_print(&timer);

  //imlib_context_set_color(0xa0, 0x00, 0xa0, 0xff);
  //imlib_image_fill_rectangle(img->width/4, img->height/4, img->width/2, img->height/2);
  //imlib_image_set_format("png");
  //Imlib_Load_Error err;
  //imlib_save_image_with_error_return("pokus.png",&err);
  //printf("Ret: %d\n", err);
  return 0;
}

// ************************************
// the rendering thread
// ************************************

void *mr_thread(void *t)
{
  mr_image *img = (mr_image*)t;
  int needrender, needresize;
  int newwidth, newheight;
  long newlat, newlon;
  int newx, newy, newangle, newzoom;
  
  printf("Render start\n");

  img->image1 = draw_create(img->width, img->height);
  img->image2 = draw_create(img->width, img->height);

  while (img->terminate == 0)
  {
    //wait for signal
    mcsema_wait(&img->sema);

    //check the required work
    mcmutex_lock(&img->mutex);
    needresize = (img->newwidth == img->width) && (img->newheight == img->height) ? 0 : 1;
    needrender = (img->rendered == 1) || 
                 ((img->newx == img->x) && (img->newy == img->y) && 
                 (img->newlat == img->lat) && (img->newlon == img->lon) && 
                 (img->newangle == img->angle) && (img->newzoom == img->zoom) &&
                 (img->newroute==0))
                 ? 0 : 1;
    if (needresize)
    {
      newwidth = img->newwidth;
      newheight = img->newheight;
      img->rendered = 0;
      needrender = 1;
    }
    if (needrender)
    {
      newlat = img->newlat;
      newlon = img->newlon;
      newx = img->newx;
      newy = img->newy;
      newangle = img->newangle;
      newzoom = img->newzoom;
      img->newroute = 0;
      img->redraw = 0;
    }
    mcmutex_unlock(&img->mutex);

    //do the work
    if (needresize)
    {
      //printf("Set %dx%d\n", newwidth, newheight);
      if (img->image1)
      {
        draw_destroy(img->image1);
      }
      img->image1 = draw_create(newwidth, newheight);
      if (img->image2)
      {
        draw_destroy(img->image2);
      }
      img->image2 = draw_create(newwidth, newheight);
      img->width = newwidth;
      img->height = newheight;
    }

    if (needrender)
    {
      //printf("Start\n");
      img->x = newx;
      img->y = newy;
      img->lat = newlat;
      img->lon = newlon;
      img->angle = newangle;
      img->zoom = newzoom;
      do_render(img);
      //printf("Stop\n");
      img->rendered = 1;
    }
  }
  printf("Render stop\n");
  if (img->image1)
  {
    draw_destroy(img->image1);
    img->image1 = NULL;
  }
  if (img->image2)
  {
    draw_destroy(img->image2);
    img->image2 = NULL;
  }
  return NULL;
}


// ************************************
// the interface
// ************************************

int mr_resize(mr_image *img, int width, int height, int x, int y)
{
  //printf("Resize\n");
  mcmutex_lock(&img->mutex);
  img->newwidth = width;
  img->newheight = height;
  img->newx = x;
  img->newy = y;
  mcmutex_unlock(&img->mutex);
  mcsema_signal(&img->sema);
  return 0;
}

int mr_render(mr_image *img, map_position *pos, int x, int y, int angle, int zoom, int route)
{
  //printf("Render\n");
  mcmutex_lock(&img->mutex);
  img->newlat = pos->lat;
  img->newlon = pos->lon;
  img->newx = x;
  img->newy = y;
  img->newangle = angle;
  img->newzoom = zoom;
  img->newroute |= route;
  img->redraw = 1;
  mcmutex_unlock(&img->mutex);
  mcsema_signal(&img->sema);
  return 0;
}

int mr_rendered(mr_image *img)
{
  mcmutex_lock(&img->mutex);
  int r = img->rendered;
  mcmutex_unlock(&img->mutex);
  return r;
}

int *mr_getdata(mr_image *img, int *width, int *height)
{
  int *p;
  mcmutex_lock(&img->mutex);
  if (img->rendered)
  {
    img->select ^= 1;
    p = img->select ? draw_getdata(img->image2) : draw_getdata(img->image1);
    img->rendered = 0;
    *width = img->width;
    *height = img->height;
    //printf("Get %d\n", img->select);
  }
  else
  {
    p = NULL;
  }
  mcmutex_unlock(&img->mutex);
  if (p)
    mcsema_signal(&img->sema);
  return p;
}
