/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * navigation.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 *
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "mctime.h"
#include "mcthread.h"
#include "projection.h"
#include "functions.h"
#include "navigation.h"
#include "gpx.h"
#include "sounds.h"
#include "mcconf.h"
#include "route_file.h"

//#define NAVTESTING 1

#ifdef NAVTESTING
#define DRAGTIMEOUT 1000       //time to go to map gps tracking mode after map dragging in miliseconds
#define GPSTIMEOUT 200000      //no gps signal limit for computing position, in miliseconds
#else
#define DRAGTIMEOUT 10000      //time to go to map gps tracking mode after map dragging in miliseconds
#define GPSTIMEOUT 20000       //no gps signal limit for computing position, in miliseconds
#endif

#define NAV_FPS 10
#define ROUTESNAPDISTANCE 35   //maximum distance to route for snapping on roads in meters, should me greater than WAYSNAPDISTANCE
#define WAYSNAPDISTANCE 15     //maximum distance to way for snapping on roads in meters
#define REACHEDDISTANCE 15     //maximum distance to mark node as reached
#define ANGLESUM 3

route_route *route = NULL;
navdata *nav = NULL;

int compgpspos(mgpsdata *gps, map_position *pos, int *angle)  //time in miliseconds
{
  int res;
  mgps_lock(gps);
  if (gps->time.tv_sec)
  {
    long time = timer_get(&gps->time);
    if (time<GPSTIMEOUT)
    {
      long dist = time * gps->speed * 10 / 38;  // should be div 3.6, but drive a little slower, it seems to be smoother
      long x = dist * sin((float)gps->angle * pi / 180.0);
      long y = dist * cos((float)gps->angle * pi / 180.0);
      wgs84_screen2pos(x, y, gps->lat, gps->lon, 0, &pos->lat, &pos->lon);
      *angle = gps->angle;
      res = 0;
    }
    else
    {
      //pos->lat = gps->lat;
      //pos->lon = gps->lon;
      res = -1;
    }
  }
  else
  {
    //pos->lat = gps->lat;
    //pos->lon = gps->lon;
    res = -1;
  }
  mgps_unlock(gps);
  return res;
}

//snap on route and update togo and resttime
int snap_on_route(navdata *nav, route_node *route, map_position *pos)
{
  route_node *node1, *node2, *nodex = NULL;
  map_position posx = {0,0};
  int d, dist, maxdist;
  int cnt, eph, speed;
  route_dir dir = dir_none;
  int len, lenx = 0, l;
  long latx, lonx;
  long time;

  //get position
  mgps_lock(nav->gps);
  time = timer_get(&nav->gps->time);
  eph = nav->gps->eph;
  speed = nav->gps->speed;
  pos->lat = nav->gps->lat;
  pos->lon = nav->gps->lon;
  mgps_unlock(nav->gps);
  if (time>GPSTIMEOUT)
    time = GPSTIMEOUT;

  maxdist = ROUTESNAPDISTANCE+2*eph;
  dist = maxdist;

  //snap to road with last known GPS position
  //if previous position known, then start with last known segment
again:
  if (nav->routepos)
  {
    cnt = 3;
    node1 = nav->routepos;
    dir = nav->route_nextturn;
  }
  else
  {
    node1 = route;
    cnt = -1;
  }
  node2 = node1->next_node;

  //find nearest route segment nodex
  //set lenx=distance to next node
  while (node2 && cnt)
  {
    wgs84_waysnap(pos->lat, pos->lon, node1->pos.lat, node1->pos.lon, node2->pos.lat, node2->pos.lon, &latx, &lonx, &d, &l, &len);
    if (d<dist)
    {
      dist = d;
      lenx = len;
      posx.lat = latx;
      posx.lon = lonx;
      nodex = node1;
    }
    node1 = node2;
    node2 = node2->next_node;
    if (cnt>0)
      cnt--;
  }

  //don't found, try the full search or return
  if (dist>=maxdist)
  {
    if (nav->routepos)
    {
      nav->routepos = NULL;
      goto again; //don't like it, but easiest
    }
    return -1;
  }
  nav->routepos = nodex;

  //move by speed in road for smooth movement
  node1 = nodex;
  node2 = nodex->next_node;
  if ((speed>0)&&(time>0))
  {
    //distance we ride by actual speed
    len = time * speed * 10 / 38000;    // should be div 3.6, but drive a little slower, it seems to be smoother
    d = 0;
    while (node2->next_node && (len>lenx))
    {
      len -= lenx;
      node1 = node2;
      node2 = node1->next_node;
      lenx = wgs84_distance(node1->pos.lat, node1->pos.lon, node2->pos.lat, node2->pos.lon);
      node1->reached = 1;
    }
    //length of segment to compute position
    if (d == 0)
      d = wgs84_distance(node1->pos.lat, node1->pos.lon, node2->pos.lat, node2->pos.lon);
    else
      d = lenx;
    if (d == 0)
      d = 1;
    //distance from end of segment
    lenx -= len;
    if (lenx<0)
      lenx = 0;
    //compute position
    posx.lat = node2->pos.lat + lenx * (node1->pos.lat - node2->pos.lat) / d;
    posx.lon = node2->pos.lon + lenx * (node1->pos.lon - node2->pos.lon) / d;
  }
  nodex = node1;

  //mark reached
  if (lenx<REACHEDDISTANCE)
  {
    node2->reached = 1;
  }
  node1 = nodex;
  while (node1 && (node1->reached == 0))
  {
    node1->reached = 1;
    node1 = node1->prev_node;
  }

  //get direction mark
  node1 = nodex->next_node;
  while (node1)
  {
    if (node1->direction!=dir_none)
    {
      dir = node1->direction;
      break;
    }
    node1 = node1->next_node;
  }
  
  nav->route_nextturn = dir;
  nav->route_toturn = lenx + node2->toturn;
  nav->route_togo = lenx + node2->togo;
  if (node2->speed)
    nav->route_totake = node2->totake + lenx * 36 / node2->speed;
  else
    nav->route_totake = node2->totake + lenx * 36;
  *pos = posx;
  return 0;
}

//snap on way
int snap_on_way(map_position *pos, int eph)
{
  map_position snap;
  int dist;
  if (map_nearestwaysnap(pos, nav->minway, nav->maxway, &snap, &dist) !=0)
    return -1;
  if (dist>(WAYSNAPDISTANCE+eph))
    return -1;

  *pos = snap;
  return 0;
}

//average angle
int angle_avg(int orig, int angle)
{
  int a;
  if (abs(orig-angle)>180)
  {
    if (angle>0)
      angle-=360;
    else
      angle+=360;
  }
  a = (orig*(ANGLESUM-1) + angle)/ANGLESUM;
  if (a>180)
    a-=360;
  if (a<=-180)
    a+=360;
  return a;
}

// ****************************************************
// routing
// ****************************************************

//add new destination to repository
void nav_route_destination(navdata *nav, map_position pos, char *name)
{
  if (route == NULL)
  {
    free(name);
    return;
  }

  routingresult rres;
  route_node *rt;
  route_itinerary *ri, *ri2;

  //find the route
  if (route->route)
  {
    //append to existing route
    route_node *node = route_lastnode(route->route);
    nav_unlock();
    rres = route_find_node2pos(node, &pos, NULL, -1000, config->route_vehicle, config->route_mode, &rt, NULL, &nav->route_prgs);
    nav_lock();
  }
  else
  {
    //route from current position
    nav_unlock();
    rres = route_find_pos2pos(&nav->map_pos, &pos, NULL, -1000, config->route_vehicle, config->route_mode, &rt, NULL, &nav->route_prgs);
    nav_lock();
    route->vehicle = config->route_vehicle;
  }

  nav->minway = way_minallowed(route->vehicle);
  nav->maxway = way_maxallowed(route->vehicle);
  nav->routepos = NULL;

  if (rres == RTS_DEST)
  {
    if (route->route)
    {
      if (route_append(route->route, rt) != NULL)
      {
        ri2 = route_lastitinerary(route->itinerary);
        ri = route_itinerary_new();
        if (name)
        {
          ri->name = name;
        }
        else
        {
          //route_node *rtl = route_lastnode(rt);
          //TODO ri->name = map_segment_name(ri->nodeid, );
        }
        ri->nodetype = nt_destination;
        ri->pos = pos;
        ri->node = route_lastnode(rt);
        ri2->next = ri;
        ri->prev = ri2;
      }
      else
      {
        route_freeroute(rt);
      }
    }
    else
    {
      route->route = rt;
      nav->route_togo = rt->togo;
      nav->route_totake = rt->totake;
      ri = route_itinerary_new();
      ri->name = NULL;
      ri->nodetype = nt_start;
      ri->pos = rt->pos;
      ri->node = rt;
      route->itinerary = ri;
      ri = route_itinerary_new();
      if (name)
      {
        ri->name = name;
      }
      else
      {
        //route_node *rtl = route_lastnode(rt);
        //TODO ri->name = map_segment_name(rtl->wayid);
      }
      ri->nodetype = nt_destination;
      ri->pos = pos;
      ri->node = route_lastnode(rt);
      ri->prev = route->itinerary;
      route->itinerary->next = ri;
    }
  }
  else
  {
    free(name);
  }
}

void nav_route_waypoint(navdata *nav, map_position pos, char *name)
{
  if ((route == NULL) || (route->route==NULL) || (route->itinerary == NULL))
  {
    free(name);
    return;
  }

  //find nearest route section from itinerary (destination/waypoint/direction)
  //TODO may be check if it is unambiguous, otherwise exit and tell the messagebox
  route_itinerary *ri1, *ri2, *rix1 = NULL, *rix2 = NULL;
  int dist=-1;
  ri1 = route->itinerary;
  ri2 = ri1->next;
  while (ri2)
  {
    int d = wgs84_waydistance(pos.lat, pos.lon, ri1->pos.lat, ri1->pos.lon, ri2->pos.lat, ri2->pos.lon);
    if ((d<dist) || (dist<0))
    {
      rix1 = ri1;
      rix2 = ri2;
      dist = d;
    }
    ri1 = ri2;
    ri2 = ri1->next;
  }
  if (dist<0)  //this should not happen, but to be sure...
  {
    return;
  }

  //find first waypoint/destination
  ri1 = rix1;
  while ((ri1->nodetype==nt_direction) && (ri1->prev))
  {
    ri1 = ri1->prev;
  }

  //find second waypoint/destination
  ri2 = rix2;
  while ((ri2->nodetype==nt_direction) && (ri2->next))
  {
    ri2 = ri2->next;
  }

  //reroute
  route_node *nodex1 = ri1->node;
  route_node *nodex2 = ri2->node;
  route_directions *directions;

  //find route from first node to position
  routingresult rres;
  route_node *rt1;
  directions = route_createdirections(ri1, rix1);
  nav_unlock();
  rres = route_find_node2pos(nodex1, &pos, directions, -1000, route->vehicle, config->route_mode, &rt1, NULL, &nav->route_prgs);
  if (rres==RTS_ERROR)
  {
    nav_lock();
    return;
  }

  //find route from position to second node
  route_node *rt2;
  route_node *rt1l = route_lastnode(rt1);
  directions = route_createdirections(rix2, ri2);
  rres = route_find_node2node(rt1l, nodex2, directions, -1000, route->vehicle, config->route_mode, &rt2, NULL, &nav->route_prgs);
  if (rres==RTS_ERROR)
  {
    route_freeroute(rt1);
    nav_lock();
    return;
  }
  nav_lock();

  //append rt2 to rt1
  if (route_append(rt1, rt2) == NULL)
  {
    route_freeroute(rt1);
    route_freeroute(rt2);
    return;
  }
  rt1l->nodetype = nt_waypoint;

  //add itinerary
  route_itinerary *ri = route_itinerary_new();
  if (name)
  {
    ri->name = name;
  }
  else
  {
    //route_node *rtl = route_lastnode(rt1);
    //TODO ri->name = map_wayseg_name(rt1l->wayid);
  }
  ri->nodetype = nt_waypoint;
  ri->node = rt1l;
  ri->pos = pos;
  ri->next = rix2;
  ri->prev = rix1;
  rix1->next = ri;
  rix2->prev = ri;

  //TODO check if new route has only two nodes
  //replace original route section with the new
  //free original segment
  route_node *node1 = nodex1->next_node;
  route_node *node2 = nodex2->prev_node;
  node2->next_node = NULL;
  route_freeroute(node1);

  //relink first node
  nodex1->next_node = rt1->next_node;
  rt1->next_node->prev_node = nodex1;
  free(rt1);

  //relink last node
  rt1l = route_lastnode(rt1);
  nodex2->prev_node = rt1l->prev_node;
  rt1l->prev_node->next_node = nodex2;
  free(rt1l);

  //make turns
  route_maketurn(route->vehicle, route->route, nodex2);

  nav->routepos = NULL;
  nav->route_togo = route->route->togo;
  nav->route_totake = route->route->totake;
}

void nav_route_direction(navdata *nav, map_position pos, char *name)
{
  if ((route == NULL) || (route->route==NULL) || (route->itinerary == NULL))
  {
    free(name);
    return;
  }

  //find nearest route section from itinerary
  route_itinerary *ri, *ri1, *ri2, *rix1 = NULL, *rix2 = NULL;
  int dist=-1;
  ri1 = route->itinerary;
  ri2 = ri1->next;
  while (ri2)
  {
    int d = wgs84_waydistance(pos.lat, pos.lon, ri1->pos.lat, ri1->pos.lon, ri2->pos.lat, ri2->pos.lon);
    if ((d<dist) || (dist<0))
    {
      rix1 = ri1;
      rix2 = ri2;
      dist = d;
    }
    ri1 = ri2;
    ri2 = ri1->next;
  }
  if (dist<0)  //this should not happen, but to be sure...
  {
    return;
  }

  //add new point between
  ri = route_itinerary_new();
  ri->name = name;
  ri->nodetype = nt_direction;
  ri->pos = pos;
  ri->next = rix2;
  ri->prev = rix1;
  rix1->next = ri;
  rix2->prev = ri;

  //find first waypoint/destination
  ri1 = rix1;
  while ((ri1->nodetype==nt_direction) && (ri1->prev))
  {
    ri1 = ri1->prev;
  }

  //find second waypoint/destination
  ri2 = rix2;
  while ((ri2->nodetype==nt_direction) && (ri2->next))
  {
    ri2 = ri2->next;
  }

  //reroute segment between nearest waypoints/destinations
  routingresult rres;
  route_node *rt;
  route_node *nodex1 = ri1->node;
  route_node *nodex2 = ri2->node;
  route_directions *directions = route_createdirections(ri1, ri2);
  nav_unlock();
  rres = route_find_node2node(nodex1, nodex2, directions, -1000, route->vehicle, config->route_mode, &rt, NULL, &nav->route_prgs);
  nav_lock();
  if (rres==RTS_ERROR)
  {
    return;
  }

  //TODO check if new route has only two nodes
  //replace original route section with the new
  //free original segment
  route_node *node1 = nodex1->next_node;
  route_node *node2 = nodex2->prev_node;
  node2->next_node = NULL;
  route_freeroute(node1);

  //relink first node
  nodex1->next_node = rt->next_node;
  rt->next_node->prev_node = nodex1;
  free(rt);

  //relink last node
  route_node *rtl = route_lastnode(rt);
  nodex2->prev_node = rtl->prev_node;
  rtl->prev_node->next_node = nodex2;
  free(rtl);

  //make turns
  route_maketurn(route->vehicle, route->route, nodex2);

  nav->routepos = NULL;
  nav->route_togo = route->route->togo;
  nav->route_totake = route->route->totake;
}

void nav_route_delete(int id)
{
  route_itinerary *ri, *ri1, *ri2;
  int i;

  //get itinerary to delete
  ri = route->itinerary;
  i = 0;
  while (ri && (i<id))
  {
    ri = ri->next;
    i++;
  }

  //find first waypoint/destination
  ri1 = ri->prev;
  while (ri1 && (ri1->nodetype==nt_direction) && (ri1->prev))
  {
    ri1 = ri1->prev;
  }

  //find second waypoint/destination
  ri2 = ri->next;
  while (ri2 && (ri2->nodetype==nt_direction) && (ri2->next))
  {
    ri2 = ri2->next;
  }

  //deleting last destination? stop routing
  if ((ri1 == route->itinerary) && (ri2 == NULL))
  {
    route_freeroute(route->route);
    route->route = NULL;
    nav->route_togo = 0;
    nav->route_totake = 0;
    route_itinerary_free(route->itinerary);
    route->itinerary = NULL;
    return;
  }

  //delete last destination? delete the rest of route
  if (ri2==NULL)
  {
    route_freeroute(ri1->node->next_node);
    ri1->node->next_node = NULL;
    route_itinerary_free(ri1->next);
    ri1->next = NULL;
    return;
  }

  //reroute from ri1 to ri2
  routingresult rres;
  route_node *rt;
  route_node *nodex1 = ri1->node;
  route_node *nodex2 = ri2->node;
  route_directions *directions = route_createdirections(ri1, ri2);
  nav_unlock();
  rres = route_find_node2node(nodex1, nodex2, directions, -1000, route->vehicle, config->route_mode, &rt, NULL, &nav->route_prgs);
  nav_lock();
  if (rres==RTS_ERROR)
  {
    return;
  }

  //delete previous route
  nodex2->prev_node->next_node = NULL;
  route_freeroute(nodex1->next_node);

  //relink first node
  nodex1->next_node = rt->next_node;
  rt->next_node->prev_node = nodex1;
  free(rt);

  //relink last node
  route_node *rtl = route_lastnode(rt);
  nodex2->prev_node = rtl->prev_node;
  rtl->prev_node->next_node = nodex2;
  free(rtl);

  //make turns
  route_maketurn(route->vehicle, route->route, nodex2);

  //delete itinerary
  ri->prev->next = ri->next;
  ri->next->prev = ri->prev;
  free(ri);

  //update data
  nav->routepos = NULL;
  nav->route_togo = route->route->togo;
  nav->route_totake = route->route->totake;  
  
}

// ****************************************************
// thread
// ****************************************************

void *nav_thread(void *t)
{
  navdata *nav = (navdata*)t;
  long timer = 1000 / NAV_FPS;
  int draw_next = 0;
  while (1)
  {
    int rtn = 0;
    int draw = draw_next;
    draw_next = 0;
    //wait for signal
    mcsema_waittimed(&nav->sema, timer);
    if (nav->terminate)
      break;

    nav_lock();

    //routing
    if (nav->route)
    {
      //printf("Route\n");
      if (nav->route_delete)
      {
        nav_route_delete(nav->route_delete);
        nav->route_delete = 0;
      }
      else if (nav->route_add == routeadd_destination)
        nav_route_destination(nav, nav->route_pos, nav->route_name);
      else if (nav->route_add == routeadd_waypoint)
        nav_route_waypoint(nav, nav->route_pos, nav->route_name);
      else if (nav->route_add == routeadd_direction)
        nav_route_direction(nav, nav->route_pos, nav->route_name);

      nav->route = 0;
      nav->route_prgs.progress = -1;
      draw = 1;
      rtn = 1;
    }

    //navigation
    if (nav->map_tracking)
    {
      //read gps data and update map position
      map_position pos = nav->map_pos;
      int angle;
      int snapped = 0;
      
      if (route->route)
      {
        //printf("Snap on route\n");
        if (snap_on_route(nav, route->route, &pos) == 0)
        {
          snapped = 1;
        }
        else
        {
          //not snapped, try to reroute
          //but only if we can snap on way
          //printf("Reroute\n");
          int dist;
          int eph;
          map_position snap;
          mgps_lock(nav->gps);
          pos.lat = nav->gps->lat;
          pos.lon = nav->gps->lon;
          eph = nav->gps->eph;
          mgps_unlock(nav->gps);
          if ((pos.lat!=nav->map_pos.lat) && (pos.lon!=nav->map_pos.lon) && (map_nearestwaysnap(&pos, nav->minway, nav->maxway, &snap, &dist) == 0) && (dist<(WAYSNAPDISTANCE+eph)))
          {
            //re-route by angle if in car or truck
            routingresult rres;
            route_node *rt;
            if ((config->route_vehicle == RV_TRUCK) || (config->route_vehicle == RV_CAR))
            {
              mgps_lock(nav->gps);
              angle = nav->gps->angle;
              mgps_unlock(nav->gps);
            }
            else
            {
              angle = -1000;
            }
            //find first unreached waypoint or destination
            //if NULL, no unreached node
            route_itinerary *ri = route->itinerary;
            while (ri)
            {
              if (((ri->nodetype==nt_waypoint) || (ri->nodetype==nt_destination)) && (ri->node->reached==0))
                break;
              ri = ri->next;
            }
            if (ri)
            {
              //reroute
              nav_unlock();
              rres = route_find_pos2node(&pos, ri->node, NULL, angle, route->vehicle, config->route_mode, &rt, route->route, &nav->route_prgs);
              nav->routepos = NULL;
              nav_lock();
              if ((rres == RTS_DEST) || (rres == RTS_ROUTE))
              {
                //combine routes
                route->route = route_combine(route->vehicle, rt, route->route, ri->node);
                //delete previous itinerary
                route_itinerary *ri2 = ri->prev;
                while (ri2->prev)
                {
                  route_itinerary *ri3 = ri2->prev;
                  free(ri2);
                  ri2 = ri3;
                }
                //update start of itinerary
                ri2->next = ri;
                ri2->prev = NULL;
                ri2->node = route->route;
                ri2->pos = route->route->pos;
                ri->prev = ri2;
                draw = 1;
              }

              //end rerouting
              nav->route_prgs.progress = -1;
              if (snap_on_route(nav, route->route, &pos) == 0)
              {
                snapped = 1;
              }
            }
          }
        }
      }

      //snap on way, if we have not snapped to route
      if (snapped == 0)
      {
        //printf("Snap on way\n");
        if (compgpspos(nav->gps, &pos, &angle) == 0)
        {
          int eph;
          mgps_lock(nav->gps);
          eph = nav->gps->eph;
          mgps_unlock(nav->gps);
          if (snap_on_way(&pos, eph) == 0)
          {
            snapped = 1;
          }
        }
      }

      //sound directives
      if (snapped)
      {
        if ((route->vehicle == RV_FOOT) && (nav->route_toturn_last >= 55) && (nav->route_toturn < 55) && (nav->route_lastsound!=sound_in50m))
        {
          sounds_type sound[2];
          sound[0] = sound_in50m;
          sound[1] = (sounds_type)nav->route_nextturn;
          sounds_play(sound, 2);
          nav->route_lastsound = sound_in50m;
        }
        else if (((route->vehicle == RV_MTB) || (route->vehicle == RV_CYCLE)) && (nav->route_toturn_last >= 110) && (nav->route_toturn < 110) && (nav->route_lastsound!=sound_in100m))
        {
          sounds_type sound[2];
          sound[0] = sound_in100m;
          sound[1] = (sounds_type)nav->route_nextturn;
          sounds_play(sound, 2);
          nav->route_lastsound = sound_in100m;
        }
        else if (((route->vehicle == RV_CAR) || (route->vehicle == RV_TRUCK)) && (nav->route_toturn_last >= 220) && (nav->route_toturn < 220) && (nav->route_lastsound!=sound_in200m))
        {
          sounds_type sound[2];
          sound[0] = sound_in200m;
          sound[1] = (sounds_type)nav->route_nextturn;
          sounds_play(sound, 2);
          nav->route_lastsound = sound_in200m;
        }
        else if (((route->vehicle == RV_CAR) || (route->vehicle == RV_TRUCK)) && (nav->route_toturn_last >= 1050) && (nav->route_toturn < 1050) && (nav->route_lastsound!=sound_in1km))
        {
          sounds_type sound[2];
          sound[0] = sound_in1km;
          sound[1] = (sounds_type)nav->route_nextturn;
          sounds_play(sound, 2);
          nav->route_lastsound = sound_in1km;
        }
        else if (((route->vehicle == RV_FOOT) && (nav->route_toturn_last >= 20) && (nav->route_toturn < 20)) ||
                 (((route->vehicle == RV_MTB) || (route->vehicle == RV_CYCLE)) && (nav->route_toturn_last >= 30) && (nav->route_toturn < 30)) ||
                 (((route->vehicle == RV_CAR) || (route->vehicle == RV_TRUCK)) && (nav->route_toturn_last >= 40) && (nav->route_toturn < 40)))
        {
          sounds_type sound[1];
          sound[0] = (sounds_type)nav->route_nextturn;
          sounds_play(sound, 1);
          nav->route_lastsound = sound_none;
        }
        nav->route_toturn_last = nav->route_toturn;
      }

      if ((nav->map_pos.lat != pos.lat) || (nav->map_pos.lon != pos.lon))
      {
        mgps_lock(nav->gps);
        angle = nav->gps->angle;
        mgps_unlock(nav->gps);
        nav->map_angle = angle_avg(nav->map_angle, angle);
        //nav->map_angle = angle;
        nav->map_pos = pos;
        //printf("GPS %ld,%ld %d\n", nav->map_pos.lat, nav->map_pos.lon, nav->map_angle);
        draw = 1;
      }
    }
    else
    {
      //check timer if we want to go to tracking mode
      if ((nav->map_tracking_disable==0) && (nav->map_dragtime.tv_sec) && (nav->gps->time.tv_sec))
      {
        long gt = mgps_timer(nav->gps);
        if ((timer_get(&nav->map_dragtime) > DRAGTIMEOUT) && (gt <= GPSTIMEOUT))
        {
          nav->map_tracking = 1;
          nav->map_img.anchor = 1;
          nav->map_x = nav->map_width / 2;
          nav->map_y = nav->map_height * 9 / 10;
          draw_next = 1;
        }
      }
    }
    nav_unlock();

    //render map if needed
    if (draw)
    {
      mr_render(&nav->map_img, &nav->map_pos, nav->map_x, nav->map_y, -nav->map_angle, nav->map_zoom, rtn);
    }
  }
  printf("Navigation end\n");
  return NULL;
}


// ****************************************************
// init & deinit
// ****************************************************

int nav_init()
{
  //create the map
  nav = (navdata*)malloc(sizeof(navdata));
  memset(nav, 0, sizeof(navdata));
  nav->route_prgs.progress = -1;
  timer_start(&nav->map_dragtime);

  //check home position is on the map
  nav->map_pos = config->home;
  /*if (map_segindexinfo(nav->map_pos.lat / sectrect * sectrect, nav->map_pos.lon / sectrect * sectrect, NULL, NULL) != 0)
  {
    map_segindexpos(-1, &nav->map_pos);
  }*/

  nav->map_angle = 0;
  nav->map_zoom = 9;

  nav->map_img.style = mapstyle_get("Default");
  nav->map_img.newlat = nav->map_pos.lat;
  nav->map_img.newlon = nav->map_pos.lon;
  nav->map_img.newangle = -nav->map_angle;
  nav->map_img.newzoom = nav->map_zoom;
  mr_create(&nav->map_img);

  //init sounds
  sounds_init(config->sounds_style);
  
  //init gps
  nav->gps = mgps_init();
  nav->routepos = NULL;

  //init routing
  route = route_init();
  nav->minway = way_minallowed(config->route_vehicle);
  nav->maxway = way_maxallowed(config->route_vehicle);

  //create navigation thread
  mcsema_init(&nav->sema);
  mcmutex_init(&nav->mutex);
  mcthread_create(&nav->thread, nav_thread, nav, 1);
  return 0;
}

int nav_deinit()
{
  nav->terminate = 1;
  mcsema_signal(&nav->sema);
  mcthread_join(&nav->thread);
  printf("Navigation gone\n");
  mcsema_destroy(&nav->sema);
  mcmutex_destroy(&nav->mutex);
  mgps_deinit(nav->gps);
  sounds_deinit();
  gpx_deinit();
  nav->gps = NULL;
  route_deinit(route);
  mapstyle_free(nav->map_img.style);
  mr_free(&nav->map_img);
  free(nav);
  return 0;
}


// ****************************************************
// control
// ****************************************************

int nav_lock()
{
  return mcmutex_lock(&nav->mutex);
}

int nav_unlock()
{
  return mcmutex_unlock(&nav->mutex);
}

int nav_route(map_position *pos, routing_add add, char *name)
{
  if (pos==NULL)
    return -1;
  
  nav_lock();
  if (nav->route)
  {
    nav_unlock();
    return -1;
  }

  nav->route_pos = *pos;
  nav->route = 1;
  nav->route_add = add;
  nav->route_name = name;
  nav_unlock();
  mcsema_signal(&nav->sema);
  return 0;
}

int nav_delete(int itinerary)
{
  if (itinerary == 0)
    return -1;

  nav_lock();
  if (nav->route)
  {
    nav_unlock();
    return -1;
  }
  nav->route_delete = itinerary;
  nav->route = 1;
  
  nav_unlock();
  mcsema_signal(&nav->sema);
  return 0;
}

int nav_stop()
{
  nav_lock();
  if (nav->route)
  {
    route_stop();
    nav->route = 0;
  }
  if (route->route)
  {
    route_freeroute(route->route);
    route->route = NULL;
    nav->route_togo = 0;
    nav->route_totake = 0;
  }
  if (route->itinerary)
  {
    route_itinerary_free(route->itinerary);
    route->itinerary = NULL;
  }
  timer_start(&nav->map_dragtime);
  mr_render(&nav->map_img, &nav->map_pos, nav->map_x, nav->map_y, -nav->map_angle, nav->map_zoom, 1);
  nav_unlock();
  return 0;
}

int nav_save(char *filename)
{
  nav_lock();
  if (route)
  {
    char *fullname = (char*)malloc(strlen(config->dir_routes)+strlen(filename)+6);
    sprintf(fullname, "%s/%s.gpx", config->dir_routes, filename);
    route_save(fullname, route);
  }
  nav_unlock();
  return 0;
}

int nav_load(char *filename)
{
  route_route *rt;
  nav_lock();
  rt = route_load(filename);
  if (rt)
  {
    if (route->route)
    {
      route_freeroute(route->route);
    }
    if (route->itinerary)
    {
      route_itinerary_free(route->itinerary);
    }
    free(route);
    route = rt;
    nav->route_togo = route->route->togo;
    nav->route_totake = route->route->totake;
    mr_render(&nav->map_img, &nav->map_pos, nav->map_x, nav->map_y, -nav->map_angle, nav->map_zoom, 1);
  }
  nav_unlock();
  return 0;
}

int nav_progress(route_progress *progress)
{
  route_lock();
  *progress = nav->route_prgs;
  route_unlock();
  return 0;
}

int nav_getpos(map_position *pos, int *angle)
{
  nav_lock();
  *pos = nav->map_pos;
  *angle = -nav->map_angle;
  nav_unlock();
  return 0;
}

int nav_setpos(map_position *pos, int angle)
{
  nav_lock();
  nav->map_pos = *pos;
  nav->map_tracking = 0;
  nav->map_angle = -angle;
  nav->map_img.anchor = 0;
  timer_start(&nav->map_dragtime);
  mr_render(&nav->map_img, &nav->map_pos, nav->map_x, nav->map_y, -nav->map_angle, nav->map_zoom, route!=NULL);
  nav_unlock();

  //testing
#ifdef NAVTESTING
  timer_start(&nav->gps->time);
  nav->gps->lat = pos->lat;
  nav->gps->lon = pos->lon;
  nav->gps->eph = 5;
  nav->gps->speed = 30;
  nav->gps->angle = 0;
#endif
  /*if (nav->gps->angle==0)
    nav->gps->angle = 90;
  else
    nav->gps->angle = -nav->gps->angle;*/

  return 0;
}

int nav_getzoom(void)
{
  int zoom;
  nav_lock();
  zoom = nav->map_zoom;
  nav_unlock();
  return zoom;
}

int nav_setzoom(int zoom)
{
  nav_lock();
  nav->map_zoom = zoom;
  mr_render(&nav->map_img, &nav->map_pos, nav->map_x, nav->map_y, -nav->map_angle, nav->map_zoom, route!=NULL);
  nav_unlock();
  return 0;
}

int nav_setsize(int width, int height)
{
  nav_lock();
  nav->map_width = width;
  nav->map_height = height;
  if (nav->map_tracking)
  {
    nav->map_x = width / 2;
    nav->map_y = height * 9 / 10;
  }
  else
  {
    nav->map_x = width / 2;
    nav->map_y = height / 2;
  }
  mr_resize(&nav->map_img, nav->map_width, nav->map_height, nav->map_x, nav->map_y);
  nav_unlock();
  return 0;
}

int nav_getanchor(int *x, int *y)
{
  *x = nav->map_x;
  *y = nav->map_y;
  return 0;
}

int nav_gpsinfo(int *fix, long *lat, long *lon, int *speed, int *alt, int *angle, int *eph, int *epv)
{
  nav_lock();
  if (nav->gps == NULL)
  {
    nav_unlock();
    return -1;
  }
  *fix = nav->gps->fix;
  *eph = nav->gps->eph;
  *epv = nav->gps->epv;
  if ((nav->gps->time.tv_sec==0) || (timer_get(&nav->gps->time)>GPSTIMEOUT))
  {
    nav_unlock();
    return -1;
  }
  *lat = nav->gps->lat;
  *lon = nav->gps->lon;
  *speed = nav->gps->speed;
  *alt = nav->gps->alt;
  *angle = nav->gps->angle;
  nav_unlock();
  return 0;
}

//call nav_unlock after use
int nav_satinfo(int *sats_num, mgpssat **sats)
{
  nav_lock();
  if (nav->gps == NULL)
  {
    nav_unlock();
    return -1;
  }
  *sats_num = nav->gps->sats_num;
  *sats = nav->gps->sats;

  return 0;
}

int nav_routeinfo(route_dir *nextturn, long *toturn, long *togo, long *totake)
{
  nav_lock();
  if ((route==NULL) || (route->route==NULL))
  {
    nav_unlock();
    return -1;
  }
  *nextturn = nav->route_nextturn;
  *toturn = nav->route_toturn;
  *togo = nav->route_togo;
  *totake = nav->route_totake;
  nav_unlock();
  return 0;
}

int nav_tracking(int tracking)
{
  nav_lock();
  if ((tracking == 0) && ((nav->map_x!=nav->map_width/2) || (nav->map_y!=nav->map_height/2)))
  {
    //recalculate map center
    int x = nav->map_x - nav->map_width/2;
    int y = nav->map_y - nav->map_height/2;
    long lat, lon;
    long size = 1 << nav->map_zoom;
    wgs84_screen2pos(x*size, y*size, nav->map_pos.lat, nav->map_pos.lon, -nav->map_angle, &lat, &lon);
    nav->map_pos.lat = lat;
    nav->map_pos.lon = lon;
    nav->map_x = nav->map_width/2;
    nav->map_y = nav->map_height/2;
  }
  nav->map_tracking_disable = tracking ? 0 : 1;
  if (tracking == 0)
    nav->map_tracking = 0;
  else
    timer_start(&nav->map_dragtime);
  nav_unlock();
  return 0;
}

int nav_setvehicle(routing_vehicle vehicle)
{
  nav_lock();
  if ((route==NULL)||(route->route==NULL))
  {
    nav->minway = way_minallowed(vehicle);
    nav->maxway = way_maxallowed(vehicle);
  }
  nav_unlock();
  return 0;
}
