/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * route.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 *
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <expat.h>
#include "safefile.h"
#include "route.h"
#include "map_read.h"
#include "projection.h"
#include "mctime.h"
#include "mcthread.h"
#include "mcconf.h"
#include "gpx.h"

#include "route_sector.h"

#define minpassdist 5000
#define dirindexlen 1000
#define distforsect 30000          //min distance for using route_sector

#define BUFF_SIZE 1000

typedef struct
{
  route_node *start;               //start node - start of routing, this is typicly false node on any way segment
  route_node *dest1;               //destination node - the first node that leads to goal node
  route_node *dest2;               //destination node - the second node that leads to goal node
  route_node *goal;                //goal end node - this is typicly false node on any way segment
  route_directions *directions;    //routing directions
  route_endnodes endnodes;         //list of end nodes sorted by cost
  route_visited visited;           //list of visited node
  routing_vehicle vehicle;
  routing_mode mode;
  int nodes_total;
  int nodes;
  routingresult status;
  int angle;                        //requested ride angle, will be checked to first way segment angle
  long maxdist;
  long maxdistid;
} routing_info;

typedef struct
{
  mcthread thread;
  mcmutex mutex;
  mcsema sema;
  int terminate;
} routing_data;

routing_data *routed = NULL;

int fixangle(int angle)
{
  while (angle>180)
    angle -= 360;
  while (angle<-180)
    angle += 360;
  return angle;
}

//direction distance
int directions_distance(route_directions *directions, map_position *pos)
{
  int i;
  int dist = -1;
  for (i=1; i<directions->count; i++)
  {
    int d = wgs84_waydistance(pos->lat, pos->lon, directions->pos[i-1].lat, directions->pos[i-1].lon, directions->pos[i].lat, directions->pos[i].lon);
    if ((dist<0) || (d<dist))
    {
      dist = d;
    }
  }
  return dist;
}

//extend route directions with start and end position
route_directions *append_directions(map_position *start, map_position *stop, route_directions *directions)
{
  if (directions==NULL)
    return NULL;

  int count = directions->count + 2;
  route_directions *rd = (route_directions *)malloc(sizeof(route_directions)+sizeof(map_position)*count);

  rd->count = count;
  rd->pos[0] = *start;
  rd->pos[count-1] = *stop;
  memcpy(&rd->pos[1], directions->pos, directions->count*sizeof(map_position));
  return rd;
}

//compute direction cost index (0-100)
int directions_index(route_directions *directions, map_position *pos)
{
  if (directions==NULL)
    return dirindexlen;

  int i, idx;
  int d, dist = -1, maxdist=-1;
  for (i=1; i<directions->count; i++)
  {
    //distance of my position to direction segment
    d = wgs84_waydistance(pos->lat, pos->lon, 
                          directions->pos[i-1].lat, directions->pos[i-1].lon, 
                          directions->pos[i].lat, directions->pos[i].lon);
    if ((dist<0) || (d<dist))
    {
      dist = d;
    }
    //maximum distance of direction point from start/stop line
    if (i<directions->count-1)
    {
      d = wgs84_waydistance(directions->pos[i].lat, directions->pos[i].lon, 
                            directions->pos[0].lat, directions->pos[0].lon, 
                            directions->pos[directions->count-1].lat, directions->pos[directions->count-1].lon);
      if (d>maxdist)
      {
        maxdist = d;
      }
    }
  }

  double dd = (double)dist/(double)maxdist*3;
  dd = dirindexlen * atan(dd) / pi * 2 + 0.5;
  idx = dirindexlen - trunc(dd);
/*  int dist = directions_distance(directions, pos);
  if (dist<1000)
    idx = 100;
  else
    idx = 100000 / dist;*/

  return idx;
}

long destination_distance(map_position *pos, map_position *destination, route_directions *directions)
{
  long togo;
  //printf("%ld,%ld->%ld,%ld\n", pos->lat, pos->lon, destination->lat, destination->lon);
  //togo = wgs84_distance(pos->lat, pos->lon, destination->lat, destination->lon);
  /*togo = wgs84_distance(pos->lat, pos->lon, destination->lat, destination->lon);
  int idx = directions_index(directions, pos);
  togo = idx * dirindexlen / togo;*/
  togo = wgs84_distance(pos->lat, pos->lon, destination->lat, destination->lon);
/*
  if (directions == NULL)
  {
    //simple distance
    togo = wgs84_distance(pos->lat, pos->lon, destination->lat, destination->lon);
  }
  else
  {
    //distance to nearest direction segment
    int i, j;
    int d, dist = -1;
    for (i=1; i<directions->count; i++)
    {
      d = wgs84_waydistance(pos->lat, pos->lon, directions->pos[i-1].lat, directions->pos[i-1].lon, directions->pos[i].lat, directions->pos[i].lon);
      //printf("%d = %ld,%ld->%ld,%ld\n", d, directions->pos[i-1].lat, directions->pos[i-1].lon, directions->pos[i].lat, directions->pos[i].lon);
      if ((dist<0) || (d<dist))
      {
        dist = d;
        j = i;
      }
    }

    //distance to destination over all remaining direction marks
    togo = dist;
    d = wgs84_waysection(pos->lat, pos->lon, directions->pos[j-1].lat, directions->pos[j-1].lon, directions->pos[j].lat, directions->pos[j].lon);
    dist = wgs84_distance(directions->pos[j-1].lat, directions->pos[j-1].lon, directions->pos[j].lat, directions->pos[j].lon);
    togo += dist - d;
    for (i=j+1; i<directions->count; i++)
    {
      togo += wgs84_distance(directions->pos[i-1].lat, directions->pos[i-1].lon, directions->pos[i].lat, directions->pos[i].lon);
    }
  }*/
  return togo;
}


// ********************* make turn direction **************************

route_dir angle2dir(int angle)
{
  angle = fixangle(angle);
  if (angle<-120)
    return dir_hardleft;
  else if (angle<-60)
    return dir_left;
  else if (angle<-23)
    return dir_softleft;
  else if (angle<23)
    return dir_straight;
  else if (angle<60)
    return dir_softright;
  else if (angle<120)
    return dir_right;
 else
    return dir_hardright;
}

int make_turn(route_node *node, int angle, int roundabout)
{
  if (roundabout)
  {
    node->direction = angle2dir(angle);
    if (roundabout == 2)
      node->direction = (route_dir)((int)node->direction + (int)dir_rl_straight - 1);
    else
      node->direction = (route_dir)((int)node->direction + (int)dir_rr_straight - 1);
  }
  else if (node->nodetype == nt_waypoint)
  {
    node->direction = dir_waypoint;
  }
  else if ((node->waytype!=node->next_node->waytype) &&
      ((node->waytype == WT_MOTORWAY)||(node->waytype == WT_TRUNK)))
  {
    //always mark leaving highways, we go always right or left - depends on driving side
    //if (rideonright)
      node->direction = dir_softright;
    //else
    //  node->direction = dir_softleft;
  }
  else if ((node->waytype!=node->next_node->waytype) &&
           ((node->next_node->waytype == WT_MOTORWAYLINK)||(node->next_node->waytype == WT_TRUNKLINK)))
  {
    //always mark turning to highways
    node->direction = angle2dir(angle);
    /*if (angle>0)
      node->direction = dir_softright;
    else
      node->direction = dir_softleft;*/
  }
  else if ((node->waytype!=node->next_node->waytype) &&
           ((node->next_node->waytype == WT_MOTORWAY)||(node->next_node->waytype == WT_TRUNK)))
  {
    //do not mark going to highways
    node->direction = dir_none;
  }
  else if (node->cross>2)
  {
    //analyze cross to create best turn direction
    int i;
    mcross *cross;
    int crosscount;
    node->direction = dir_none;
    if (map_nodecross(node->id, &node->pos, &crosscount, &cross)==0)
    {
      int cnt_same = 0;
      int cnt_higher = 0;
      int cnt_lower = 0;
      maptypes highest = node->waytype;
      for (i=0; i<crosscount; i++)
      {
        if (cross[i].waytype==node->waytype)
          cnt_same++;
        if (cross[i].waytype<node->waytype)
          cnt_higher++;
        if (cross[i].waytype>node->waytype)
          cnt_lower++;
        if (highest>cross[i].waytype)
          highest = cross[i].waytype;
      }
      //we pass higher roads, always make direction, if we don't pass links
      if ((cnt_higher>0) && (((highest!=WT_PRIMARYLINK) && (highest!=WT_TRUNKLINK) && (highest!=WT_MOTORWAYLINK)) || (abs(angle)>15)))
      {
        node->direction = angle2dir(angle);
      }
      //we ride on the same road type, do not make direction
      else if ((cnt_same == 2) && (node->waytype==node->next_node->waytype) && (abs(angle)<60))
        node->direction = dir_none;
      //more roads with the same type, make a direction
      else if ((cnt_same > 2) && (node->waytype==node->next_node->waytype))
      {
        node->direction = angle2dir(angle);
        if (node->direction == dir_straight)
          node->direction = dir_none;
      }
      else
        node->direction = angle2dir(angle);
      map_nodeunlock(node->id, &node->pos);
    }
    //printf("Node %ld %d %d dir %d %ld\n", node->id, angle, angle2, node->direction, node->toturn);
  }
  else
  {
    node->direction = dir_none;
  }
  return 0;
}

int route_maketurn(routing_vehicle vehicle, route_node *route, route_node *end)
{
  if (route == NULL)
    return -1;

  if (end == NULL)
  {
    end = route_lastnode(route);
  }

  int angle = 0;
  long toturn = end->toturn;
  long togo = end->togo;
  long totake = end->totake;
  int speed = end->speed;
  int roundabout = 0;

  if (end->nodetype == nt_destination)
  {
    end->direction = dir_finish;
  }
  else if (end->next_node)
  {
    end = end->next_node;
  }

	if (end->flags & WF_FORD)
		speed = 1;
	else
    vehiclespeed(vehicle, end->waytype, &speed);
  map_position pos = end->pos;
  end = end->prev_node;
  angle = wgs84_angle(end->pos.lat, end->pos.lon, pos.lat, pos.lon);
  while (end && end->nodetype!=nt_destination)
  {
    long len = wgs84_distance(end->pos.lat, end->pos.lon, pos.lat, pos.lon);
    if (end->speed != 0)
      speed = end->speed;
    if (end->flags & WF_FORD)
      speed = 1;
    else
      vehiclespeed(vehicle, end->waytype, &speed);
    if (end->prev_node)
    {
      //mark routing nodes
      route_node *node = end->prev_node;
      if (end->flags & WF_ROUNDABOUT)
      {
        roundabout = 1;
      }
      else
      {
        int angle2 = wgs84_angle(node->pos.lat, node->pos.lon, end->pos.lat, end->pos.lon);
        int a = fixangle(angle-angle2);
        make_turn(end, a, roundabout);
        angle = angle2;
        roundabout = 0;
      }
    }
    else
    {
      end->direction = dir_none;
    }
    toturn = (end->direction != dir_none) && (end->direction != dir_waypoint) ? 0 : toturn + len;
    togo += len;
    totake += len * 36 / speed;
    end->toturn = toturn;
    end->togo = togo;
    end->totake = totake;
    pos = end->pos;
    //printf("Node %ld %d dir %d %ld\n", end->id, angle, end->direction, end->toturn);
    end = end->prev_node;
  }
  return 0;
}

// *********************** last node **************************
route_node *route_lastnode(route_node *route)
{
  if (route == NULL)
    return NULL;

  route_node *node = route;
  while (node->next_node)
  {
    node = node->next_node;
  }

  return node;
}

// ********************* reached route **************************
//returns, if the node is part of the route
route_node * reached_route(long nodeid, route_node *route)
{
  while (route)
  {
    if (route->id == nodeid)
      return route;
    route = route->next_node;
  }
  return NULL;
}

// ********************* route step **************************
int route_step(routing_info *routing, route_node *node)
{
  int i;

  //printf("Node %ld Cost %ld Len %ld\n", node->id, node->cost, node->length);

  //find all ways from this node
  mcross *cross = NULL;
  routecross *rcross = NULL;
  int crosscount = 0;;
  mturn *turn = NULL;
  int turncount = 0;

  //max distance from direct start-goal abscisse
  long d = wgs84_waydistance(node->pos.lat, node->pos.lon, routing->start->pos.lat, routing->start->pos.lon, routing->goal->pos.lat, routing->goal->pos.lon);
  if (d>routing->maxdist)
  {
    routing->maxdist = d;
    routing->maxdistid = node->id;
  }

  //get cross
  map_position nsec, rsec, gsec;
  nsec.lat = MAPSECTOR(node->pos.lat);
  nsec.lon = MAPSECTOR(node->pos.lon);
  rsec.lat = MAPSECTOR(routing->start->pos.lat);
  rsec.lon = MAPSECTOR(routing->start->pos.lon);
  gsec.lat = MAPSECTOR(routing->goal->pos.lat);
  gsec.lon = MAPSECTOR(routing->goal->pos.lon);

  //TODO what about +-180 cross ?
  /*if ((abs(nsec.lat-rsec.lat)<=MAPSECTRECT) || (abs(nsec.lon-rsec.lon)<=MAPSECTRECT) ||
      (abs(nsec.lat-gsec.lat)<=MAPSECTRECT) || (abs(nsec.lon-gsec.lon)<=MAPSECTRECT))
  {
    //printf("Node\n");
    if (map_nodecross(node->id, &node->pos, &crosscount, &cross)<0)
    {
      printf("Warning: cross for node %ld not found\n", node->id);
    }
  }
  else
  {
    printf("Sector %ld,%ld\n", node->pos.lat, node->pos.lon);
    route_nodecross(node->id, &node->pos, &crosscount, &rcross);
  }*/
  
  if ((config->route_sector==0) || (node->length<distforsect) || (node->togo<distforsect))
  {
    if (map_nodecrossturn(node->id, &node->pos, &crosscount, &cross, &turncount, &turn)<0)
    {
      routing->status = RTS_ERROR;
      printf("ERROR: cross for node %ld not found\n", node->id);
      return -1;
    }
  }
  else
  {
    if (route_nodecross(node->id, &node->pos, &crosscount, &rcross)!=0)
    {
      if (map_nodecross(node->id, &node->pos, &crosscount, &cross)!=0)
      {
        routing->status = RTS_ERROR;
        printf("ERROR: cross for node %ld not found\n", node->id);
        return -1;
      }
    }
    else
    {
      //printf("OK\n");
    }
  }

  node->cross = crosscount;
  node->direction = dir_none;
  //if (node->id == 359928492)
  //  printf("Stop\n");

  //add as visited
  //add_visited(&routing->visited, node->prev_node->wayid, node->wayid);
  add_visited(&routing->visited, node->prev_node->id, node->id);

  //create nodes for it and update the end list
  remove_endnode(&routing->endnodes, node);
  route_node *node3 = NULL;
  for (i=0; i<crosscount; i++)
  {
    route_node *node2;
    long crossid = cross ? cross[i].id : rcross[i].id;
    int speed = cross ? cross[i].speed : rcross[i].speed;
    unsigned int flags = cross ? cross[i].flags : rcross[i].flags;
    int waydir = flags & WF_DIR;
    maptypes waytype = cross ? cross[i].waytype : rcross[i].waytype;
    map_position endpos;
    int length;

    //do not go back
    //TODO allow turn back on crossroads without semaphore
    if ((cross!=NULL) && (node->prev_node) && (cross[i].id == node->prev_node->id))
    {
      continue;
    }
    
    //disallow forbidden routing
    if ((routing->vehicle!=RV_FOOT) && (flags & WF_ONEWAY) && waydir)
      continue;
    if ((routing->vehicle==RV_CAR) && (flags & WF_NOMOTORCAR))
      continue;
    if ((routing->vehicle==RV_TRUCK) && (flags & WF_NOTRUCKS))
      continue;
    if ((routing->vehicle==RV_CYCLE) && (flags & WF_NOBICYCLE))
      continue;
    if ((routing->vehicle<=RV_FOOT) && (flags & WF_NOFOOT))
      continue;
    if ((flags & WF_TOLL) && (routing->mode!=routemode_toll) && ((routing->vehicle==RV_TRUCK)||(routing->vehicle==RV_CAR)))
    //if (((waytype == WT_MOTORWAY) || (waytype == WT_TRUNK)) && (routing->mode!=routemode_toll) && ((routing->vehicle==RV_TRUCK)||(routing->vehicle==RV_CAR)))
      continue;
		//TODO ask to allow water fords
    if (((routing->vehicle==RV_TRUCK) || (routing->vehicle==RV_CAR)) && (flags & WF_FORD))
			continue;
    if (way_allowed(routing->vehicle, waytype) == 0)
      continue;
    if (((routing->vehicle==RV_TRUCK)||(routing->vehicle==RV_CAR)) && (node->length>3000) && (node->togo>3000) && /*(node->waytype<=WT_TERTIARY) &&*/ (waytype>WT_TERTIARY))
      continue;
    if (((routing->vehicle==RV_TRUCK)||(routing->vehicle==RV_CAR)) && (node->length>15000) && (node->togo>15000) && /*(node->waytype<=WT_SECONDARY) &&*/ (waytype>WT_SECONDARY))
      continue;
    if (cross && (map_nodeposition(cross[i].id, &node->pos, &endpos)!=0))
      continue;
    if (rcross && (route_nodeposition(rcross[i].id, &node->pos, &endpos)!=0))
      continue;
    //check that we don't ride from link in bad direction - check angle of link
    if ((node->prev_node) &&
        (((node->waytype == WT_MOTORWAYLINK) && (waytype == WT_MOTORWAY)) ||
         ((node->waytype == WT_TRUNKLINK) && (waytype == WT_TRUNK)) ||
         ((node->waytype == WT_PRIMARYLINK) && (waytype == WT_PRIMARY)) ||
         ((node->waytype == WT_MOTORWAY) && (waytype == WT_MOTORWAYLINK)) ||
         ((node->waytype == WT_TRUNK) && (waytype == WT_TRUNKLINK)) ||
         ((node->waytype == WT_PRIMARY) && (waytype == WT_PRIMARYLINK))))
    {
      int angle = wgs84_angle(endpos.lat, endpos.lon, node->pos.lat, node->pos.lon);
      int angle2 = wgs84_angle(node->pos.lat, node->pos.lon, node->prev_node->pos.lat, node->prev_node->pos.lon);
      int a = fixangle(angle-angle2);
      if (abs(a)>120)
        continue;
    }
    //check turn restriction
    if ((turncount>0) && turn)
    {
      int allow = 1;
      for (int r=0; r<turncount; r++)
      {
        if (turn[r].flags & TRF_ALLOW)
        {
          //this will be the only allowed way, do not allow other ways
        }
        else
        {
          //this is turn restriction
          if (turn[r].nodes[0] == crossid)
          {
            //printf("Got it\n");
            //check the way
            route_node *tnode = node->prev_node;
            for (int t=1; t<turn[r].count; t++)
            {
              if ((tnode == NULL) || (tnode->id != turn[r].nodes[t]))
              {
                allow = 1;
                break;
              }
              allow = 0;
              tnode = tnode->prev_node;
            }
            if (allow == 0)
            {
              break;
            }
          }
        }
      }
      if (allow == 0)
        continue;
    }
    

    length = wgs84_distance(node->pos.lat, node->pos.lon, endpos.lat, endpos.lon);
    //printf("C %i %ld %ld (%ld,%ld) -> %ld (%ld,%ld)\n", i, cross->segments[i], startnode, startpos.lat, startpos.lon, endnode, endpos.lat, endpos.lon);

    //check if we have found the destination, update data, link it and exit
    if (((node->id==routing->dest1->id) && (crossid == routing->dest2->id)) ||
        ((node->id==routing->dest2->id) && (crossid == routing->dest1->id)))
    {
      printf("We have reached the destination! (%ld)\n", cross[i].id);
      routing->status = RTS_DEST;

      //update goal node
      routing->goal->cost_src = node->cost_src + cost(routing->vehicle, routing->mode, routing->goal->waytype, node->length, speed);
      routing->goal->cost= routing->goal->cost_src;
      routing->goal->length = node->length + wgs84_distance (node->pos.lat, node->pos.lon, routing->goal->pos.lat, routing->goal->pos.lon);
      routing->goal->prev_node = node;
      routing->dest1->length = node->length + length;
      if (node->next_node==NULL)
      {
        node->next_node = routing->goal;
      }
      else
      {
        node3->next_list = routing->goal;
      }
      routing->nodes++;
      routing->nodes_total++;
      free(routing->dest1);
      routing->dest1 = NULL;
      free(routing->dest2);
      routing->dest2 = NULL;
      printf_nodeinfo(routing->goal);
      break;
    }

    //route to next node
    //if (is_visited(routing, node->wayid, cross->segments[i]) == 0)
    if (is_visited(&routing->visited, node->id, crossid) == 0)
    {
      node2 = new_node();
      node2->id = crossid;
      node2->pos = endpos;
      node2->cross = 0;
      node2->flags = flags;
      node2->direction = dir_none;
      node2->length = node->length + length;
      node2->speed = speed;
      node2->waytype = waytype;
      int dirindex = directions_index(routing->directions, &node2->pos);
      node2->cost_src = node->cost_src + cost(routing->vehicle, routing->mode, waytype, length, speed) * dirindexlen / dirindex;
      node2->togo = destination_distance(&node2->pos, &routing->goal->pos, routing->directions);
      node2->cost = node2->cost_src + cost(routing->vehicle, routing->mode, (maptypes)-1, node2->togo, aerospeed(routing->vehicle))/* * dirindexlen / dirindex*/;
      node2->prev_node = node;
      if (node->next_node==NULL)
      {
        node->next_node = node2;
      }
      else
      {
        node3->next_list = node2;
        node2->prev_list = node3;
      }

      node3 = node2;
      routing->nodes++;
      routing->nodes_total++;
      printf_nodeinfo(node2);

      //append route from sector
      if (rcross)
      {
        if (route_sector_append(node2, &routing->endnodes) != 0)
        {
          //TODO should clear the node, it has no follow roads
          //currently add as endnode, it will be cleared in next step
          add_endnode(&routing->endnodes, node2);
        }
      }
      else
      {
        add_endnode(&routing->endnodes, node2);
      }
    }
  }
  if (cross)
  {
    map_nodeunlock(node->id, &node->pos);
  }

  //if there is no way, this is closed street, so clear the current path
  if (node->next_node==NULL)
  {
    remove_endnode(&routing->endnodes, node);
    clear_node(node);
  }

  //may be we can simply free very far end nodes to save memory
  while (routing->endnodes.count > config->limit_endnodes)
  {
    route_endnode *curendnode = routing->endnodes.last;
    curendnode->prev->next = NULL;
    routing->endnodes.last = curendnode->prev;
    clear_node(curendnode->node);
    free(curendnode);
    routing->endnodes.count--;
  }

  return 0;
}

// ********************* find route **************************
routingresult route_find(routing_info *routing, route_node **route, route_node *prevroute, route_progress *progress)
{
  route_node *node;
  int length;
  int startdist1, startdist2, goaldist1, goaldist2;
  int d;

  if ((routing->start == NULL) || (routing->goal == NULL))
    return RTS_ERROR;

  //init progress
  progress->progress = 0;
  progress->nodes = 0;
  progress->length = 0;

  struct timeval timer;
  timer_start(&timer);
  routed->terminate = 0;
  maptypes waytype_min = way_minallowed(routing->vehicle);
  maptypes waytype_max = way_maxallowed(routing->vehicle);

  //printf("vehicle: %d\n", routing->vehicle);
  //int truedist = wgs84_distance(routing->lat, source->lon, dest->lat, dest->lon);
  //printf("Route: true distance %dm\n", truedist);

  //find nearest start map way segment and snap to it
  long startnode1, startnode2;
  map_position startpos1, startpos2;

  if (map_nearestwayseg(&routing->start->pos, waytype_min, waytype_max, &startnode1, &startnode2, &routing->start->waytype, &routing->start->speed, &routing->start->flags) != 0)
    return RTS_ERROR;
  if (map_nodeposition(startnode1, &routing->start->pos, &startpos1) != 0)
    return RTS_ERROR;
  if (map_nodeposition(startnode2, &routing->start->pos, &startpos2) != 0)
    return RTS_ERROR;
  routing->start->nodetype = nt_start;
  routing->start->id = -1;
  routing->start->speed = 0;
  routing->start->direction = dir_none;
  routing->start->cross = 0;
  wgs84_waysnap(routing->start->pos.lat, routing->start->pos.lon,
                startpos1.lat, startpos1.lon, startpos2.lat, startpos2.lon, 
                &routing->start->pos.lat, &routing->start->pos.lon, &d, &startdist1, &startdist2);
  /*printf("Start %ld,%ld -> %ld %ld,%ld - %ld %ld,%ld\n", routing->start->pos.lat, routing->start->pos.lon,
         startnode1, startpos1.lat, startpos1.lon,
         startnode2, startpos2.lat, startpos2.lon
         );*/
  //routing->start->length = 0;

  //find nearest goal map way segment and snap to it
  long goalnode1, goalnode2;
  map_position goalpos1, goalpos2;
  if (map_nearestwayseg(&routing->goal->pos, waytype_min, waytype_max, &goalnode1, &goalnode2, &routing->goal->waytype, &routing->goal->speed, &routing->goal->flags) != 0)
    return RTS_ERROR;
  if (map_nodeposition(goalnode1, &routing->goal->pos, &goalpos1) != 0)
    return RTS_ERROR;
  if (map_nodeposition(goalnode2, &routing->goal->pos, &goalpos2) != 0)
    return RTS_ERROR;

  routing->goal->nodetype = nt_goal;
  routing->goal->id = -1;
  routing->goal->cross = 0;
  wgs84_waysnap(routing->goal->pos.lat, routing->goal->pos.lon,
                goalpos1.lat, goalpos1.lon, goalpos2.lat, goalpos2.lon, 
                &routing->goal->pos.lat, &routing->goal->pos.lon, &d, &goaldist1, &goaldist2);
  /*printf("Goal %ld,%ld -> %ld %ld,%ld - %ld %ld,%ld\n", routing->goal->pos.lat, routing->goal->pos.lon,
         goalnode1, goalpos1.lat, goalpos1.lon,
         goalnode2, goalpos2.lat, goalpos2.lon
         );*/

  //check requested ride angle
  int difangle;
  int sourceangle = wgs84_angle(startpos1.lat, startpos1.lon, startpos2.lat, startpos2.lon);
  if ((routing->angle>=-180) && (routing->angle<=180))
  {
    difangle = abs(sourceangle - routing->angle);
    while (difangle>360)
      difangle -= 360;
  }
  else
  {
    difangle = -1;
  }

  //if source and destination on the same way, then return the simple route
  //TODO apply angle
  if ((startnode1==goalnode1) && (startnode2==goalnode2))
  {
    if ((startdist1<=goaldist1) || (((routing->start->flags & WF_ONEWAY)==0) && (startdist1>goaldist1)))
    {
      routing->goal->length = abs(goaldist1 - startdist1); // ???
      routing->start->next_node = routing->goal;
      routing->goal->prev_node = routing->start;
      *route = routing->start;
      return RTS_DEST;
    }
  }

  //init sectors
  route_sector_init(routing->vehicle, routing->mode);

  //otherwise create two nodes - start and end node of source way,
  //but limit it by required angle and flags
  //length = wgs84_distance(snode->pos.lat, snode->pos.lon, source_seg.end.lat, source_seg.end.lon);
  if ((difangle==-1)||(difangle>=90) || (routing->start->flags & WF_ONEWAY))
  {
    node = new_node();
    node->id = startnode2;
    node->nodetype = nt_node;
    node->pos = startpos2;
    node->length = startdist2;
    node->waytype = routing->start->waytype;
    node->speed = routing->start->speed;
    node->flags = routing->start->flags;
    node->direction = dir_none;
    node->cross = 0;
    node->cost_src = cost(routing->vehicle, routing->mode, node->waytype, node->length, node->speed);
    length = wgs84_distance(node->pos.lat, node->pos.lon, routing->goal->pos.lat, routing->goal->pos.lon);
    node->cost = node->cost_src + cost(routing->vehicle, routing->mode, (maptypes)-1, length, aerospeed(routing->vehicle));
    add_endnode(&routing->endnodes, node);
    routing->start->next_node = node;
    node->prev_node = routing->start;
    if (difangle!=-1)  //disallow routing back in next step
      routing->start->id = startnode1;
    routing->nodes++;
    routing->nodes_total++;
    printf_nodeinfo(node);
  }

  if ((difangle<=90) && ((routing->start->flags & WF_ONEWAY)==0))
  {
    node = new_node();
    node->id = startnode1;
    node->nodetype = nt_node;
    node->pos = startpos1;
    node->length = startdist1;
    node->waytype = routing->start->waytype;
    node->speed = routing->start->speed;
    node->flags = routing->start->flags | WF_DIR;
    node->direction = dir_none;
    node->cross = 0;
    node->cost_src = cost(routing->vehicle, routing->mode, node->waytype, node->length, node->speed);
    length = wgs84_distance(node->pos.lat, node->pos.lon, routing->goal->pos.lat, routing->goal->pos.lon);
    node->cost = node->cost_src + cost(routing->vehicle, routing->mode, (maptypes)-1, length, aerospeed(routing->vehicle));
    add_endnode(&routing->endnodes, node);
    node->prev_node = routing->start;
    if (difangle!=-1)  //disallow routing back in next step
      routing->start->id = startnode2;
    if (routing->start->next_node)
    {
      routing->start->next_node->next_list = node;
      node->prev_list = routing->start->next_node;
    }
    else
    {
      routing->start->next_node = node;
    }
    routing->nodes++;
    routing->nodes_total++;
    printf_nodeinfo(node);
  }

  //create two destination nodes, this is used to detect way segment to reach
  node = new_node();
  node->id = goalnode1;
  node->nodetype = nt_node;
  node->pos = goalpos1;
  node->length = goaldist1;
  node->waytype = routing->goal->waytype;
  node->speed = routing->goal->speed;
  node->flags = routing->goal->flags;
  node->direction = dir_none;
  node->next_node = routing->goal;
  routing->dest1 = node;

  //if ((routing->goal->flags & WF_ONEWAY)==0)
  //{
    node = new_node();
    node->id = goalnode2;
    node->nodetype = nt_node;
    node->pos = goalpos2;
    node->length = goaldist2;
    node->waytype = routing->goal->waytype;
    node->speed = routing->goal->speed;
    node->flags = routing->goal->flags | WF_DIR;
    node->direction = dir_none;
    node->next_node = routing->goal;
    routing->dest2 = node;
  //}

  //lets go
  int maxprg = 0;
  int same = 0;
  map_position sect;
  while ((routing->status==RTS_ROUTING) && (routing->endnodes.first) && (routing->endnodes.first->node) && (routed->terminate==0))
  {
    //route step
    if (same == 0)
    {
      //continue with best node
      sect.lat = MAPSECTOR(routing->endnodes.first->node->pos.lat);
      sect.lon = MAPSECTOR(routing->endnodes.first->node->pos.lon);
      route_step(routing, routing->endnodes.first->node);
      same = 1;
    }
    else
    {
      //find node with the same sector
      same = 0;
      route_endnode *endnode = routing->endnodes.first;
      while (endnode)
      {
        if ((sect.lat == MAPSECTOR(endnode->node->pos.lat)) && (sect.lon == MAPSECTOR(endnode->node->pos.lon)))
        {
          same = 1;
          route_step(routing, endnode->node);
          break;
        }
        endnode = endnode->next;
      }
    }

    //check if we pass the previous route
    if (prevroute && (routing->endnodes.first && (routing->endnodes.first->node->length > minpassdist)))
    {
      route_node *passnode = reached_route(routing->endnodes.first->node->id, prevroute);
      if (passnode)
      {
        node = routing->endnodes.first->node;
        printf("We have reached the route! (node %ld)\n", node->id);
        remove_endnode(&routing->endnodes, node);
        //node->cross = map_nodecrosscnt(node->id);
        node->direction = dir_none;
        routing->status = RTS_ROUTE;
        free_node(routing->dest1);
        routing->dest1 = node;
        free_node(routing->dest2);
        routing->dest2 = node;
        break;
      }
    }

    //create the statistics
    route_lock();
    if (routing->endnodes.first && routing->endnodes.first->node->togo)
    {
      int prg = routing->endnodes.first->node->length * 1000 / (routing->endnodes.first->node->length + routing->endnodes.first->node->togo);
      if (prg>maxprg)
      {
        progress->progress = prg / 10;
        progress->length = routing->endnodes.first->node->length;
        maxprg = prg;
      }
      progress->nodes = routing->nodes;
    }
    route_unlock();

    //printf("Nodes %8d/%8d visited %8d ends %5d len %8ld cost %8ld/%8ld %d%%\r", routing->nodes, routing->nodes_total, routing->visitedcnt, routing->endnodes, routing->end->node->length, routing->end->node->cost_src, routing->end->node->cost, *progress);
    //fflush(stdout);
  }
 //printf("Done, free memory\n");

  //terminated? 
  if (routing->status==RTS_ROUTING)
  {
    routing->status = RTS_ERROR;
  }

  //free all unused nodes
  clear_visited(&routing->visited);
  while (routing->endnodes.first)
  {
    node = routing->endnodes.first->node;
    printf_nodeinfo(node);
    remove_endnode(&routing->endnodes, routing->endnodes.first->node);
    clear_node(node);
  }

  if (routing->status==RTS_DEST)
  {
    routing->goal->nodetype = nt_destination;
    route_maketurn(routing->vehicle, routing->start, routing->goal);
  }

  //print the result and return
  if ((routing->status==RTS_DEST) || (routing->status==RTS_ROUTE))
  {
    node = routing->start;
    printf("Found the route: %ldm %d nodes, %d analyzed\n", routing->goal->length, routing->nodes, routing->nodes_total);
    //printf("Max distance %ld on %ld\n", routing->maxdist, routing->maxdistid);

    /*while (node)
    {
      //printf("Node %10ld OSM %10ld Way %10ld Length %d Cost %ld Turn %d %ld\n", node->id, node->osmid, node->wayid, node->length, node->cost_src, node->direction, node->toturn);
      if (node->direction)
        printf("Node %ld Dist %ld Turn %d %ld dir %d\n", node->id, node->togo, node->direction, node->toturn, node->direction);
      if (node->next_list)
      {
        printf("NEXT LIST %ld !!!!!!!!!!!\n", node->next_list->id);
      }
      if (node->prev_list)
        printf("PREV LIST !!!!!!!!!!!\n");
      node = node->next_node;
    }*/

    *route = routing->start;
  }
  else
  {
    printf("The route was not found, %d nodes analyzed\n", routing->nodes_total);
    //route_freeroute(routing->start);
    free_node(routing->goal);
    free_node(routing->dest1);
    free_node(routing->dest2);
    *route = NULL;
    //*route = routing->source;
  }
  timer_print(&timer);
  return routing->status;
}

routingresult route_find_pos2pos(map_position *source, map_position *dest, route_directions *directions, int angle,
                              routing_vehicle vehicle, routing_mode route_mode, 
                              route_node **route, route_node *prevroute, route_progress *progress)
{
  route_node *node;
  routing_info routing;
  //long osmid1, osmid2;

  if ((source==NULL) || (dest==NULL))
    return RTS_ERROR;

  //printf("Find %ld,%ld -> %ld,%ld\n", source->lat, source->lon, dest->lat, dest->lon);

  //first init routing
  memset(&routing, 0, sizeof(routing_info));
  routing.mode = route_mode;
  routing.vehicle = vehicle;
  routing.angle = angle;
  routing.directions = append_directions(source, dest, directions);

  //start node
  node = new_node();
  node->id = -1;
  node->pos = *source;
  node->nodetype = nt_start;
  routing.start = node;

  //goal node
  node = new_node();
  node->id = -1;
  node->pos = *dest;
  node->nodetype = nt_goal;
  routing.goal = node;

  return route_find(&routing, route, prevroute, progress);
}

routingresult route_find_node2pos(route_node *source, map_position *dest, route_directions *directions, int angle,
                              routing_vehicle vehicle, routing_mode route_mode, 
                              route_node **route, route_node *prevroute, route_progress *progress)
{
  route_node *node;
  routing_info routing;

  if ((source==NULL) || (dest==NULL))
    return RTS_ERROR;

  printf("Find %ld,%ld -> %ld,%ld\n", source->pos.lat, source->pos.lon, dest->lat, dest->lon);

  //first init routing
  memset(&routing, 0, sizeof(routing_info));
  routing.mode = route_mode;
  routing.vehicle = vehicle;
  routing.angle = angle;
  routing.directions = append_directions(&source->pos, dest, directions);

  //start node - copy
  routing.start = new_node();
  *routing.start = *source;
  routing.start->next_node = NULL;
  routing.start->prev_node = NULL;
  routing.start->next_list = NULL;
  routing.start->prev_list = NULL;

  //goal node
  node = new_node();
  node->id = -1;
  node->pos = *dest;
  node->nodetype = nt_goal;
  routing.goal = node;

  return route_find(&routing, route, prevroute, progress);
}

routingresult route_find_pos2node( map_position *source, route_node *dest, route_directions *directions, int angle,
                              routing_vehicle vehicle, routing_mode route_mode, 
                              route_node **route, route_node *prevroute, route_progress *progress)
{
  route_node *node;
  routing_info routing;

  if ((source==NULL) || (dest==NULL))
    return RTS_ERROR;

  printf("Find %ld,%ld -> %ld,%ld\n", source->lat, source->lon, dest->pos.lat, dest->pos.lon);

  //first init routing
  memset(&routing, 0, sizeof(routing_info));
  routing.mode = route_mode;
  routing.vehicle = vehicle;
  routing.angle = angle;
  routing.directions = append_directions(source, &dest->pos, directions);

  //start node
  node = new_node();
  node->id = -1;
  node->pos = *source;
  node->nodetype = nt_start;
  routing.start = node;

  //goal node - copy
  routing.goal = new_node();
  *routing.goal = *dest;
  routing.goal->next_node = NULL;
  routing.goal->prev_node = NULL;
  routing.goal->next_list = NULL;
  routing.goal->prev_list = NULL;

  return route_find(&routing, route, prevroute, progress);
}

routingresult route_find_node2node( route_node *source, route_node *dest, route_directions *directions, int angle,
                              routing_vehicle vehicle, routing_mode route_mode, 
                              route_node **route, route_node *prevroute, route_progress *progress)
{
  routing_info routing;

  if ((source==NULL) || (dest==NULL))
    return RTS_ERROR;

  printf("Find %ld,%ld -> %ld,%ld\n", source->pos.lat, source->pos.lon, dest->pos.lat, dest->pos.lon);

  //first init routing
  memset(&routing, 0, sizeof(routing_info));
  routing.mode = route_mode;
  routing.vehicle = vehicle;
  routing.angle = angle;
  routing.directions = append_directions(&source->pos, &dest->pos, directions);

  //start node - copy
  routing.start = new_node();
  *routing.start = *source;
  routing.start->next_node = NULL;
  routing.start->prev_node = NULL;
  routing.start->next_list = NULL;
  routing.start->prev_list = NULL;

  //goal node - copy
  routing.goal = new_node();
  *routing.goal = *dest;
  routing.goal->next_node = NULL;
  routing.goal->prev_node = NULL;
  routing.goal->next_list = NULL;
  routing.goal->prev_list = NULL;

  return route_find(&routing, route, prevroute, progress);
}

route_node *route_combine(routing_vehicle vehicle, route_node *route, route_node *pass, route_node *node)
{
  if ((route==NULL) || (pass==NULL) || (node==NULL))
    return NULL;

  //find the last node of route
  route_node *last = route_lastnode(route);

  //find the passing node
  while (pass)
  {
    //if ((pass->id == node->id) && (wgs84_compare(pass->pos.lat, pass->pos.lon, node->pos.lat, node->pos.lon)==0))
    if ((pass == node) || ((pass->id == last->id) && (wgs84_compare(pass->pos.lat, pass->pos.lon, last->pos.lat, last->pos.lon)==0)))
    {
      pass->prev_node = last->prev_node;
      last->prev_node->next_node = pass;
      route_maketurn(vehicle, route, pass);
      free(last);
      break;
    }

    route_node *node2 = pass->next_node;
    free(pass);
    pass = node2;
  }

  return route;
}

route_node *route_append(route_node *route, route_node *append)
{
  if ((route==NULL) || (append==NULL))
    return NULL;

  //find the last node of route
  route_node *node = route_lastnode(route);

  //check that lastnode and first append node are the same
  //printf("%ld-%ld-%ld\n", node->pos.lat, node->pos.lon, node->wayid);
  //printf("%ld-%ld-%ld\n", append->pos.lat, append->pos.lon, append->wayid);
  /*if ((node->pos.lat != append->pos.lat) || (node->pos.lon != append->pos.lon))
  {
    return NULL;
  }*/

  //copy parameters and join
  node->length = append->length;
  node->toturn = append->toturn;
  node->togo = append->togo;
  node->totake = append->totake;
  node->cost_src = append->cost_src;
  node->cost = append->cost;
  node->next_node = append->next_node;
  node->next_node->prev_node = node;
  free(append);

  return route;
}

int route_lock()
{
  if (routed == NULL)
    return -1;
  return mcmutex_lock(&routed->mutex);
}

int route_unlock()
{
  if (routed == NULL)
    return -1;
  return mcmutex_unlock(&routed->mutex);
}

route_route *route_init()
{
  route_route *route;

  routed = (routing_data*)malloc(sizeof(routing_data));
  mcmutex_init(&routed->mutex);

  route = (route_route*)malloc(sizeof(route_route));
  memset(route, 0, sizeof(route_route));
  return route;
}

int route_deinit(route_route *route)
{
  if (routed == NULL)
    return -1;
  if (route)
  {
    if (route->route)
    {
      route_freeroute(route->route);
      route->route = NULL;
    }
    free(route);
  }
  route_sector_deinit();
  mcmutex_destroy(&routed->mutex);
  free(routed);
  routed = NULL;
  return 0;
}

int route_stop()
{
  routed->terminate = 1;
  return 0;
}

//itinerary
route_itinerary *route_itinerary_new()
{
  route_itinerary *itinerary = (route_itinerary *)malloc(sizeof(route_itinerary));
  memset(itinerary, 0, sizeof(route_itinerary));
  return(itinerary);
}

void route_itinerary_free(route_itinerary *itinerary)
{
  while (itinerary)
  {
    route_itinerary *itinerary2 = itinerary->next;
    if (itinerary->name)
      free(itinerary->name);
    free(itinerary);
    itinerary = itinerary2;
  }
}

route_itinerary *route_lastitinerary(route_itinerary *itinerary)
{
  if (itinerary == NULL)
    return NULL;

  while (itinerary->next)
  {
    itinerary = itinerary->next;
  }
  return itinerary;
}

route_directions *route_createdirections(route_itinerary *ri1, route_itinerary *ri2)
{
  int marks = 0;
  route_directions *directions = NULL;
  route_itinerary *rit = ri1;
  while (rit!= ri2)
  {
    if (rit->nodetype == nt_direction)
      marks++;
    rit = rit->next;
  }
  if (marks>0)
  {
    directions = (route_directions*)malloc(sizeof(route_directions)+sizeof(map_position)*marks);
    directions->count = marks;
    rit = ri1;
    while (rit!= ri2)
    {
      marks = 0;
      if (rit->nodetype == nt_direction)
      {
        directions->pos[marks] = rit->pos;
        marks++;
      }
      rit = rit->next;
    }
  }
  return directions;
}
