/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * route_common.cc
 * Copyright (C) Mike Crash 2011 <mike@mikecrash.com>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdlib.h>
#include <string.h>
#include "route_common.h"

//common functions
int vehiclespeed(routing_vehicle vehicle, maptypes waytype, int *speed)
{
  switch (vehicle)
  {
    case RV_TRAIN:
      *speed = 80;
      break;
    case RV_TRUCK:
      //allow lower roads, but do not prefer it, set speed to minimum
      if (waytype==WT_SERVICE)
        *speed = 30;
      else if (waytype>=WT_UNCLASSIFIED)
        *speed = 1;
      else if (waytype>=WT_TERTIARY)
        *speed = 20;
      else if (*speed>90)
        *speed = 90;
      break;
    case RV_CAR:
      if (*speed==0)
      {
        //printf("Warning: car way speed is 0 (type %d)\n", waytype);
        switch (waytype)
        {
          case WT_MOTORWAY: 
            *speed = 130;
            break;
          case WT_MOTORWAYLINK:
            *speed = 80;
            break;
          case WT_TRUNK: 
            *speed = 120;
            break;
          case WT_TRUNKLINK: 
            *speed = 70;
            break;
          case WT_PRIMARY:
            *speed = 90;
            break;
          case WT_SECONDARY:
            *speed = 75;
            break;
          case WT_TERTIARY:
            *speed = 60;
            break;
          case WT_UNCLASSIFIED:
            *speed = 50;
            break;
          case WT_RESIDENTIAL:
            *speed = 25;
            break;
          case WT_LIVING: 
            *speed = 20;
            break;
          case WT_SERVICE:
            *speed = 20;
            break;
          case WT_TRACK1: 
            *speed = 20;
            break;
          case WT_TRACK2: 
            *speed = 15;
            break;
          case WT_TRACK3:
            *speed = 10;
            break;
          case WT_TRACK4:
            *speed = 6;
            break;
          case WT_TRACK5:
            *speed = 3;
            break;
          default:
            *speed = 1;
            break;
        }
      }
      break;
    case RV_CYCLE:
      if (waytype<=WT_PRIMARY)
        *speed = 25;
      else if (waytype<=WT_SECONDARY)
        *speed = 30;
      else if (waytype<=WT_TERTIARY)
        *speed = 28;
      else if (waytype<=WT_SERVICE)
        *speed = 22;
      else
        *speed = 15;
      break;
    case RV_MTB:
      //prefer tertiary and tracks before primary and secondary roads
      if (waytype<=WT_PRIMARY)
        *speed = 1;
      else if (waytype<=WT_SECONDARY)
        *speed = 18;
      else if (waytype<=WT_TERTIARY)
        *speed = 20;
      else if (waytype<=WT_TRACK1)
        *speed = 22;
      else if (waytype<=WT_TRACK2)
        *speed = 20;
      else if (waytype<=WT_TRACK3)
        *speed = 17;
      else if (waytype<=WT_TRACK4)
        *speed = 13;
      else if (waytype<=WT_TRACK5)
        *speed = 9;
      else
        *speed = 5;
      break;
    case RV_FOOT:
      *speed = 5;
      break;
  }
  if (*speed==0)
  {
    *speed = 1;
    printf("Warning: way speed is 0 (type %d)\n", waytype);
  }
  return 0;
}

maptypes way_minallowed(routing_vehicle vehicle)
{
  switch (vehicle)
  {
    case RV_TRAIN:
      return NT_RAIL;
    case RV_TRUCK:
    case RV_CAR:
      return WT_MOTORWAY;
    case RV_CYCLE:
      return WT_PRIMARY;
    case RV_MTB:
      return WT_PRIMARY;
    case RV_FOOT:
      return WT_PRIMARY;
  }
  return (maptypes)0;
}

maptypes way_maxallowed(routing_vehicle vehicle)
{
  switch (vehicle)
  {
    case RV_TRAIN:
      return NT_RAIL;
    case RV_TRUCK:
      return WT_SERVICE;
    case RV_CAR:
      return WT_TRACK5;
    case RV_CYCLE:
      return WT_TRACK1;
    case RV_MTB:
      return WT_PATH;
    case RV_FOOT:
      return WT_STEPS;
  }
  return (maptypes)0;
}

int way_allowed(routing_vehicle vehicle, maptypes waytype)
{
  if ((vehicle<=RV_CAR) && (waytype == WT_CYCLE))
    return 0;

  if ((waytype>=way_minallowed(vehicle)) && (waytype<=way_maxallowed(vehicle)))
    return 1;
  return 0;
}

//should be maximum vehicle speed, but for faster routing decrease to something above avarage vehicle speed
int aerospeed(routing_vehicle vehicle)
{
  switch (vehicle)
  {
    case RV_TRAIN:
      return 130;
    case RV_TRUCK:
      return 85;
    case RV_CAR:
      return 85;
    case RV_CYCLE:
      return 25;
    case RV_MTB:
      return 18;
    case RV_FOOT:
      return 5;
  }
  return 90;
}

long cost(routing_vehicle vehicle, routing_mode mode, maptypes waytype, int length, int speed)
{
  if (mode == routemode_length)
    return length;

  //check unusual parameters)
  if (waytype!=-1)
  {
    vehiclespeed(vehicle, waytype, &speed);
  }

  if (speed<=0)
  {
    printf("ERROR: invalid speed (%d)\n", speed);
    //speed = 1;
    return 0;
  }

  long cost = length*100/speed;
  //printf("Type %d len %d speed %d cost %ld\n", waytype, length, speed, cost);
  return cost;
}

// ********************* nodes and routes **************************
route_node *new_node(void)
{
  route_node *node = (route_node *)malloc(sizeof(route_node));
  memset(node, 0, sizeof(route_node));
  return node;
}

void free_node(route_node *node)
{
  if (node)
    free(node);
}

void clear_node(route_node *node)
{
  route_node *prev_node;

  if ((node==NULL) || (node->id < 0))
  {
    return;
  }

  //printf("Clear node %ld OSM %10ld way %ld\n", node->id, node->osmid, node->wayid);
  prev_node = node->prev_node;

  //delete it from the list
  if (node->next_list != NULL)
  {
    node->next_list->prev_list = node->prev_list;
  }
  if (node->prev_list != NULL)
  {
    node->prev_list->next_list = node->next_list;
  }

  //update prev_node, if we are the first
  if (prev_node && (prev_node->next_node == node))
  {
    prev_node->next_node = node->next_list;
  }

  //if there is no node in the list, clear the prev_node too
  if (prev_node && (node->next_list == NULL) && (node->prev_list == NULL))
  {
    //printf("Clear prev %ld\n", prev_node->id);
    clear_node(prev_node);
  }

  //free
  //printf("Cleared %ld\n", node->id);
  free_node(node);
}

int route_freeroute(route_node *route)
{
  route_node *node = route;
  while (node)
  {
    route_node *node2 = node;
    node2 = node->next_node;
    free(node);
    node = node2;
  }
  return 0;
}

// ********************* end nodes **************************
void add_endnode(route_endnodes *endnodes, route_node *node)
{
  route_endnode *endnode, *curendnode, *prevendnode;
  endnode = (route_endnode *)malloc(sizeof(route_endnode));
  endnode->node = node;

  //printf("  Endnode cost %ld\n", node->cost);

  //if empty list then return
  if (endnodes->first == NULL)
  {
    endnode->next = NULL;
    endnode->prev = NULL;
    endnodes->first = endnode;
    endnodes->last = endnode;
    endnodes->count++;
    return;
  }

  //if be the first, then return
  if (node->cost < endnodes->first->node->cost)
  {
    endnode->next = endnodes->first;
    endnode->prev = NULL;
    endnodes->first->prev = endnode;
    endnodes->first = endnode;
    endnodes->count++;
    return;
  }

  //iterate and insert
  curendnode = endnodes->first->next;
  prevendnode = endnodes->first;
  while ((curendnode != NULL) && (node->cost >= curendnode->node->cost))
  {
    //printf("    %ld\n", curendnode->next->node->cost);
    prevendnode = curendnode;
    curendnode = curendnode->next;
  }
  endnode->next = curendnode;
  endnode->prev = prevendnode;
  if (curendnode)
    curendnode->prev = endnode;
  prevendnode->next = endnode;
  if (curendnode==NULL)
    endnodes->last = endnode;
  endnodes->count++;
}

void remove_endnode(route_endnodes *endnodes, route_node *node)
{
  route_endnode *curendnode = endnodes->first;
  route_endnode *prevendnode = NULL;
  if ((node==NULL) || (curendnode == NULL))
  {
    return;
  }
  while ((curendnode!= NULL) && (curendnode->node != node))
  {
    prevendnode = curendnode;
    curendnode = curendnode->next;
  }
  if (curendnode!= NULL)
  {
    if (curendnode->next)
      curendnode->next->prev = prevendnode;
    else
      endnodes->last = prevendnode;
    if (prevendnode == NULL)
      endnodes->first = curendnode->next;
    else
      prevendnode->next = curendnode->next;
    free(curendnode);
    endnodes->count--;
  }
}

// ********************* visited **************************
void add_visited(route_visited *visited, long fromid, long toid)
{
  if (visited==NULL)
    return;

  route_visited_block *block = visited->blocks;

  if (block==NULL)
  {
    block = (route_visited_block *)malloc(sizeof(route_visited_block));
    block->count = 0;
    block->next = NULL;
    visited->blocks = block;
  }
  while (block->next)
  {
    block = block->next;
  }
  if (block->count==visited_max)
  {
    route_visited_block *visited2 = (route_visited_block *)malloc(sizeof(route_visited_block));
    visited2->count = 0;
    visited2->next = NULL;
    block->next = visited2;
    block = visited2;
  }
  block->turn[block->count].fromid = fromid;
  block->turn[block->count].toid = toid;
  block->count++;
  visited->count++;
}

int is_visited(route_visited *visited, long fromid, long toid)
{
  route_visited_block *block = visited->blocks;
  while (block!=NULL)
  {
    int i = 0;
    route_turn *turn = block->turn;
    while (i<block->count)
    {
      if ((turn->fromid == fromid) && (turn->toid == toid))
      {
        return 1;
      }
      i++;
      turn++;
    }
    block = block->next;
  }
  return 0;
}

int clear_visited(route_visited *visited)
{
  route_visited_block *block = visited->blocks;
  while (block!=NULL)
  {
    route_visited_block *visited2 = block;
    block = block->next;
    free(visited2);
  }
  visited->count = 0;
  visited->blocks = NULL;
  return 0;
}

void printf_nodeinfo(route_node *node)
{
  //printf("Node %5ld Length %ld Cost %ld/%ld\n", node->id, node->length, node->cost_src, node->cost);
}
