/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * route.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 *
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <expat.h>
#include "safefile.h"
#include "route_file.h"
#include "map_read.h"
#include "projection.h"
#include "gpx.h"

#define BUFF_SIZE 1000

//common
typedef struct
{
  maptypes maptype;
  const char *label;
} routefiletype;

routefiletype routefiletypearr [] =
{
  { WT_MOTORWAY, "motorway"},
  { WT_MOTORWAYLINK, "motorwaylink"},
  { WT_TRUNK, "trunk"},
  { WT_TRUNKLINK, "trunklink"},
  { WT_PRIMARY, "primary"},
  { WT_PRIMARYLINK, "primarylink"},
  { WT_SECONDARY, "secondary"},
  { WT_TERTIARY, "tertiary"},
  { WT_UNCLASSIFIED, "unclassified"},
  { WT_RESIDENTIAL, "residential"},
  { WT_LIVING, "living"},
  { WT_SERVICE, "service"},
  { WT_CYCLE, "cycle"},
  { WT_TRACK1, "track1"},
  { WT_TRACK2, "track2"},
  { WT_TRACK3, "track3"},
  { WT_TRACK4, "track4"},
  { WT_TRACK5, "track5"},
  { WT_BRIDLE, "bridle"},
  { WT_PATH, "path"},
  { WT_FOOT, "foot"},
  { WT_STEPS, "steps"}
};

const char *maptypelabel(maptypes maptype)
{
  int cnt = sizeof(routefiletypearr) / sizeof(routefiletype);
  for (int i=0; i<cnt; i++)
  {
    if (maptype == routefiletypearr[i].maptype)
      return routefiletypearr[i].label;
  }

  return NULL;
}

typedef struct
{
  routing_vehicle vehicle;
  const char *label;
} routevehicletype;

routevehicletype routevehiclearr[] =
{
  { RV_TRAIN, "train" },
  { RV_TRUCK, "truck" },
  { RV_CAR, "car" },
  { RV_CYCLE, "cycle" },
  { RV_MTB, "mtb" },
  { RV_FOOT, "foot" }
};

const char *vehiclelabel(routing_vehicle vehicle)
{
  int cnt = sizeof(routevehiclearr) / sizeof(routevehicletype);
  for (int i=0; i<cnt; i++)
  {
    if (vehicle == routevehiclearr[i].vehicle)
      return routevehiclearr[i].label;
  }

  return NULL;
}

routing_vehicle vehicletype(char *label)
{
  int cnt = sizeof(routevehiclearr) / sizeof(routevehicletype);
  for (int i=0; i<cnt; i++)
  {
    if (strcmp(label, routevehiclearr[i].label) == 0)
      return routevehiclearr[i].vehicle;
  }

  return (routing_vehicle)-1;
}

typedef struct
{
  route_nodetype nodetype;
  const char *label;
} routenodetype;

routenodetype routenodetypearr[] =
{
  { nt_node, "node" },
  { nt_start, "start" },
  { nt_direction, "direction" },
  { nt_destination, "destination" },
  { nt_waypoint, "waypoint" },
  { nt_goal, "goal" }
};

const char *itinodelabel(route_nodetype nodetype)
{
  int cnt = sizeof(routenodetypearr) / sizeof(routenodetype);
  for (int i=0; i<cnt; i++)
  {
    if (nodetype == routenodetypearr[i].nodetype)
      return routenodetypearr[i].label;
  }

  return NULL;
}

route_nodetype itinodetype(char *label)
{
  int cnt = sizeof(routenodetypearr) / sizeof(routenodetype);
  for (int i=0; i<cnt; i++)
  {
    if (strcmp(label, routenodetypearr[i].label) == 0)
      return routenodetypearr[i].nodetype;
  }

  return nt_node;
}
//load route

typedef struct
{
  int depth;
  int gpx;
  int gpx_trk;
  int gpx_trkseg;
  int gpx_trkpt;
  int vehicle;
  int nodetype;
  int waytype;
  int direction;
  int name;
  char *nodename;
  route_route *route;
  route_node *node;
  route_itinerary *iti;
  int error;
} route_load_data;


static void XMLCALL route_load_startElement(void *userData, const char *name, const char **attr)
{
  route_load_data *data = (route_load_data *)userData;
  if (data->error)
    return;

  if ((data->depth == 0) && (strcmp(name, "gpx") == 0))
  {
    data->gpx = 1;
  }
  else if ((data->depth == 1) && (strcmp(name, "trk") == 0) && (data->gpx == 1))
  {
    data->gpx_trk = 1;
  }
  else if ((data->depth == 2) && (data->gpx_trk == 1))
  {
    if (strcmp(name, "trkseg") == 0)
    {
      if ((int)data->route->vehicle == -1)
        data->error = 1;
      data->gpx_trkseg = 1;
    }
    else if (strcmp(name, "vehicle") == 0)
    {
      data->vehicle = 1;
    }
  }
  else if ((data->depth == 3) && (strcmp(name, "trkpt") == 0) && (data->gpx_trkseg == 1))
  {
    double lat, lon;
    long id;
    if ((attr_long(attr, "id", &id) == 0) && (attr_double(attr, "lat", &lat) == 0) && (attr_double(attr, "lon", &lon) == 0))
    {
      data->gpx_trkpt = 1;
      route_node *node = new_node();
      node->prev_node = data->node;
      if (data->node)
        data->node->next_node = node;
      if (data->route->route == NULL)
        data->route->route = node;
      data->node = node;
      node->id = id;
      node->pos.lat = lat * 1000000;
      node->pos.lon = lon * 1000000;
      //printf("%d: %f,%f\n", gpxtrack->count, lat, lon);
    }
  }
  else if ((data->depth == 4) && (data->gpx_trkpt == 1))
  {
    if (data->node)
    {
      if (strcmp(name, "nodetype") == 0)
      {
        data->nodetype = 1;
      }
      else if (strcmp(name, "waytype") == 0)
      {
        data->waytype = 1;
      }
      else if (strcmp(name, "direction") == 0)
      {
        data->direction = 1;
      }
      else if (strcmp(name, "name") == 0)
      {
        data->name = 1;
      }
    }
  }
  data->depth++;
}

void route_load_charElement(void *userData, const char *txt, int txtlen)
{
  route_load_data *data = (route_load_data *)userData;
  if (data->error)
    return;

  char *s = (char *)malloc(txtlen+1);
  memcpy(s, txt, txtlen);
  s[txtlen] = 0;

  if (data->depth == 3)
  {
    if (data->gpx_trk && data->vehicle)
    {
      data->route->vehicle = vehicletype(s);
      if ((int)data->route->vehicle == -1)
        data->error = 1;
    }
  }
  else if (data->depth == 5)
  {
    if (data->nodetype)
    {
      data->node->nodetype = itinodetype(s);
    }
    else if (data->name)
    {
      if (data->nodename)
        free(data->nodename);
      data->nodename = (char *)malloc(strlen(s)+1);
      strcpy(data->nodename, s);
    }
    /*else if (data->waytype)
    {
      data->node->waytype = (maptypes)atoi(s);
    }
    else if (data->direction)
    {
      data->node->direction = (route_dir)atoi(s);
    }*/
  }
  free(s);
  //fwrite(txt, txtlen, sizeof(char), stdout);
}

static void XMLCALL route_load_endElement(void *userData, const char *name)
{
  route_load_data *data = (route_load_data *)userData;
  data->depth--;
  if (data->error)
    return;

  if ((data->depth == 0) && (strcmp(name, "gpx") == 0))
  {
    data->gpx = 0;
  }
  else if ((data->depth == 1) && (strcmp(name, "trk") == 0))
  {
    data->gpx_trk = 0;
  }
  else if (data->depth == 2)
  {
    if ((strcmp(name, "trkseg") == 0))
    {
      data->gpx_trkseg = 0;
    }
    else if ((strcmp(name, "vehicle") == 0))
    {
      data->vehicle = 0;
    }
  }
  else if ((data->depth == 3) && (strcmp(name, "trkpt") == 0))
  {
    data->gpx_trkpt = 0;
    //check previous node and good connection
    if ((data->node) && (data->node->prev_node))
    {
      route_node *prevnode = data->node->prev_node;
      if ((prevnode->id != -1) && (data->node->id != -1))
      {
        mcross *cross;
        int count;
        if (map_nodecross(prevnode->id, &prevnode->pos, &count, &cross) == 0)
        {
          prevnode->cross = count;
          int found = 0;
          for (int i=0; i<count; i++)
          {
            if (cross[i].id == data->node->id)
            {
              data->node->speed = cross[i].speed;
              vehiclespeed(data->route->vehicle, cross[i].waytype, &data->node->speed);
              data->node->waytype = cross[i].waytype;
              data->node->flags = cross[i].flags;
              found = 1;
              break;
            }
          }
          map_nodeunlock(data->node->prev_node->id, &data->node->prev_node->pos);
          if (found == 0)
          {
            data->error = 1;
            printf("Node %ld has no way to node %ld!\n", prevnode->id, data->node->id);
            return;
          }
        }
        else
        {
          //node not found
          data->error = 1;
          printf("Node %ld not found in map!\n", prevnode->id);
          return;
        }
      }
      else
      {
        long node1, node2;
        maptypes waytype;
        int speed;
        unsigned int flags;
        maptypes min = way_minallowed(data->route->vehicle);
        maptypes max = way_maxallowed(data->route->vehicle);
        if (map_nearestwayseg(&prevnode->pos, min, max, &node1, &node2, &waytype, &speed, &flags) == 0)
        {
          data->node->waytype = waytype;
        }
        else
        {
          data->node->waytype = min;
        }
        vehiclespeed(data->route->vehicle, data->node->waytype, &speed);
        data->node->speed = speed;
      }
      //compute length and time
      int len = wgs84_distance(prevnode->pos.lat, prevnode->pos.lon, data->node->pos.lat, data->node->pos.lon);
      data->node->length = prevnode->length + len;
    }

    //add itinerary
    if ((data->node->nodetype == nt_start) ||
        (data->node->nodetype == nt_direction) ||
        (data->node->nodetype == nt_destination) ||
        (data->node->nodetype == nt_waypoint) ||
        (data->node->prev_node == NULL))
    {
      if (data->node->prev_node == NULL)
        data->node->nodetype = nt_start;
      route_itinerary *iti = route_itinerary_new();
      iti->node = data->node;
      iti->pos = data->node->pos;
      iti->nodetype = data->node->nodetype;
      iti->prev = data->iti;
      if (data->iti)
      {
        data->iti->next = iti;
      }
      if (data->route->itinerary == NULL)
      {
        data->route->itinerary = iti;
      }
      if (data->nodename)
      {
        iti->name = data->nodename;
        data->nodename = NULL;
      }
      data->iti = iti;
    }
  }
  else if (data->depth == 4)
  {
    if ((strcmp(name, "nodetype") == 0))
    {
      data->nodetype = 0;
    }
    else if ((strcmp(name, "waytype") == 0))
    {
      data->waytype = 0;
    }
    else if ((strcmp(name, "direction") == 0))
    {
      data->direction = 0;
    }
    else if ((strcmp(name, "name") == 0))
    {
      data->name = 0;
    }
  }
}

route_route *route_load(char *filename)
{
  FILE *f;
  if ((f = fopen(filename, "r")) == NULL)
  {
    printf("Error opening route file!\n");
    return NULL;
  }

  //create route
  route_route *route = (route_route *)malloc(sizeof(route_route));
  memset(route, 0, sizeof(route_route));

  //open XML parser
  route_load_data data;
  memset(&data, 0, sizeof(data));
  data.route = route;
  data.route->vehicle = (routing_vehicle)-1;
  printf("Load route from %s\n", filename);
  XML_Parser parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser, &data);
  XML_SetElementHandler(parser, route_load_startElement, route_load_endElement);
  XML_SetCharacterDataHandler(parser, route_load_charElement);

  //iterate items
  int done;
  do
  {
    void *buf = XML_GetBuffer(parser, BUFF_SIZE);
    if (buf == NULL)
    {
      printf("Error alloc buffer\n");
      gpx_unlock();
      if (route->route)
      {
        route_freeroute(route->route);
      }
      free(route);
      fclose(f);
      return NULL;
    }
    int len = (int)fread(buf, 1, BUFF_SIZE, f);
    //printf("LEN: %d\n", len);
    done = (int)(len < BUFF_SIZE);
    if (! XML_ParseBuffer(parser, len, done))
    {
      fprintf(stderr,
              "%s at line %lu\n",
              XML_ErrorString(XML_GetErrorCode(parser)),
              XML_GetCurrentLineNumber(parser));
      if (route->route)
      {
        route_freeroute(route->route);
      }
      free(route);
      fclose(f);
      return NULL;
    }
  } while (!done && (data.error==0));
  fclose(f);

  //free_all
  if (data.nodename)
  {
    free(data.nodename);
    data.nodename = NULL;
  }

  //error?
  if (data.error)
  {
    if (route->route)
    {
      route_freeroute(route->route);
    }
    free(route);
    return NULL;
  }

  //create togo
  /*route_node *end = route_lastnode(route->route);
  if (end)
  {
    end->togo = 0;
    end->totake = 0;
    end = end->prev_node;
    
    while (end)
    {
      int len = wgs84_distance(end->pos.lat, end->pos.lon, end->next_node->pos.lat, end->next_node->pos.lon);
      int t = len * 36 / end->next_node->speed;
      end->togo = end->next_node->togo + len;
      end->totake = end->next_node->totake + t;
      end = end->prev_node;
    }
  }*/
  
  //create directions
  route_maketurn(route->vehicle, route->route, NULL);
  
  return route;
}

//save route
int route_save(char *filename, route_route *route)
{
  FILE *f;
  char s[255];
  if ((f = fsafeopen(filename, "w")) == NULL)
  {
    printf("Error opening %s file for writing!\n", filename);
    return -1;
  }

  fprintf(f, "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>\n");
  fprintf(f, "<gpx version=\"1.1\" creator=\"MC Navi\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.topografix.com/GPX/1/1\" xsi:schemaLocation=\"http://www.topografix.com/GPX/1/1\">\n");
  fprintf(f, "<trk>\n");
  fprintf(f, "  <name>Noname</name>\n");
  fprintf(f, "  <vehicle>%s</vehicle>\n", vehiclelabel(route->vehicle));
  fprintf(f, "  <trkseg>\n");

  route_node *node = route->route;
  route_itinerary *iti = route->itinerary;
  time_t t = time(0);
  float tt = 0;
  map_position pos;
  int first = 1;
  while (node)
  {
    //node
    fprintf(f, "    <trkpt id=\"%ld\" lat=\"%ld.%06ld\" lon=\"%ld.%06ld\">\n", node->id,
             node->pos.lat/1000000, node->pos.lat%1000000, node->pos.lon/1000000, node->pos.lon%1000000);
    if (node->nodetype)
    {
      fprintf(f, "      <nodetype>%s</nodetype>\n", itinodelabel(node->nodetype));
    }
    const char *waytype = maptypelabel(node->waytype);
    if (waytype)
    {
      fprintf(f, "      <waytype>%s</waytype>\n", waytype);
    }
    fprintf(f, "      <speed>%d</speed>\n", node->speed);
    if (node->direction)
    {
      fprintf(f, "      <direction>%d</direction>\n", node->direction);
    }
    //time
    if (first)
    {
      first = 0;
    }
    else
    {
      int len = wgs84_distance(pos.lat, pos.lon, node->pos.lat, node->pos.lon);
      int speed = node->speed;
      vehiclespeed(route->vehicle, node->waytype, &speed);
      tt += (float)len / (float)speed * 3.6;
    }
    time_t t2 = t + (int)tt;
    fprintf(f, "      <time>");
    struct tm lt;
    gmtime_r (&t2, &lt);
    int milisec = (t2 - trunc(t2)) * 100;
    strftime(s, sizeof(s), "%Y-%m-%dT%H:%M:%S", &lt);
    fputs(s, f);
    snprintf(s, sizeof(s), ".%02dZ", milisec);
    fputs(s, f);
    fprintf(f, "</time>\n");
    if (iti && ((node->nodetype == nt_start) ||
        (node->nodetype == nt_direction) ||
        (node->nodetype == nt_destination) ||
        (node->nodetype == nt_waypoint)))
    {
      if (iti->node == node)
      {
        if (iti->name)
          fprintf(f, "      <name>%s</name>\n", iti->name);
      }
      else
      {
        printf("Error in itinerary !\n");
      }
      iti = iti->next;
    }
    pos = node->pos;
    //snprintf(s, sizeof(s), "      <ele>%1.1f</ele>\n", p->points[i].altitude);
    //fprintf(f, s);
    //snprintf(s, sizeof(s), "      <speed>%1.1f</speed>\n", p->points[i].speed);
    //fprintf(f, s);
    fprintf(f, "    </trkpt>\n");
    node = node->next_node;
  }

  fprintf(f, "  </trkseg>\n");
  fprintf(f, "</trk>\n");
  fprintf(f, "</gpx>\n");
  fclose(f);

  return 0;
}

