/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * sounds.cc
 * Copyright (C) Mike Crash 2011 <mike@mikecrash.com>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include "sounds.h"
#include "mcthread.h"
#include "mcconf.h"

#define soundsqueuesize 30
typedef struct
{
  //queue
  int queue_index;
  int queue_len;
  sounds_type queue[soundsqueuesize];

  //internal
  int terminate;
  mcmutex mutex;
  mcsema sema;
  pthread_t thread;
} soundsdata;

char *sounds_files[sounds_type_count];

soundsdata *sounds_data = NULL;

void *sounds_thread(void *p)
{
  printf("Sounds start\n");
  sounds_data = (soundsdata*)p;
 
  while (sounds_data->terminate==0)
  {
    //wait for signal
    mcsema_wait(&sounds_data->sema);
    if (sounds_data->terminate)
      break;

    //check the required work
    sounds_type sound = sound_none;
    char **argv = NULL;
    mcmutex_lock(&sounds_data->mutex);
    if (sounds_data->queue_len > 0)
    {
      argv = (char **)malloc(sizeof(char*) * (sounds_data->queue_len + 3));
      argv[0] = config->sounds_cmd;
      argv[1] = (char*)"-q";
      int i = 2;
      while (sounds_data->queue_len>0)
      {
        int sound = (int)sounds_data->queue[sounds_data->queue_index];
        if (sounds_files[(int)sound] != NULL)
        {
          argv[i] = sounds_files[(int)sound];
          i++;
        }
        sounds_data->queue_index++;
        if (sounds_data->queue_index>=soundsqueuesize)
          sounds_data->queue_index -= soundsqueuesize;
        sounds_data->queue_len--;
      }
      argv[i] = NULL;
    }
    mcmutex_unlock(&sounds_data->mutex);

    //play required sound
    if (argv)
    {
      pid_t pid = fork();
      if (pid == 0)
      {
        execvp(argv[0], argv);
        exit(0);
      }
      else if (pid<0)
      {
        printf("Error running program %s\n", config->sounds_cmd);
      }
      else
      {
        int status;
        waitpid(pid, &status, 0);
      }
      free(argv);
    }
  }
  printf("Sounds end\n");
  return NULL;
}

void sounds_load(char *filename)
{
  char s[500];
  FILE *f;
  
  if ((f = fopen(filename, "r"))==NULL)
    return;

  while (!feof(f))
  {
    if (fgets(s, sizeof(s), f) == NULL)
      break;
    if (s[0] == 0)
      continue;
    if (s[0] == '#')
      continue;

    //get cmd and value
    char *cmd, *value, *strip;
    cmd = s;
    value = strchr(s, '=');
    if (value == NULL)
      continue;
    *value = 0;
    value++;
    strip = strchr(value, '\n');
    if (strip)
      *strip = 0;
    strip = strchr(value, '\r');
    if (strip)
      *strip = 0;
    /*cmd = strtok(s, "= \n\r");
    if (cmd == NULL)
      continue;
    value = strtok(NULL, "= \n\r");*/

    int index = atoi(cmd);
    if ((index>0) && (index<sounds_type_count))
    {
      if (sounds_files[index])
      {
        free(sounds_files[index]);
        sounds_files[index] = NULL;
      }
      char *sf = (char*)malloc(strlen(filename)+strlen(value)-8);
      strcpy(sf, filename);
      char *sf2 = sf + strlen(sf) - 12;
      strcpy(sf2, value);
      sounds_files[index] = sf;
    }
  }
  fclose(f);
}

int sounds_init(char *style)
{
  if (sounds_data || (style==NULL))
    return -1;

  printf("Sounds init %s\n", style);

  sounds_data = (soundsdata*)malloc(sizeof(soundsdata));
  memset(sounds_data, 0, sizeof(soundsdata));
  memset(sounds_files, 0, sizeof(sounds_files));

  //load sound style
  char *sf = (char*)malloc(strlen(style)+strlen(config->datadir)+25);
  sprintf(sf, "%s/sounds/%s/sounds.style", config->datadir, style);
  printf("%s\n", sf);
  sounds_load(sf);
  free(sf);
 
  mcmutex_init(&sounds_data->mutex);
  mcsema_init(&sounds_data->sema);
  mcthread_create(&sounds_data->thread, sounds_thread, sounds_data, 0);

  return 0;
}

int sounds_play(sounds_type *sound, int count)
{
  if (sounds_data == NULL)
    return -1;

  mcmutex_lock(&sounds_data->mutex);
  while ((sounds_data->queue_len < soundsqueuesize) && (count>0))
  {
    int index = sounds_data->queue_index + sounds_data->queue_len;
    if (index>=soundsqueuesize)
      index -= soundsqueuesize;
    sounds_data->queue[index] = *sound;
    sounds_data->queue_len++;
    sound++;
    count--;
  }
  if (count>0)
  {
    printf("Sounds queue full!\n");
  }
  mcmutex_unlock(&sounds_data->mutex);
  mcsema_signal(&sounds_data->sema);

  return 0;
}

int sounds_deinit()
{
  if (sounds_data == NULL)
    return -1;
  
  sounds_data->terminate = 1;
  mcsema_signal(&sounds_data->sema);
  mcthread_join(&sounds_data->thread);
  mcsema_destroy(&sounds_data->sema);
  mcmutex_destroy(&sounds_data->mutex);
  free(sounds_data);
  sounds_data = NULL;

  for (int i=0; i<sounds_type_count; i++)
  {
    if (sounds_files[i])
    {
      free(sounds_files[i]);
      sounds_files[i] = NULL;
    }
  }

  printf("Sounds gone\n");

  return 0;
}

