/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * map_planes.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <typeio.h>
#include <math.h>
#include "polygon.h"
#include "map_common.h"
#include "map_text.h"
#include "map_planes.h"
#include "mcconf.h"


// read all plane sectors from file
int readplaneectors()
{
  map_part part; 
  fseek(map->f, map->info.planesindex, SEEK_SET);
  map_readpart(&part);
  if (part.partid!=MPID_PLANESINDEX)
    return -1;

  int i = 0;
  plane_sector *prev = NULL;
  while (i<part.count)
  {
    //alloc sector
    plane_sector *sector = (plane_sector*)malloc(sizeof(plane_sector));
    sector->next = NULL;
    sector->prev = prev;
    if (i==0)
      map->planes.sectors = sector;
    if (prev)
      prev->next = sector;

    //read
    fread_long(&sector->lat, 1, map->f);
    fread_long(&sector->lon, 1, map->f);
    fread_char(&sector->zoom, 1, map->f);
    fread_long(&sector->count, 1, map->f);
    fread_long(&sector->filepos, 1, map->f);
    fread_long(&sector->size, 1, map->f);

    //init
    sector->used = 0;
    sector->mem = 0;
    sector->planes = NULL;
    prev = sector;

    i++;
  }
  
  return 0;
}

int readplanesector(plane_sector *sector)
{
  int size = 0;
  int count = sector->count;
  char *buf = (char*)malloc(sector->size);

  //printf("Read plane sector %ld,%ld,%d pos %ld size %ld count %ld\n", sector->lat, sector->lon, sector->zoom, sector->filepos, sector->size, sector->count);

  map_lock(map);
  fseek(map->f, map->info.planes + map_part_size + sector->filepos, SEEK_SET);
  fread_char(buf, sector->size, map->f);

  sector->planes = (mplanecache **)malloc(sizeof(mplanecache*) * cntplanetypes);
  memset(sector->planes, 0, sizeof(mplanecache*) * cntplanetypes);

  int check = 0;
  int i;
  char *p = buf;
  for (i=0; i<count; i++)
  {
    int ch = (int)read_char(&p);
    maptypes mt = (maptypes)ch;
    ch -= firstplanetype;
    int mpcount = read_int(&p);
    check += map_plane_size + map_multipolygon_size;
    //printf("  Plane %d - %d - %d\n", mt, mpcount, check);

    mplanecache *mc = (mplanecache *)malloc(sizeof(mplanecache));
    mc->plane.planetype = mt;
    mc->plane.mp = (map_multipolygon *)malloc(sizeof(map_multipolygon)+mpcount*sizeof(map_polygon *));
    mc->plane.mp->count = mpcount;
    int j;
    for (j=0; j<mpcount; j++)
    {
      int pcount = read_int(&p);
      //printf("    Poly %d\n", pcount);
      map_polygon *mp = (map_polygon *)malloc(sizeof(map_polygon)+pcount*sizeof(map_position));
      mp->count = pcount;
      int k;
      for (k=0; k<pcount; k++)
      {
        mp->points[k].lat = read_long(&p);
        mp->points[k].lon = read_long(&p);
      }
      mc->plane.mp->polygon[j] = mp;
      size += sizeof(map_polygon)+pcount*sizeof(map_position);
      check += map_multipolygon_size + pcount * map_position_size;
    }
    map_polygonarea(mc->plane.mp->polygon[0], &mc->lat1,  &mc->lon1,  &mc->lat2,  &mc->lon2);
    mc->next = sector->planes[ch];
    sector->planes[ch] = mc;
    size += sizeof(mplanecache)+mpcount*sizeof(map_polygon *);
  }
  
  //printf("  Size %d\n", check);
  free(buf);
  sector->mem = size + sizeof(mplanecache*) * cntplanetypes;
  map->planes.mem += sector->mem;
  map->planes.count++;
  map_unlock(map);
  return 0;
}

int leaveplanesector(plane_sector *sector)
{
  int i;
  for (i=0; i<cntplanetypes; i++)
  {
    mplanecache *mc = sector->planes[i];
    while (mc)
    {
      mplanecache *mc2 = mc->next;
      map_freemultipolygon(mc->plane.mp);
      free(mc);
      mc = mc2;
    }
  }
  free(sector->planes);
  sector->planes = NULL;
  map->planes.mem -= sector->mem;
  map->planes.count--;
  sector->mem = 0;
  return 0;
}

//bring to top
int planesector2top(plane_sector *sector)
{
  if (sector == map->planes.sectors)
    return 0;
  if (sector->next)
    sector->next->prev = sector->prev;
  if (sector->prev)
    sector->prev->next = sector->next;
  map->planes.sectors->prev = sector;
  sector->next = map->planes.sectors;
  map->planes.sectors = sector;
  sector->prev = NULL;
  return 0;
}

//find all required sectors and update it
int update_planecache(long lat1, long lon1, long lat2, long lon2)
{
  int updated = 0;
  int checked = 0;
  plane_sector *sector = map->planes.sectors;
  long zlat = (lat2 - lat1) / MAPSECTRECT;
  long zlon = (lon2 - lon1) / MAPSECTRECT;
  int zoom = zlat>zlon ? zlat : zlon;
  if (zoom>0)
  {
    zoom = (int)(log((double)zoom)/log(2.0)+1);
  }
  else
  {
    zoom = 0;
  }
  if (zoom>5)
    zoom = zoom - 5;
  else
    zoom = 0;
  //lat1 = lat1 / MAPSECTRECT * MAPSECTRECT - MAPSECTRECT;
  //lon1 = lon1 / MAPSECTRECT * MAPSECTRECT - MAPSECTRECT;
  //lat2 = lat2 / MAPSECTRECT * MAPSECTRECT + MAPSECTRECT;
  //lon2 = lon2 / MAPSECTRECT * MAPSECTRECT + MAPSECTRECT;

  while (sector)
  {
    plane_sector *next = sector->next;
    //TODO check
    long zoomd = (long)pow(2, (double)sector->zoom);
    //long slat1 = sector->lat / (zoomd * MAPSECTRECT) * (zoomd * MAPSECTRECT);
    //long slon1 = sector->lon / (zoomd * MAPSECTRECT) * (zoomd * MAPSECTRECT);
    //long slat2 = (sector->lat / (zoomd * MAPSECTRECT) + 1) * (zoomd * MAPSECTRECT);
    //long slon2 = (sector->lon / (zoomd * MAPSECTRECT) + 1) * (zoomd * MAPSECTRECT);
    long slat1 = sector->lat - zoomd * MAPSECTRECT / 2;
    long slon1 = sector->lon - zoomd * MAPSECTRECT / 2;
    long slat2 = sector->lat + zoomd * MAPSECTRECT * 3 / 2;
    long slon2 = sector->lon + zoomd * MAPSECTRECT * 3 / 2;
    //if ((sector->zoom>=zoom) && (sector->lat>=slat1) && (sector->lon>=slon1) && (sector->lat<=slat2) && (sector->lon<=slon2))
    if ((sector->zoom>=zoom) && !((lat1>slat2) || (lon1>slon2) || (lat2<slat1) || (lon2<slon1)))
    {
      //printf("Sect %9ld,%9ld  %9ld,%9ld %9ld,%9ld %d\n", slat1, slon1, slat2, slon2, sector->lat, sector->lon, sector->zoom);
      if (sector->planes==NULL)
      {
        readplanesector(sector);
        updated = 1;
      }
      planesector2top(sector);
      checked++;
    }
    sector = next;
  }
  
  if (updated)
  {
    printf("Cached planes: %d memory: %d kB\n", map->planes.count, map->planes.mem/1024);
  }

  return checked;
}

// *****************************
// init and deinit
// *****************************

int map_planes_init(void)
{
  if (readplaneectors() != 0)
    return -1;
  return 0;
}

int map_planes_clear(void)
{
  plane_sector *sector = map->planes.sectors;
  while (sector)
  {
    plane_sector *sector2 = sector->next;
    if (sector->planes)
      leaveplanesector(sector);
    free(sector);
    sector = sector2;
  }
  map->planes.sectors = NULL;
  map->planes.count = 0;
  map->planes.mem = 0;
  return 0;
}

// *******************
// plane search
// *******************
//TODO include zoom
mapplanesearch *map_planes_start(long lat1, long lon1, long lat2, long lon2)
{
  mapplanesearch *ps;
  //printf("Area %9ld,%9ld  %9ld,%9ld\n", lat1, lon1, lat2, lon2);
  ps = (mapplanesearch *)malloc(sizeof(mapplanesearch));
  ps->lat1 = lat1;
  ps->lon1 = lon1;
  ps->lat2 = lat2;
  ps->lon2 = lon2;
  ps->check = update_planecache(lat1, lon1, lat2, lon2);
  //if (ps->planetype == PT_SEA)
  //printf("Find plane %ld,%ld->%ld,%ld (%d-%d)\n", ps->lat1, ps->lon1, ps->lat2, ps->lon2, firstplanetype, maxplanetype);
  return ps;
}

void map_planes_type(mapplanesearch *ps, maptypes planetype)
{
  ps->planetype = planetype;
  ps->sector = (void*)map->planes.sectors;
  ps->plane = NULL;
  ps->checked = 0;
  //if (ps->planetype == PT_SEA)
  //  printf("Type %d\n", ps->planetype);
}

mplane *map_planes_get(mapplanesearch *ps)
{
  plane_sector *sector = (plane_sector *)ps->sector;
  mplanecache *plane = (mplanecache *)ps->plane;
  //printf("Area %9ld,%9ld  %9ld,%9ld\n", ps->lat1, ps->lon1, ps->lat2, ps->lon2);

  while (sector && (ps->checked<ps->check))
  {
    //is cache area part of requested area?
    if (sector->planes)  //TODO check sector coordinates and requested area
    {
      //printf("  Search sector %ld,%ld,%d type %d\n", sector->lat, sector->lon, sector->zoom, ps->planetype);
      int pi = ps->planetype - firstplanetype;
      if (plane==NULL)
        plane = sector->planes[pi];
      while (plane)
      {
        //if ((plane->lat1==61300000)&&(plane->lon1==-6800000)&&(plane->lat2==61500000)&&(plane->lon2==-6600000))
        //  printf("???\n");
        //if (ps->planetype == PT_SEA)
        //  printf(" Sea %9ld,%9ld  %9ld,%9ld %d\n", plane->lat1, plane->lon1, plane->lat2, plane->lon2, sector->zoom);
        if (!((plane->lat2<ps->lat1) ||
            (plane->lat1>ps->lat2) ||
            (plane->lon2<ps->lon1) ||
            (plane->lon1>ps->lon2)))
        {
//          if (ps->planetype == PT_SEA)
//            printf(" Sea %9ld,%9ld  %9ld,%9ld\n", plane->lat1, plane->lon1, plane->lat2, plane->lon2);
//            printf("   From %9ld,%9ld  %d\n", sector->lat, sector->lon, sector->zoom);
//          if (ps->planetype == PT_SEA)
//            printf("YES\n");
          ps->plane = plane->next;
          if (ps->plane == NULL)
          {
            ps->sector = sector->next;
            ps->checked++;
          }
          else
            ps->sector = sector;
          return &plane->plane;
        }
//        if (ps->planetype == PT_SEA)
//          printf("NO\n");
        plane = plane->next;
      }
    }
    sector = sector->next;
    plane = NULL;
    ps->checked++;
  }
  ps->sector = NULL;
  ps->plane = NULL;
  return NULL;
}

int map_planes_stop(mapplanesearch *ps)
{
  free(ps);
  /*int updated = 0;
  while (map->planes.mem > config->limit_planes)
  {
    free_planesect(map->planes.sectcacheend);
    updated = 1;
  }
  if (updated)
  {
    printf("Free planes cached %ld, points %ld (%ld kB)\n", map->planes.count, map->planes.size, (map->planes.count * sizeof(mplane) + map->planes.size * sizeof(map_position))/1024);
  }*/
  //printf("Done\n");
  return 0;
}
