/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * map_pois.cc
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 * 
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <typeio.h>
#include <string.h>
#include "map_common.h"
#include "map_pois.h"
#include "map_text.h"
#include "mcconf.h"


// *****************************
// free cache functions
// *****************************
void poicache_free(mpoicache *cache)
{
  mpoicache *next;
  while (cache)
  {
    next = cache->next;
    free(cache);
    cache = next;
  }
}

// *****************************
// read all plane sectors from file
// *****************************
int map_poiindexread(void)
{
  map_part part;
  long cnt;

  map_lock(map);
  fseek(map->f, map->info.poisindex, SEEK_SET);
  //printf("Planes: %ld\n",map->info.planes);

  //check index type
  map_readpart(&part);
  //printf("Part: %d , %ld\n", part.partid, part.count);
  if (part.partid != MPID_POISINDEX)
  {
    map_unlock(map);
    printf("Bad POI part\n");
    return -1;
  }

  //alloc new index
  cnt = part.count;

  //read the index
  long i = 0;
  poi_sector *prev = NULL;
  while ((i<part.count) && !feof(map->f))
  {
    poi_sector *sector = (poi_sector *)malloc(sizeof(poi_sector));
    fread_long(&sector->lat, 1, map->f);
    fread_long(&sector->lon, 1, map->f);
    fread_long(&sector->count, 1, map->f);
    fread_long(&sector->filepos, 1, map->f);
    fread_long(&sector->size, 1, map->f);
    sector->mem = 0;
    sector->used = 0;
    sector->pois = NULL;
    sector->next = NULL;
    sector->prev = prev;
    if (i==0)
      map->pois.sectors = sector;
    if (prev)
      prev->next = sector;
    prev = sector;
    //printf("POI Sector %ld,%ld: %ld %ld\n", pi->pos.lat, pi->pos.lon, pi->count, pi->filepos);
    i++;
  }
  map->pois.count = 0;
  map->pois.mem = 0;
  map_unlock(map);
  return 0;
}

//read sector from file
int readpoisect(poi_sector *sector)
{
  //printf("POI %ld,%ld Count: %ld Filepos: %ld Size: %ld Pos: %lx\n", sector->lat, sector->lon, sector->count, sector->filepos, sector->size, 
  //       map->info.pois + map_part_size + sector->filepos);
  //read it
  char *buf = (char*)malloc(sector->size);
  map_lock(map);
  fseek(map->f, map->info.pois + map_part_size + sector->filepos, SEEK_SET);
  fread_char(buf, sector->size, map->f);
  char *p = buf;

  sector->pois = (mpoicache **)malloc(sizeof(mpoicache*) * cntpoitypes);
  memset(sector->pois, 0, sizeof(mpoicache*) * cntpoitypes);

  int i = 0;
  while (i<sector->count)
  {
    mpoicache *pc = (mpoicache *)malloc(sizeof(mpoicache));
    pc->poi.poitype = (maptypes)read_uchar(&p);
    pc->poi.pos.lat = read_long(&p);
    pc->poi.pos.lon = read_long(&p);
    pc->poi.flags = read_uchar(&p);
    pc->poi.namepos = read_long(&p);
    pc->poi.name = NULL;

    int typ = (int)pc->poi.poitype - firstpoitype;
    pc->next = sector->pois[typ];
    sector->pois[typ] = pc;
    sector->mem += sizeof(mpoicache);
    i++;
  }
  map->pois.mem += sector->mem;
  map->pois.count++;
  free(buf);
  map_unlock(map);
  return 0;
}

int leavepoisector(poi_sector *sector)
{
  int i;
  for (i=0; i<cntpoitypes; i++)
  {
    mpoicache *mc = sector->pois[i];
    while (mc)
    {
      mpoicache *mc2 = mc->next;
      if (mc->poi.namepos && mc->poi.name)
        map_leavetext(mc->poi.namepos);
      free(mc);
      mc = mc2;
    }
  }
  free(sector->pois);
  map->pois.mem -= sector->mem;
  map->pois.count--;
  sector->pois = NULL;
  sector->mem = 0;
  return 0;
}

//bring to top
int poisector2top(poi_sector *sector)
{
  if (sector == map->pois.sectors)
    return 0;
  if (sector->next)
    sector->next->prev = sector->prev;
  if (sector->prev)
    sector->prev->next = sector->next;
  map->pois.sectors->prev = sector;
  sector->next = map->pois.sectors;
  map->pois.sectors = sector;
  sector->prev = NULL;
  return 0;
}

//find all required sectors and update it
int update_poicache(long lat1, long lon1, long lat2, long lon2)
{
  lat1 = MAPSECTOR(lat1);
  lon1 = MAPSECTOR(lon1);
  lat2 = MAPSECTOR(lat2);
  lon2 = MAPSECTOR(lon2);
  poi_sector *sector = map->pois.sectors;
  int checked = 0;
  int check = ((lat2-lat1) / MAPSECTRECT + 1) * ((lon2-lon1) / MAPSECTRECT + 1);
  int updated = 0;

  while (sector && (checked<check))
  {
    poi_sector *next = sector->next;
    if ((sector->lat>=lat1) && (sector->lat<=lat2) && (sector->lon>=lon1) && (sector->lon<=lon2))
    {
      if ((sector->pois==NULL))
      {
        readpoisect(sector);
        updated++;
      }
      poisector2top(sector);
      checked++;
    }
    sector = next;
  }
  if (updated)
  {
    printf("Cached POIs: %d memory: %d kB\n", map->pois.count, map->pois.mem/1024);
  }
  return checked;
}


//free cache for maximum count
int freepoicache()
{
  int updated = 0;
  //printf("%ld - %ld\n", map->nodes.mem, config->limit_nodes);
  //check limit
  if (map->pois.mem < config->limit_pois)
    return 0;

  //fins last cached sector
  poi_sector *sector = map->pois.sectors;
  while (sector && sector->next && sector->next->pois)
  {
    sector = sector->next;
  }

  //free sectors
  while (sector && (map->pois.mem > config->limit_pois))
  {
    if (sector->pois)
    {
      leavepoisector(sector);
      updated = 1;
    }
    sector = sector->prev;
  }

  if (updated)
  {
    printf("Cached POIs: %d memory: %d kB\n", map->pois.count, map->pois.mem/1024);
  }
  return 0;
}


// *****************************
// init and deinit
// *****************************

int map_pois_init(void)
{
  if (map_poiindexread() != 0)
    return -1;

  //check POIs part
  map_part part; 
  map_lock(map);
  fseek(map->f, map->info.pois, SEEK_SET);
  map_readpart(&part);
  map_unlock(map);
  if (part.partid!=MPID_POIS)
    return -1;

  return 0;
}

int map_pois_clear(void)
{
  poi_sector *sector = map->pois.sectors;
  while (sector)
  {
    poi_sector *sector2 = sector->next;
    if (sector->pois)
    {
      leavepoisector(sector);
    }
    free(sector);
    sector = sector2;;
  }

  map->pois.sectors = NULL;
  map->pois.count = 0;
  map->pois.mem = 0;
  return 0;
}

// *******************
// plane search
// *******************

mappoisearch *map_pois_start(long lat1, long lon1, long lat2, long lon2)
{
  mappoisearch *ps;
  
  ps = (mappoisearch *)malloc(sizeof(mappoisearch));
  ps->lat1 = lat1;
  ps->lon1 = lon1;
  ps->lat2 = lat2;
  ps->lon2 = lon2;
  ps->names = 0;
  ps->checked = 0;
  ps->check = update_poicache(lat1, lon1, lat2, lon2);

  ps->sector = map->pois.sectors;
  ps->poi = NULL;
  //printf("Find POI %ld,%ld->%ld,%ld %d (%d-%d)\n", ps->lat1, ps->lon1, ps->lat2, ps->lon2, ps->poitype, firstpoitype, maxpoitype);
  return ps;
}

void map_pois_type(mappoisearch *ps, maptypes poitype, int names)
{
  ps->poitype = poitype;
  ps->names = names;
  ps->checked = 0;
  ps->sector = map->pois.sectors;
  ps->poi = NULL;
}

mpoi *map_pois_get(mappoisearch *ps)
{
  poi_sector *sector = (poi_sector *)ps->sector;
  mpoicache *poi = (mpoicache *)ps->poi;

  while (sector && (ps->checked<ps->check))
  {
    //printf("Find plane %ld,%ld->%ld,%ld\n", plane->lat1, plane->lon1, plane->lat2, plane->lon2);
    //is cache area part of requested area?
    if (sector->pois)
    {
      if (poi==NULL)
        poi = sector->pois[ps->poitype - firstpoitype];
      while (poi)
      {
        if ((poi->poi.pos.lat>=ps->lat1) &&
            (poi->poi.pos.lat<=ps->lat2) &&
            (poi->poi.pos.lon>=ps->lon1) &&
            (poi->poi.pos.lon<=ps->lon2))
        {
          ps->poi = poi->next;
          if (ps->poi == NULL)
          {
            ps->sector = sector->next;
            ps->checked++;
          }
          else
            ps->sector = sector;
          if (ps->names && poi->poi.namepos && (poi->poi.name == NULL))
          {
            poi->poi.name = map_gettext(poi->poi.namepos);
          }
          return &poi->poi;
        }
        poi = poi->next;
      }
    }
    sector = sector->next;
    ps->checked++;
    poi = NULL;
  }
  ps->sector = sector;
  ps->poi = poi;
  return NULL;
}

int map_pois_stop(mappoisearch *ps)
{
  free(ps);
  freepoicache();
  return 0;
}
