/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * main.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 *
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>
#if !defined(WIN32)
#include <unistd.h>
#endif
#include <expat.h>
#include <string.h>
#include "projection.h"
#include "polygon.h"
#include "db_mysql.h"
#include "map_common.h"
#include "map_relation.h"
#include "map_ways.h"
#include "map_convert.h"
#include "map_coastlines.h"
#include "map_write.h"


#ifdef XML_LARGE_SIZE
#if defined(XML_USE_MSC_EXTENSIONS) && _MSC_VER < 1400
#define XML_FMT_INT_MOD "I64"
#else
#define XML_FMT_INT_MOD "ll"
#endif
#else
#define XML_FMT_INT_MOD "l"
#endif


#if !defined(WIN32)
#define _FILE_OFFSET_BITS 64
#endif

#ifdef _WIN32
#define fseeko _fseeki64
#define ftello _ftelli64
//typedef __int64 off_t;
#else
#define fseeko fseek
#define ftello ftell
typedef long off_t;
#endif

#define BUFF_SIZE 1000

#define MAXWAYNODES 500000
#define MAXRELMEMBERS 50000

#define ELM_NODE 1
#define ELM_WAY 2
#define ELM_RELATION 3

int topelement;
int osmerror;

long long node_id = 0;

long long way_id = 0;
long long *way_nodes;
int way_nodes_cnt = 0;

long long rel_id = 0;
relation_member *rel_members;
int rel_members_cnt = 0;



static char *attr_str(const char **attr, const char *name)
{
  int i;
  for (i = 0; attr[i]; i += 2)
  {
    if (strcmp(attr[i], name) == 0)
      return (char*)attr[i+1];
  }
  return NULL;
}

static int attr_int(const char **attr, const char *name, long *val)
{
  char *an;
  if ((an = attr_str(attr, name)) != NULL)
  {
    *val = atoi(an);
    return 0;
  }
  return -1;
}

static int attr_longlong(const char **attr, const char *name, long long *val)
{
  char *an;
  if ((an = attr_str(attr, name)) != NULL)
  {
    *val = atoi(an);
    return 0;
  }
  return -1;
}

static double attr_float(const char **attr, const char *name, float *val)
{
  char *an;
  if ((an = attr_str(attr, name)) != NULL)
  {
    *val = (float)atof(an);
    return 0;
  }
  return -1;
}

static double attr_double(const char **attr, const char *name, double *val)
{
  char *an;
  if ((an = attr_str(attr, name)) != NULL)
  {
    *val = atof(an);
    return 0;
  }
  return -1;
}

void save_tag(const char *table, long long id, const char *name, const char *val)
{
  long sl = (long)strlen(name);
  char *name2 = (char*)malloc(sl*2+1);
  sql_escape(name, name2, sl);
  sl = (long)strlen(val);
  char *val2 = (char*)malloc(sl*2+1);
  sql_escape(val, val2, sl);
  char *sql = (char *)malloc(strlen(table)+strlen(name2)+strlen(val2)+100);
  sprintf(sql, "INSERT INTO %s (id, name, val) VALUES ('%lld', '%s', '%s')", table, id, name2, val2);
  if (sql_query(sql) != 0)
  {
    printf("Error inserting tag\n");
  }
  free(sql);
  free(name2);
  free(val2);
}


static void XMLCALL startElement(void *userData, const char *name, const char **attr)
{
  int *depthPtr = (int *)userData;

  if  (*depthPtr == 1)
  {
    if (strcmp(name, "node") == 0)
    {
      //add nodes
      double flat, flon;
      char sql[100];
      if (attr_longlong(attr, "id", &node_id) < 0) return;
      if (attr_double(attr, "lat", &flat) < 0) return;
      if (attr_double(attr, "lon", &flon) < 0) return;
      long lat = (long)(flat * 1000000);
      long lon = (long)(flon * 1000000);
      sprintf(sql, "INSERT INTO osm_nodes (id, lat, lon) VALUES ('%lld', '%ld', '%ld')", node_id, lat, lon);
      if (sql_query(sql) == 0)
      {
        topelement = ELM_NODE;
      }
      else
      {
        printf("Error inserting node\n");
      }
      //printf("\nN: %lld\n", id);
    }
    else if (strcmp(name, "way") == 0)
    {
      if (attr_longlong(attr, "id", &way_id) < 0) return;
      topelement = ELM_WAY;
      way_nodes_cnt = 0;
      //printf("\nW: %lld\n", id);
    }
    else if (strcmp(name, "relation") == 0)
    {
      if (attr_longlong(attr, "id", &rel_id) < 0) return;
      topelement = ELM_RELATION;
      rel_members_cnt = 0;
      //printf("\nR: %lld\n", id);
    }
  }
  else if  (*depthPtr == 2)
  {
    if (topelement == ELM_NODE)
    {
      //node tags
      if (strcmp(name, "tag") == 0)
      {
        char *name;
        char *val;
        if ((name = attr_str(attr, "k")) == NULL) return;
        if ((val = attr_str(attr, "v")) == NULL) return;
        save_tag("osm_nodes_tags", node_id, name, val);
        //printf("\n%s=%s\n", name, val);
      }
    }
    else if (topelement == ELM_WAY)
    {
      //way tags
      if (strcmp(name, "nd") == 0)
      {
        if (way_nodes_cnt<MAXWAYNODES)
        {
          long long ref;
          if (attr_longlong(attr, "ref", &ref) < 0) return;
          way_nodes[way_nodes_cnt] = ref;
          way_nodes_cnt++;
        }
        else
        {
          printf("Warning: maximum way nodes count reached\n");
          osmerror = 1;
        }
      }
      else if (strcmp(name, "tag") == 0)
      {
        char *name;
        char *val;
        if ((name = attr_str(attr, "k")) == NULL) return;
        if ((val = attr_str(attr, "v")) == NULL) return;
        save_tag("osm_ways_tags", way_id, name, val);
        //process_tags(name, val, &way->tags);
      }
    }
    else if (topelement == ELM_RELATION)
    {
      if (strcmp(name, "member") == 0)
      {
        if (rel_members_cnt<MAXRELMEMBERS)
        {
          char *s;
          if ((s = attr_str(attr, "type")) == NULL) return;
            rel_members[rel_members_cnt].type = map_relation_member(s);
          if ((s = attr_str(attr, "role")) != NULL)
            rel_members[rel_members_cnt].role = map_relation_role(s);
          else
            rel_members[rel_members_cnt].role = 0;
          if (attr_longlong(attr, "ref", &rel_members[rel_members_cnt].id) < 0) return;
          rel_members_cnt++;
        }
        else
        {
          printf("Warning: maximum relation members count reached\n");
          osmerror = 1;
        }
      }
      else if (strcmp(name, "tag") == 0)
      {
        char *name;
        char *val;
        if ((name = attr_str(attr, "k")) == NULL) return;
        if ((val = attr_str(attr, "v")) == NULL) return;
        save_tag("osm_relations_tags", rel_id, name, val);
      }
    }
  }
  else
  {

  }
  *depthPtr += 1;
}

static void XMLCALL endElement(void *userData, const char *name)
{
  int *depthPtr = (int *)userData;
  *depthPtr -= 1;

  if (*depthPtr == 1)
  {
    //NODE
    if ((strcmp(name, "node") == 0) && (topelement == ELM_NODE))
    {
      topelement = 0;
    }
    //WAY
    else if (((strcmp(name, "way") == 0) && (topelement == ELM_WAY)) && (osmerror == 0))
    {
      if (way_nodes_cnt>1)
      {
        char *sql = (char *)malloc(way_nodes_cnt*16+100);
        sprintf(sql, "INSERT INTO osm_ways (id, nodes) VALUES ('%lld', 0x", way_id);
        int i = (int)strlen(sql);
        char *arr = sql+i;
        for (i=0; i<way_nodes_cnt; i++)
        {
          sprintf(arr, "%016llx", way_nodes[i]);
          arr += 16;
        }
        sprintf(arr, ")");

        if (sql_query(sql) != 0)
        {
          printf("Error inserting way\n");
        }
        free(sql);
      }
      topelement = 0;
    }
    //RELATION
    else if ((strcmp(name, "relation") == 0) && (topelement == ELM_RELATION))
    {
      if (rel_members_cnt>=1)
      {
        char *sql = (char *)malloc(rel_members_cnt*20+100);
        sprintf(sql, "INSERT INTO osm_relations (id, members) VALUES ('%lld', 0x", rel_id);
        int i = (int)strlen(sql);
        char *arr = sql+i;
        for (i=0; i<rel_members_cnt; i++)
        {
          sprintf(arr, "%02x", rel_members[i].type);
          arr += 2;
          sprintf(arr, "%02x", rel_members[i].role);
          arr += 2;
          sprintf(arr, "%016llx", rel_members[i].id);
          arr += 16;
        }
        sprintf(arr, ")");

        if (sql_query(sql) != 0)
        {
          printf("Error inserting relation\n");
        }
        free(sql);
      }
      topelement = 0;
    }
  }
}

int process_osm(char *osmname)
{
  FILE *f;
  long long progress_total;
  long long progress_cur;
  long long progress = -1;
  XML_Parser parser;
  int done;
  int depth = 0;
  int len;
  long long prg;

  if ((f = fopen(osmname, "r")) == NULL)
  {
    printf("Error opening file!\n");
    return -1;
  }

  //get size of file
  fseeko (f , 0 , SEEK_END);
  progress_total = ftello(f);
  progress_cur = 0;
  fseeko (f , 0 , SEEK_SET);

  //open XML parser
  parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser, &depth);
  XML_SetElementHandler(parser, startElement, endElement);

  printf("Import data from %s\n", osmname);

  //iterate items
  do
  {
    void *buf = XML_GetBuffer(parser, BUFF_SIZE);
    if (buf == NULL)
    {
      printf("Error alloc buffer\n");
      return -1;
    }
    len = (int)fread(buf, 1, BUFF_SIZE, f);
    //printf("LEN: %d\n", len);
    done = (int)(len < BUFF_SIZE);
    if (! XML_ParseBuffer(parser, len, done))
    {
      fprintf(stderr,
              "%s at line %" XML_FMT_INT_MOD "u\n",
              XML_ErrorString(XML_GetErrorCode(parser)),
              XML_GetCurrentLineNumber(parser));
      return -1;
    }
    progress_cur += len;
    prg = 100 * (progress_cur+1) / progress_total;
    if (prg!=progress)
    {
      progress = prg;
      printf("Import data %lld%%\r",progress);
      fflush(stdout);
    }
  } while (!done);

  printf("\n");

  //free
  XML_ParserFree(parser);
  fclose(f);

  return 0;
}

int print_help()
{
  printf("MC Map Creator\n");
  printf("Usage:\n");
  printf("  osm2mcmap <arguments> <osmfile.osm>\n");
  printf("Arguments:\n");
  printf("  -mo file.mcm   output map to file\n");
  printf("  -bt <0|1>      boundary type to save to boundary file\n");
  printf("                   0=country\n");
  printf("                   1=city\n");
  printf("  -bo file.mcb   output boundary data to file\n");
  printf("  -bi file.mcb   input boundary data from file\n");
  printf("  -l  file.log   ouput log file\n");
  printf("  -p  file.poly  load polygon of area to convert from file,\n");
  printf("                 it is not used for cropping, but will be used\n");
  printf("                 for sea generation from coastlines\n");
  printf("  -v N           verbose log level\n");
  printf("                   0 - none\n");
  printf("                   1 - error\n");
  printf("                   2 - warning\n");
  printf("                   3 - info\n");
  printf("                   4 - debug\n");
  printf("  -rcl           use residential zone for city limit\n");
  printf("  -c N           run only convert step N, if not given, all steps are processed\n");
  printf("                 can be given more at once, for example -c 1 -c 2\n");
  printf("                 must be run in this order!\n");
  printf("                 possible values are:\n");
  printf("                   0 - skip all convert steps\n");
  printf("                   1 - process boundaries\n");
  printf("                   2 - process city limit (can be skipped if not needed)\n");
  printf("                   3 - process cities\n");
  printf("                   4 - process relations\n");
  printf("                   5 - process ways\n");
  printf("                   6 - process nodes\n");
  printf("                   7 - process coastlines\n");
  printf("                   8 - process duplicate text\n");
  printf("  -h             display this help\n");
  printf("\n");
  return 0;
}


#define CONV_ALL            0xff
#define CONV_BOUNDARIES     0x01
#define CONV_CITYLIMIT      0X02
#define CONV_CITIES         0X04
#define CONV_RELATIONS      0X08
#define CONV_WAYS           0X10
#define CONV_NODES          0X20
#define CONV_COASTLINES     0x40
#define CONV_DUPLICATETEXT  0X80


int main(int argc, char *argv[])
{
  int infiles = 0;
  int convert = 0xff;
  char *boundaryinput = NULL;
  char *boundaryoutput = NULL;
  boundary_type boundarytype = BT_ERROR;
  map_multipolygons *mp = NULL;

  conf_init(argc, argv);

  if (sql_connect("localhost", "osm", "osm", "osm") != 0)
  {
    return -1;
  }

  way_nodes = (long long*)malloc(sizeof(long long) * MAXWAYNODES);
  rel_members = (relation_member*)malloc(sizeof(relation_member) * MAXRELMEMBERS);

  //parse arguments
  int i = 1;
  while (i<argc)
  {
    if (strcmp(argv[i], "-h") == 0)
    {
      print_help();
      conf_deinit();
      return 0;
    }
    else if (strcmp(argv[i], "-c") == 0)
    {
      i++;
      if (convert == CONV_ALL)
        convert = 0;
      if (i<argc)
      {
        int step = atoi(argv[i]);
        if (step == 0)
          convert = 0;
        else
          convert |= 1 << (step-1);
      }
    }
    else if (strcmp(argv[i], "-mo") == 0)
    {
      i++;
      if ((i<argc) && (curmapname == NULL))
      {
        curmapname = argv[i];
      }
    }
    else if (strcmp(argv[i], "-bt") == 0)
    {
      i++;
      if (i<argc)
      {
        boundarytype = (boundary_type)atoi(argv[i]);
      }
    }
    else if (strcmp(argv[i], "-bo") == 0)
    {
      i++;
      if ((i<argc) && (boundaryoutput == NULL))
      {
        boundaryoutput = argv[i];
      }
    }
    else if (strcmp(argv[i], "-bi") == 0)
    {
      i++;
      if ((i<argc) && (boundaryoutput == NULL))
      {
        boundaryinput = argv[i];
      }
    }
    else if (strcmp(argv[i], "-l") == 0)
    {
      i++;
      if ((i<argc) && (logname == NULL))
      {
        logname = argv[i];
        FILE *f = fopen(logname, "r");
        if (f)
        {
          fclose(f);
          #if defined(WIN32)
          _unlink(logname);
          #else
          unlink(logname);
          #endif
        }
      }
    }
    else if (strcmp(argv[i], "-p") == 0)
    {
      i++;
      if ((i<argc) && (mp == NULL))
      {
        mp = load_multipolygons(argv[i]);
      }
    }
    else if (strcmp(argv[i], "-v") == 0)
    {
      i++;
      if (i<argc)
      {
        logwhat = (logtype)atoi(argv[i]);
      }
    }
    else if (strcmp(argv[i], "-rcl") == 0)
    {
      config->residential_citylimit = 1;
    }
    else
    {
      if (infiles == 0)
      {
#ifdef _WIN32
        if (sql_init("import.sql") != 0)
#else
        char s[200];
        snprintf(s, sizeof(s), "%s/mcnavi/osm2mcm/import.sql", PACKAGE_DATA_DIR);
        if (sql_init(s) != 0)
#endif
        {
          printf("Cannot init database schema\n");
          sql_close();
          return -1;
        }
      }

      if (process_osm(argv[i])==0)
      {
        infiles++;
      }
      else
      {
        printf("Error parsing file %s\n", argv[i]);
        conf_deinit();
        return -1;
      }
    }
    i++;
  }

  /*if (infiles==0)
  {
    print_help();
    conf_deinit();
    free(way_nodes);
    free(rel_members);
    return -1;
  }*/

  /*if ((boundaryoutput==NULL)&&(curmapname==NULL)&&(convert==0xff))
  {
    print_help();
    conf_deinit();
    free(way_nodes);
    free(rel_members);
    return -1;
  }*/

  if (convert & 0x01)
  {
    #ifdef _WIN32
    if (sql_init("convert.sql") != 0)
    #else
    char s[200];
    snprintf(s, sizeof(s), "%s/mcnavi/osm2mcm/convert.sql", PACKAGE_DATA_DIR);
    if (sql_init(s) != 0)
    #endif
    {
      printf("Cannot init database schema\n");
      sql_close();
      free(way_nodes);
      free(rel_members);
      return -1;
    }
  }

  /*long latx, lonx;
  int r = wgs84_cross(0, 100, 100, 100, 
                      50, 0, 50, 50,
                      &latx, &lonx);
  printf("%d %ld,%ld\n", r, latx, lonx);
  return 0;*/
  
  if (boundaryinput)
  {
    map_loadboundaries(boundaryinput);
  }

  if (convert & CONV_BOUNDARIES)
  {
    map_process_boundaries();
  }
  if (convert & CONV_CITYLIMIT)
  {
      map_process_citylimit();
  }

  if (boundaryoutput)
  {
    map_saveboundary(boundaryoutput, boundarytype);
  }

  map_remove_planes(PT_SEA);

  if (convert & CONV_CITIES)
  {
    map_process_cities();
  }
  if (convert & CONV_RELATIONS)
  {
    map_process_relations();
  }
  if (convert & CONV_WAYS)
  {
    map_process_ways();
  }
  if (convert & CONV_NODES)
  {
    map_process_nodes();
  }
  if (convert & CONV_COASTLINES)
  {
    map_process_coastlines(mp);
  }
  if (convert & CONV_DUPLICATETEXT)
  {
    map_fix_duplicate_text();
  }

  if (curmapname)
  {
    //map_reindex();
    map_write();
  }

  sql_close();
  conf_deinit();
  free(way_nodes);
  free(rel_members);
  map_freemultipolygons(mp);
    
  printf("Done\n");

  return 0;
}
