/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * map_coastlines.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 *
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "map_common.h"
#include "map_coastlines.h"
#include "polygon.h"
#include "projection.h"
#include "map_convert.h"
#include "db_mysql.h"
#include "map_convert.h"

#define MAXSEASIZE 200000


int merge_coastlines(long start1, long stop1, map_polygon *polygon1,
                     long start2, long stop2, map_polygon *polygon2,
                     long *start, long *stop, map_polygon **polygon)
{
  map_polygon *first, *second;
  
  if (start1 == stop2)
  {
    first = polygon2;
    second = polygon1;
    *start = start2;
    *stop = stop1;
  }
  else if (stop1 == start2)
  {
    first = polygon1;
    second = polygon2;
    *start = start1;
    *stop = stop2;
  }
  else
  {
    printf("Something wrong in coastline append !!! Gone...\n"); 
    return -1;
  }

  map_polygon *p = (map_polygon *)malloc(sizeof(map_polygon) + (polygon1->count + polygon2->count - 1) * sizeof(map_position));
  p->count = polygon1->count + polygon2->count - 1;
  for (int i=0; i<first->count; i++)
  {
    p->points[i] = first->points[i];
  }
  for (int i=1; i<second->count; i++)
  {
    p->points[i+first->count-1] = second->points[i];
  }
  *polygon = p;

  return 0;
}

int update_coastline(long id, long start, long stop, map_polygon *polygon)
{
  char *s = polygon2ascii (polygon);
  if (s==NULL)
    return -1;
  char *sql = (char*)malloc(strlen(s) + 200);
  sprintf(sql, "UPDATE mcm_coastlines SET start=%ld, stop=%ld, polygon=%s  WHERE id=%ld", start, stop, s, id);
  //printf("%s\n", sql);
  int res = sql_query(sql);
  free(sql);
  free(s);
  return res;
}

int delete_coastline(long id)
{
  char sql[100];
  sprintf(sql, "DELETE FROM mcm_coastlines WHERE id=%ld", id);
  int res = sql_query(sql);
  return res;
}

int save_coastline(osm_way *way)
{
  int res = -1;
  map_polygon *polygon;

  if (way2polygon(way, &polygon) == 0)
  {
    //check if we find existing coastline to extend
    //coast may not be reversed - it must connect start and stop nodes, but not start to start or end to end
    char sql[100];
    sprintf(sql, "SELECT id, start, stop, polygon FROM mcm_coastlines WHERE (start=%ld) OR (stop=%ld)",
            way->nodes[way->count-1], way->nodes[0]);  
    if (sql_query(sql) != 0)
    {
      free(polygon);
      return -1;
    }

    sql_result *result;
    if ((result = sql_storeresult()) == NULL)
    {
      free(polygon);
      return -1;
    }
    sql_row row;
    long start, stop;
    if (row = sql_fetch_row(result))
    {
      //append way to existing coastline
      long id = sql_record_long(row, 0);
      start = sql_record_long(row, 1);
      stop = sql_record_long(row, 2);
      unsigned char *p = (unsigned char*)sql_record_str (row, 3);

      map_polygon *append = ascii2polygon (p);
      if (append == NULL)
      {
        free(polygon);
        sql_freeresult(result);        
        return -1;
      }

      long newstart, newstop;
      map_polygon *newpolygon;
      if (merge_coastlines(way->nodes[0], way->nodes[way->count-1], polygon, 
                           start, stop, append,
                           &newstart, &newstop, &newpolygon) == 0)
      {
        update_coastline (id, newstart, newstop, newpolygon);
        free(newpolygon);
      }

      free(append);
    }
    else
    {
      //create new coastline
      char *s = polygon2ascii(polygon);
      if (s)
      {
        char *sql = (char*)malloc(strlen(s) + 200);
        sprintf(sql, "INSERT INTO mcm_coastlines (start, stop, polygon) VALUES (%ld, %ld, %s)",
                way->nodes[0], way->nodes[way->count-1], s);
        res = sql_query(sql);
        free(sql);
        free(s);
      }
    }
    sql_freeresult(result);
    free(polygon);
  }

  return res;
}

//cut outer closed "polygon" with the other closed or unclosed polygon "cut"
//cut polygon assumes, that it fills to the right, so cut off left portion
typedef struct cutcrosspoint
{
  int id;
  map_position pos;
  int pindex;       //index of second point in abscisse
  int cindex;       
  int pdist;        //distance from pindex - eg. backwards !!!
  int cdist;
  int used;
  cutcrosspoint *next;
} cutcrosspoint;

int swapcutcrosspoint(cutcrosspoint *cp1, cutcrosspoint *cp2)
{
  int tmp = cp1->pindex;
  cp1->pindex = cp2->pindex;
  cp2->pindex = tmp;
  tmp = cp1->cindex;
  cp1->cindex = cp2->cindex;
  cp2->cindex = tmp;
  map_position pos = cp1->pos;
  cp1->pos = cp2->pos;
  cp2->pos = pos;
  return 0;
}

int iscutcrosspoint(cutcrosspoint *cp, int index)
{
  cutcrosspoint *cpn = cp;
  while (cpn)
  {
    if (cpn->cindex == index)
    {
      cpn->used = 1;
      return 1;
    }
    cpn = cpn->next;
  }
  return 0;
}

int ispolycrosspoint(cutcrosspoint *cp, int index)
{
  cutcrosspoint *cpn = cp;
  while (cpn)
  {
    if (cpn->pindex == index)
    {
      cpn->used = 1;
      return 1;
    }
    cpn = cpn->next;
  }
  return 0;
}

//return next crosspoint following the index
cutcrosspoint *nextpolycrosspoint(cutcrosspoint *cross, cutcrosspoint *cp)
{ //poly is always closed
  cutcrosspoint *res = NULL;
  cutcrosspoint *cpn = cross;
  while (cpn)
  {
    //if ((cpn == cp) || (cpn->used == 1))
    if (cpn == cp)
    {
    }
    else if (res == NULL)
    {
      res = cpn;
    }
    else if ((res->pindex<cp->pindex) || ((res->pindex==cp->pindex) && (res->pdist>cp->pdist)))
    {
      if (((cpn->pindex == cp->pindex) && (cpn->pdist<cp->pdist)) ||
          (cpn->pindex > cp->pindex) ||
          (cpn->pindex < res->pindex) ||
          ((cpn->pindex == res->pindex) && (cpn->pdist > res->pdist)))
      {
        res = cpn;
      }
    }
    else if ((cpn->pindex>cp->pindex) || ((cpn->pindex == cp->pindex) && (cpn->pdist<cp->pdist)))    
    {
      if ((cpn->pindex<res->pindex) || ((cpn->pindex == res->pindex) && (cpn->pdist>res->pdist)))
      {
        res = cpn;
      }
    }
    cpn = cpn->next;
  }
  return res;
}

cutcrosspoint *nextcutcrosspoint(cutcrosspoint *cross, cutcrosspoint *cp, int closed)
{
  cutcrosspoint *res = NULL;
  cutcrosspoint *cpn = cross;
  while (cpn)
  {
    //if ((cpn == cp) || (cpn->used == 1))
    if (cpn == cp)
    {
    }
    else if (res == NULL)
    {
      if (closed || (cpn->cindex>cp->cindex) || ((cpn->cindex==cp->cindex) && (cpn->cdist<cp->cdist)))
        res = cpn;
    }
    else if ((res->cindex<cp->cindex) || ((res->cindex==cp->cindex) && (res->cdist>cp->cdist)))
    {
      if (((cpn->cindex == cp->cindex) && (cpn->cdist<cp->cdist)) ||
          (cpn->cindex>cp->cindex) ||
          (cpn->cindex<res->cindex) ||
          ((cpn->cindex==res->cindex) && (cpn->cdist>res->cdist)))
      {
        res = cpn;
      }
    }
    else if ((cpn->cindex>cp->cindex) || ((cpn->cindex == cp->cindex) && (cpn->cdist<cp->cdist)))
    {
      if ((cpn->cindex<res->cindex) || ((cpn->cindex == res->cindex) && (cpn->cdist>res->cdist)))
      {
        res = cpn;
      }
    }
    cpn = cpn->next;
  }
  return res;
}

map_multipolygon *map_polygoncut(map_polygon *polygon, map_polygon *cut, int closed, int corners=0)
{
  //polygon is always closed
  //cut polygon may be closed or open
  int i;
  int counter;
  int stop = 0;
  map_polygon *polycw;
  int found = 0;

  //printf("Cut poly with %d points with poly with %d points\n", polygon->count, cut->count);
  if ((polygon == NULL) || (cut == NULL))
    return NULL;

  //we need always clockwise ordering of polygon
  polycw = map_polygonclockwise(polygon);
  if (polycw == NULL)
  {
    map_log(LOG_DEBUG, "Bad polygon clockwise\n"); 
    return NULL;
  }
  /*if (corners==1)
  {
    for (int i=0; i<polycw->count; i++)
    {
      char cname[20];
      sprintf(cname, "P%d", i);
      save_poi(POI_STONE, polycw->points[i].lat, polycw->points[i].lon, cname);
    }
    for (int i=0; i<cut->count; i++)
    {
      char cname[20];
      sprintf(cname, "S%d", i);
      save_poi(POI_STONE, cut->points[i].lat, cut->points[i].lon, cname);
    }
  }*/
  //if (corners)
  //  printf("AAA\n");
  
  //find all crossing points of polygon and cut
  cutcrosspoint *cross = NULL;
  cutcrosspoint *cn;
  int count = 0;
  int pindex = 0;
  int pindex2 = polycw->count-1;
  while (pindex<polycw->count)
  {
    int cindex = closed ? 0 : 1;
    int cindex2 = closed ? cut->count-1 : 0;
    while (cindex<cut->count)
    {
      map_position pos;
      //if ((corners==1) && (pindex==5)&&(cindex==2870))
      //    printf("check\n");
      if (wgs84_cross (polycw->points[pindex2].lat, polycw->points[pindex2].lon,
                       polycw->points[pindex].lat, polycw->points[pindex].lon,
                       cut->points[cindex2].lat, cut->points[cindex2].lon, 
                       cut->points[cindex].lat, cut->points[cindex].lon,
                       &pos.lat, &pos.lon) == 1)
      {
        int skip = 0;
        //compute distance, use values in milimeters, because we need to be more precise than lat/lon
        int pdist = wgs84_distance_mm(pos.lat, pos.lon, polycw->points[pindex].lat, polycw->points[pindex].lon);
        int cdist = wgs84_distance_mm(pos.lat, pos.lon, cut->points[cindex].lat, cut->points[cindex].lon);
        if (cross)
        {
          //check duplicated points
          cutcrosspoint *ch = cross;
          while (ch)
          {
            if (((ch->pindex==pindex)&&(ch->pdist==pdist)) || ((ch->pos.lat==pos.lat)&&(ch->pos.lon==pos.lon)))
            {
              skip = 1;
              break;
            }
            ch = ch->next;
          }
          if (skip==0)
          {
            cn->next = (cutcrosspoint*)malloc(sizeof(cutcrosspoint));
            cn = cn->next;
          }
        }
        else
        {
          cross = (cutcrosspoint*)malloc(sizeof(cutcrosspoint));
          cn = cross;
        }

        if (skip==0)
        {
          count ++;
          cn->next = NULL;
          cn->id = count;
          cn->pos = pos;
          cn->pindex = pindex;
          cn->cindex = cindex;
          cn->pdist = pdist;
          cn->cdist = cdist;
          cn->used = 0;
          //if (count==6)
          //  printf("Cross%d P %d,%d C %d,%d [%ld,%ld]\n", count, pindex, cn->pdist, cindex, cn->cdist, pos.lat, pos.lon);
          //printf("  X %ld,%ld %ld,%ld\n", cut->points[cindex2].lat, cut->points[cindex2].lon, cut->points[cindex].lat, cut->points[cindex].lon);
          //printf("  P %ld,%ld %ld,%ld\n", polycw->points[pindex2].lat, polycw->points[pindex2].lon, polycw->points[pindex].lat, polycw->points[pindex].lon);
          if (corners==1)
          {
            char cname[20];
            printf("Cross%d P %d,%d C %d,%d [%ld,%ld]\n", count, pindex, cn->pdist, cindex, cn->cdist, pos.lat, pos.lon);
            sprintf(cname, "C%d", count);
            save_poi(POI_STONE, pos.lat, pos.lon, 0, cname);
            /*sprintf(cname, "X%d-%d", count, 1);
            save_poi(POI_STONE, cut->points[cindex2].lat, cut->points[cindex2].lon, cname);
            sprintf(cname, "              X%d-%d", count, 2);
            save_poi(POI_STONE, cut->points[cindex].lat, cut->points[cindex].lon, cname);
            sprintf(cname, "P%d-%d", count, 1);
            save_poi(POI_STONE, polycw->points[pindex2].lat, polycw->points[pindex2].lon, cname);
            sprintf(cname, "              P%d-%d", count, 2);
            save_poi(POI_STONE, polycw->points[pindex].lat, polycw->points[pindex].lon, cname);*/
          }
        }
      }
      cindex2 = cindex;
      cindex++;
    }
    pindex2 = pindex;
    pindex++;
  }
  if ((cross == NULL) || (count<2))
  {
    //printf("  No cut points\n");
    map_freepolygon(polycw);
    return NULL;
  }

  if ((closed==0) && (count == 1))
  {
    map_freepolygon(polycw);
    map_log(LOG_ERROR, "Cannot cut too short not-closed polygon, start=%ld,%ld stop=%ld,%ld\n", 
            cut->points[0].lat, cut->points[0].lon, cut->points[cut->count].lat, cut->points[cut->count].lon);
    return NULL;
  }

  //if (corners)
  //  printf("AAA\n");
  //if (corners) return NULL;
  
  //create polygons
  map_multipolygon *mp = NULL;
  while (1)
  {
    //int pidx = 0;
    cutcrosspoint *cn = cross;

    //find first unused crosspoint, going inside of poly
    while (cn && (cn->used==1))
    {
      cn = cn->next;
    }
    if (cn == NULL)
      break;
    cutcrosspoint *first = NULL;
    cn->used = 1;
    //if ((cn->id == 12) && corners)
    //  printf("AAA\n");
    //printf("  New polygon\n");

    //continue with inner - if next cross point inside polygon, go with cross to next crosspoint, 
    //                      else with polygon to next crosspoint
    //repeat until we reach first crosspoint
    map_polygon *p = NULL;
    int pcnt = 0;
    int prevcut = 0;
    int ccnt = 0;
    while ((first==NULL) || (cn->id != first->id))
    {
      int nextcut = 0;
      map_position pos1 = cn->pos;
      if (ccnt > count)
      {
        map_log(LOG_ERROR, "program error: Possible polygon cut error, contact author with this information:\n");
        save_polygon (polygon, "p_polygon.poly");
        save_polygon (cut, "p_cut.poly");
        if (p)
        {
          map_freepolygon(p);
          p = NULL;
        }
        /*cn = cross;
        while (cn)
        {
          char cname[30];
          sprintf(cname, "C%d", cn->id);
          save_poi(POI_STONE, cn->pos.lat, cn->pos.lon, 0, cname);
          map_log(LOG_ERROR, "  C%d %ld,%ld\n", cn->id, cn->pos.lat, cn->pos.lon);
          cn= cn->next;
        }*/
        break;
      }
      //printf("    Cross %d\n", cn->id);
      if (first == NULL)
        first = cn;
      cutcrosspoint *ci = nextcutcrosspoint(cross, cn, closed);
      cutcrosspoint *cp = nextpolycrosspoint(cross, cn);
      if (ci == NULL)
      {
        //last cut cross point of not closed polygon, go with polycw
        nextcut = 0;
      }
      else if (p == NULL)
      {
        //first crossing point
        if (ci->cindex == cn->cindex)
        {
          if (ci->cdist < cn->cdist)
            nextcut = 1;
        }
        //else if ((cn->cdist == 0) && 
        else if ((cn->pos.lat == cut->points[cn->cindex].lat) && (cn->pos.lon == cut->points[cn->cindex].lon))
        {
          if (((cn->cindex == cut->count-1) && (ci->cindex==0)) || (ci->cindex == (cn->cindex+1)))
          {
            nextcut = 1;
          }
          else
          {
            if (cn->cindex == cut->count-1)
            {
              if (map_pointinpolygon(&cut->points[0], polycw) != 0)
                nextcut = 1;
            }
            else
            {
              if (map_pointinpolygon(&cut->points[cn->cindex+1], polycw) != 0)
                nextcut = 1;
            }
          }
        }
        else if (map_pointinpolygon(&cut->points[cn->cindex], polycw) != 0)
        {
          nextcut = 1;
        }
      }
      else
      {
        //compute angle and use the smallest one
        int a1 = wgs84_angle(p->points[p->count-1].lat, p->points[p->count-1].lon, cn->pos.lat, cn->pos.lon);
        map_position pos = cn->cdist == 0 ? (cn->cindex == cut->count-1 ? cut->points[0] : cut->points[cn->cindex+1]) : cut->points[cn->cindex];
        int a2 = wgs84_angle(cn->pos.lat, cn->pos.lon, pos.lat, pos.lon);
        pos = cn->pdist == 0 ? (cn->pindex == polycw->count-1 ? polycw->points[0] : polycw->points[cn->pindex+1]) : polycw->points[cn->pindex];
        int a3 = wgs84_angle(cn->pos.lat, cn->pos.lon, pos.lat, pos.lon);
        if (a1<0) a1 = 360+a1;
        if (a2<0) a2 = 360+a2;
        if (a3<0) a3 = 360+a3;
        a1 += 180;
        a2 = a1-a2;
        if (a2>=360) a2 -= 360;
        if (a2<0) a2 += 360;
        a3 = a1-a3;
        if (a3>=360) a3 -= 360;
        if (a3<0) a3 += 360;
        if (a2<a3)
          nextcut = 1;
      }

      if (nextcut)
      {
        map_polygon *pt = NULL;
        if ((ci->cindex>cn->cindex) || ((ci->cindex==cn->cindex) && (ci->cdist<cn->cdist)))
        {
          int d = 0;
          int cnt = ci->cindex - cn->cindex;
          if ((cnt>0) && (cut->points[cn->cindex].lat==cn->pos.lat) && (cut->points[cn->cindex].lon==cn->pos.lon))
            { cnt--; d = 1; }
          if ((cnt>0) && (cut->points[ci->cindex-1].lat==ci->pos.lat) && (cut->points[ci->cindex-1].lon==ci->pos.lon))
            { cnt--; }
          pt = map_newpolygon (pcnt+cnt+1);
          if (p)
          {
            memcpy(pt->points, p->points, pcnt*sizeof(map_position));
            map_freepolygon(p);
          }
          pt->points[pcnt] = cn->pos;
          //printf("%ld,%ld\n",pt->points[pcnt].lat, pt->points[pcnt].lon);
          for (int i=0; i<cnt; i++)
          {
            pt->points[pcnt+i+1] = cut->points[cn->cindex+d+i];
            //printf("%ld,%ld\n",pt->points[pcnt+i+1].lat, pt->points[pcnt+i+1].lon);
          }
          pcnt += cnt+1;
        }
        else
        {
          int d1 = 0;
          int d2 = 0;
          int cnt1 = cut->count - cn->cindex;
          int cnt2 = ci->cindex;
          if ((cnt2>0) && (cnt1==0) && (polycw->points[cn->cindex].lat==cn->pos.lat) && (polycw->points[cn->cindex].lon==cn->pos.lon))
            { cnt2--; d2 = 1; }
          if ((cnt2>0) && (cut->points[ci->cindex-1].lat==ci->pos.lat) && (cut->points[ci->cindex-1].lon==ci->pos.lon))
            { cnt2--; }
          if ((cnt1>0) && (cut->points[cn->cindex].lat==cn->pos.lat) && (cut->points[cn->cindex].lon==cn->pos.lon))
            { cnt1--; d1 = 1; }
          if ((cnt1>0) && (cnt2==0) && (cut->points[ci->cindex-1].lat==ci->pos.lat) && (cut->points[ci->cindex-1].lon==ci->pos.lon))
            { cnt1--; }
          pt = map_newpolygon (pcnt+cnt1+cnt2+1);
          if (pt==NULL)
            return NULL;
          if (p)
          {
            memcpy(pt->points, p->points, pcnt*sizeof(map_position));
            map_freepolygon(p);
          }

          pt->points[pcnt] = cn->pos;
          pcnt++;
          for (int i=0; i<cnt1; i++)
          {
            pt->points[pcnt+i] = cut->points[cn->cindex+d1+i];
          }
          pcnt += cnt1;
          for (int i=0; i<cnt2; i++)
          {
            pt->points[pcnt+i] = cut->points[d2+i];
          }
          pcnt += cnt2;
        }
        p = pt;
        ci->used = 1;
        cn = ci;
      }
      else
      {
        //copy polycw to next cross point
        map_polygon *pt = NULL;
        if ((cp->pindex>cn->pindex) || ((cp->pindex==cn->pindex) && (cp->pdist<cn->pdist)))
        {
          int d = 0;
          int cnt = cp->pindex - cn->pindex;
          if ((cnt>0) && (polycw->points[cn->pindex].lat==cn->pos.lat) && (polycw->points[cn->pindex].lon==cn->pos.lon))
            { cnt--; d = 1; }
          if ((cnt>0) && (polycw->points[cp->pindex-1].lat==cp->pos.lat) && (polycw->points[cp->pindex-1].lon==cp->pos.lon))
            { cnt--; }
          pt = map_newpolygon (pcnt+cnt+1);
          if (p)
          {
            memcpy(pt->points, p->points, pcnt*sizeof(map_position));
            map_freepolygon(p);
          }
          pt->points[pcnt] = cn->pos;
          for (int i=0; i<cnt; i++)
          {
            pt->points[pcnt+i+1] = polycw->points[cn->pindex+d+i];
          }
          pcnt += cnt+1;
        }
        else
        {
          int d1 = 0;
          int d2 = 0;
          int cnt1 = polycw->count - cn->pindex;
          int cnt2 = cp->pindex;
          if ((cnt2>0) && (cnt1==0) && (polycw->points[cn->pindex].lat==cn->pos.lat) && (polycw->points[cn->pindex].lon==cn->pos.lon))
            { cnt2--; d2 = 1; }
          if ((cnt2>0) && (polycw->points[cp->pindex-1].lat==cp->pos.lat) && (polycw->points[cp->pindex-1].lon==cp->pos.lon))
            { cnt2--; }
          if ((cnt1>0) && (polycw->points[cn->pindex].lat==cn->pos.lat) && (polycw->points[cn->pindex].lon==cn->pos.lon))
            { cnt1--; d1 = 1; }
          if ((cnt1>0) && (cnt2==0) && (polycw->points[cp->pindex-1].lat==cn->pos.lat) && (polycw->points[cp->pindex-1].lon==cn->pos.lon))
            { cnt1--; }
          pt = map_newpolygon (pcnt+cnt1+cnt2+1);
          if (p)
          {
            memcpy(pt->points, p->points, pcnt*sizeof(map_position));
            map_freepolygon(p);
          }

          pt->points[pcnt] = cn->pos;
          //printf("%ld,%ld\n",pt->points[pcnt].lat, pt->points[pcnt].lon);
          pcnt++;
          for (int i=0; i<cnt1; i++)
          {
            pt->points[pcnt+i] = polycw->points[cn->pindex+d1+i];
            //printf("%ld,%ld\n",pt->points[pcnt+i].lat, pt->points[pcnt+i].lon);
          }
          pcnt += cnt1;
          for (int i=0; i<cnt2; i++)
          {
            pt->points[pcnt+i] = polycw->points[d2+i];
            //printf("%ld,%ld\n",pt->points[pcnt+i].lat, pt->points[pcnt+i].lon);
          }
          pcnt += cnt2;
        }
        p = pt;
        cp->used = 1;
        cn = cp;
      }
      prevcut = nextcut;
      ccnt++;
    }

    //append to multipolygon
    if (p)
    {
      int mpcnt = mp ? mp->count + 1 : 1;
      map_multipolygon *mpt = map_newmultipolygon (mpcnt);
      if (mp)
      {
        for (int i=0; i<mp->count; i++)
        {
          mpt->polygon[i] = mp->polygon[i];
        }
        free(mp);
      }
      mpt->polygon[mpcnt-1] = p;
      mp = mpt;
    }
  }

  //free cross
  while (cross)
  {
    cn = cross;
    cross = cross->next;
    free(cn);
  }
  
  return mp;
}

//cut multipolygon to the cut polygon
//assume cut is inside of mp
map_multipolygons *map_multipolygoncut(map_multipolygon *mp, map_polygon *cut)
{
  if ((mp == NULL) || (cut == NULL))
    return NULL;

  
  //check outer polygon and cut it if needed
  //mpc will be be array of outer polygons
  map_multipolygon *mpc = NULL;
  if (map_polygoninpolygon (cut, mp->polygon[0]))
  {
    mpc = map_newmultipolygon(1);
    mpc->polygon[0] = map_duplicatepolygon (cut);
  }
  else
  {
    mpc = map_polygoncut(cut, mp->polygon[0], 1);
  }

  if (mpc == NULL)
    return NULL;

  /*map_multipolygon *mpt = map_newmultipolygon (1);
  mpt->polygon[0] = mpc->polygon[0];
  save_plane(PT_FOREST, mpt);
  free(mpt);
  return NULL;*/
    
  /*map_multipolygons *mpx = map_newmultipolygons(mpc->count);
  for (int mi=0; mi<mpc->count; mi++)
  {
    mpx->mp[mi] = map_newmultipolygon(1);
    mpx->mp[mi]->polygon[0] = mpc->polygon[mi];
  }
  free(mpc);
  return mpx;*/

  //if we are inside of inner polygon, omit it
  for (int i=1; i<mp->count; i++)
  {
    for (int j=0; j<mpc->count; j++)
    {
      if (mpc->polygon[j] && map_polygoninpolygon (mpc->polygon[j], mp->polygon[i]))
      {
        map_freepolygon(mpc->polygon[j]);
        mpc->polygon[j] = NULL;
        break;
      }
    }
  }

  //cut it with inner polygons too
  for (int i=1; i<mp->count; i++)
  {
    //cut...
    //map_log(LOG_DEBUG, "Poly %d\n", mp->polygon[i]->count);
    map_multipolygons *mpo = map_newmultipolygons (mpc->count);
    int count = 0;
    for (int j=0; j<mpc->count; j++)
    {
      if (mpc->polygon[j] == NULL)
      {
        mpo->mp[j] = NULL;
      }
      else if (map_polygoncrosspolygon(mp->polygon[i], mpc->polygon[j]))
      {
        mpo->mp[j] = map_polygoncut(mpc->polygon[j], mp->polygon[i], 1);
        if (mpo->mp[j])
          count += mpo->mp[j]->count;
        else
          count++;
      }
      else
      {
        mpo->mp[j] = NULL;
        count++;
      }
    }
    //... and join to one array
    map_multipolygon *mpc2 = map_newmultipolygon(count);
    count = 0;
    for (int j=0; j<mpc->count; j++)
    {
      if (mpo->mp[j])
      {
        for (int k=0; k<mpo->mp[j]->count; k++)
        {
          mpc2->polygon[count] = mpo->mp[j]->polygon[k];
          count++;
        }
        map_freepolygon(mpc->polygon[j]);
      }
      else if (mpc->polygon[j])
      {
        mpc2->polygon[count] = mpc->polygon[j];
        count++;
      }
    }
    free(mpc);
    mpc = mpc2;
  }

  //assign inner polygons
  map_multipolygons *mps = map_newmultipolygons(mpc->count);
  for (int i=0; i<mpc->count; i++)
  {
    mps->mp[i] = map_newmultipolygon(1);
    mps->mp[i]->polygon[0] = mpc->polygon[i];
    for (int j=1; j<mp->count; j++)
    {
      if (map_polygoninpolygon(mp->polygon[j], mps->mp[i]->polygon[0]))
      {
        map_multipolygon *mpn = map_newmultipolygon (mps->mp[i]->count+1);
        for (int k=0; k<mps->mp[i]->count; k++)
          mpn->polygon[k] = mps->mp[i]->polygon[k];
        mpn->polygon[mps->mp[i]->count] = map_duplicatepolygon (mp->polygon[j]);
        free(mps->mp[i]);
        mps->mp[i] = mpn;
      }
    }
  }
  free(mpc);
  
  return mps;
}

//find islands and insert it to sea multipolygons
map_multipolygon *update_islands(map_polygon *sea, int areaindex, int areacount)
{
  char sql[200];
  map_multipolygon *seares = NULL;

  if (sea == NULL)
    return NULL;

  if (sql_query("SELECT id, start, stop, polygon FROM mcm_coastlines WHERE `start`=`stop`") != 0)
  {
    return NULL;
  }

  sql_result *result;
  if ((result = sql_storeresult()) == NULL)
  {
    return NULL;
  }
  long rowcnt = sql_numrows (result);
  long rowindex = 0;

  seares = map_newmultipolygon(1);
  seares->polygon[0] = sea;

  sql_row row;
  while (row = sql_fetch_row(result))
  {
    long id = sql_record_long(row, 0);
    long start = sql_record_long(row, 1);
    long stop = sql_record_long(row, 2);
    unsigned char *p = (unsigned char*)sql_record_str (row, 3);

    map_polygon *polygon = ascii2polygon (p);
    if (polygon)
    {
      if (map_polygoninpolygon (polygon, sea))
      {
        //append island as polygon to multipolygon
        map_multipolygon *mp = map_newmultipolygon(seares->count+1);
        for (int k=0; k<seares->count; k++)
        {
          mp->polygon[k] = seares->polygon[k];
        }
        mp->polygon[seares->count] = polygon;
        free(seares);
        seares = mp;
      }
      else
      {
        map_freepolygon(polygon);
      }
    }
    rowindex++;
    long proc = 100 * areaindex / areacount + (50 + (50 * rowindex) / rowcnt) / areacount;
    printf("Create coastlines %ld%%\r", proc);
    fflush(stdout);
  }
  sql_freeresult(result);

  return seares;
}

map_multipolygon *cut_sea(map_polygon *area, map_multipolygon *sea, int maxrow, int level, int areaindex, int areacount)
{
  char sql[200];
  if (maxrow>0)
  {
    sprintf(sql, "SELECT id, start, stop, polygon FROM mcm_coastlines LIMIT 0,%d", maxrow-1);
  }
  else
  {
    sprintf(sql, "SELECT id, start, stop, polygon FROM mcm_coastlines");
  }
  if (sql_query(sql) != 0)
  {
    return NULL;
  }

  sql_result *result;
  if ((result = sql_storeresult()) == NULL)
  {
    return NULL;
  }

  long rowcnt = sql_numrows (result);
  long rowindex = 0;
  long lastproc = -1;

  //create multipolygon from outer polygon and cut it with coastlines
  sql_row row;
  while (row = sql_fetch_row(result))
  {
    //get coastline
    long id = sql_record_long(row, 0);
    long start = sql_record_long(row, 1);
    long stop = sql_record_long(row, 2);
    unsigned char *p = (unsigned char*)sql_record_str (row, 3);
    int closed = (start==stop) ? 1 : 0;

    /*if (id == 1190)
    {
      unsigned long *lengths = sql_fetch_lengths(result);
      printf("AAA %ld\n", lengths[3]);
    }*/

    map_polygon *polygon = ascii2polygon (p);
    if (polygon == NULL)
    {
      map_freemultipolygon (sea);
      sql_freeresult(result);        
      return NULL;
    }
    if (closed)
    {
      polygon->count--;
    }
    if (map_polygonselfcross(polygon))
    {
      map_log(LOG_ERROR, "Warning: Coastline polygon %ld self cross! This may lead to errors in processing!\n", id);
    }

    if (closed==0)
      map_log(LOG_WARNING, "Coastline not closed, start=%ld,%ld, stop=%ld,%ld\n", 
              polygon->points[0].lat, polygon->points[0].lon, 
              polygon->points[polygon->count-1].lat, polygon->points[polygon->count-1].lon);

    /*if (id == 1190)
    {
      for (int i=0; i<sea->count; i++)
      {
        char f[100];
        sprintf(f, "%d.poly", i);
        save_polygon(sea->polygon[i], f);
      }
      save_polygon(polygon, "polygon");
    }*/

    //cut sea parts with coastline
    map_multipolygons *seas = map_newmultipolygons (sea->count);
    int cnt=0;
    int cutted = 0;
    for (int i=0; i<sea->count; i++)
    {
      //cut already cutted sea
      seas->mp[i] = map_polygoncut(sea->polygon[i], polygon, closed);
      if (seas->mp[i])
      {
        //save_plane(PT_SEA, seas->mp[i]);
        cnt += seas->mp[i]->count;
        cutted++;
      }
      else
      {
        cnt++;
      }
    }

    //if no parts cutted, try to cut directly area, because it may create another part
    //TODO what if one coastline cuts already cutted area and also another area part?
    //TODO what if this create duplicate part?
    if ((cutted == 0) && (maxrow==-1))
    {
      map_multipolygon *np = map_polygoncut(area, polygon, closed);
      if (np)
      {
        //save_plane(PT_SEA, seas->mp[i]);
        printf("\nLevel %d %d\n", level+1, rowindex);
        seas->mp[0] = cut_sea(area, np, rowindex, level+1, -1, -1);
        if (seas->mp[0])
          cnt += seas->mp[0]->count - 1;
        printf("\nLevel %d\n", level);
      }      
    }
            
    //create new sea from cutted parts - append to existing parts
    map_multipolygon *newsea = map_newmultipolygon (cnt);
    cnt = 0;
    for (int i=0; i<sea->count; i++)
    {
      if (seas->mp[i])
      {
        for (int j=0; j<seas->mp[i]->count; j++)
        {
          newsea->polygon[cnt] = seas->mp[i]->polygon[j];
          //printf("S %d\n", newsea->polygon[cnt]->count);
          cnt++;
        }
        free(seas->mp[i]);
        free(sea->polygon[i]);
      }
      else
      {
        newsea->polygon[cnt] = sea->polygon[i];
        cnt++;
      }
    }
    free(sea);
    sea = newsea;
    free(seas);

    rowindex++;
    if (areaindex>=0)
    {
      long proc = 100 * areaindex / areacount + 50 * rowindex / rowcnt / areacount;
      if (lastproc != proc)
      {
        printf("Create coastlines %ld%%\r", proc);
        fflush(stdout);
        lastproc = proc;
      }
    }
    else
    {
      long proc = 1000 * rowindex / rowcnt;
      if (lastproc != proc)
      {
        printf("  Cut subsea %ld%%\r", proc);
        fflush(stdout);
        lastproc = proc;
      }
    }
  }
  return sea;
}

//cut area multipolygon with coastlines and create array of large seas
map_multipolygons *create_sea(map_multipolygon *area, int areaindex, int areacount)
{
  if (area==NULL)
    return NULL;

  if (map_polygonselfcross(area->polygon[0]))
  {
    map_log(LOG_ERROR, "Warning: Sea area %d self cross! This may lead to errors in processing!", areaindex);
  }

  //set clockwise ordering of outer polygon and counterclockwise of inner polygons
  map_multipolygon *mp = map_newmultipolygon (area->count);
  mp->polygon[0] = map_polygonclockwise (area->polygon[0], 0);
  for (int i=1; i<mp->count; i++)
  {
    mp->polygon[i] = map_polygonclockwise (area->polygon[i], 1);
  }

  //create multipolygon from outer polygon and cut it with coastlines
  map_multipolygon *sea = map_newmultipolygon (1);
  sea->polygon[0] = map_duplicatepolygon(mp->polygon[0]);
  sea = cut_sea(mp->polygon[0], sea, -1, 0, areaindex, areacount);
  
  //append islands
  map_multipolygons *mps = map_newmultipolygons (sea->count);
  for (int i=0; i<sea->count; i++)
  {
    mps->mp[i] = update_islands(sea->polygon[i], areaindex, areacount);
  }

  return mps;
}

//cut sea to smaller pieces and insert to map
int save_sea(map_multipolygon *sea, int areaindex, int areacount, int seaindex, int seacount)
{
  if (sea == NULL)
    return -1;
  map_log(LOG_DEBUG, "Save sea\n");
  
  long lat1, lon1, lat2, lon2;
  if (map_polygonarea(sea->polygon[0], &lat1, &lon1, &lat2, &lon2) == 0)
  {
    long zlat = lat1<0 ? 1 : 0;
    long zlon = lon1<0 ? 1 : 0;
    long latc = (lat2-lat1) / MAXSEASIZE + 1;
    long lonc = (lon2-lon1) / MAXSEASIZE + 1;
    if ((latc>1) || (lonc>1))
    {
      //cut sea to smaller pieces
      map_log(LOG_DEBUG, "Cut sea to smaller pieces\n");
      for (int lati=0; lati<latc; lati++)
      {
        for (int loni=0; loni<lonc; loni++)
        {
          map_log(LOG_DEBUG, "Cut part %d,%d\n", lati, loni);
          map_polygon *cut = map_newpolygon(4);
          cut->points[0].lat = lat1 + lati * (lat2-lat1) / latc;
          cut->points[0].lon = lon1 + loni * (lon2-lon1) / lonc;
          cut->points[1].lat = lat1 + (lati+1) * (lat2-lat1) / latc;
          cut->points[1].lon = lon1 + loni * (lon2-lon1) / lonc;
          cut->points[2].lat = lat1 + (lati+1) * (lat2-lat1) / latc;
          cut->points[2].lon = lon1 + (loni+1) * (lon2-lon1) / lonc;
          cut->points[3].lat = lat1 + lati * (lat2-lat1) / latc;
          cut->points[3].lon = lon1 + (loni+1) * (lon2-lon1) / lonc;
          /*char cname[50];
          sprintf(cname, "S %d,%d", lati, loni);
          save_poi(POI_STONE, lat1 + lati * (lat2-lat1) / latc + (lat2-lat1) / latc / 2, lon1 + loni * (lon2-lon1) / lonc + (lon2-lon1) / lonc / 2, 0, cname);*/
          /*map_multipolygon *mpt = map_newmultipolygon (1);
          mpt->polygon[0] = cut;
          save_plane(PT_SEA, mpt);
          free(mpt);*/
          map_multipolygons *cutsea = map_multipolygoncut(sea, cut);
          if (cutsea)
          {
            for (int i=0; i<cutsea->count; i++)
            {
              save_plane(PT_SEA, cutsea->mp[i]);
            }
            map_freemultipolygons(cutsea);
          }
          map_freepolygon(cut);
          long proc = 100 * areaindex / areacount + 100 * seaindex / seacount / areacount + 100 * (loni + lati * lonc + 1) / latc / lonc / seacount / areacount;
          printf("Split coastlines %ld%%\r", proc);
          fflush(stdout);
        }
      }
      //save_plane(PT_SEA, sea->mp[i]);
    }
    else
    {
      save_plane(PT_SEA, sea);
      long proc = 100 * areaindex / areacount + 100 * (seaindex+1) / seacount / areacount;
      printf("Split coastlines %ld%%\r", proc);
      fflush(stdout);
    }
  }
  return 0;
}

//generate plane data from coast lines, detect islands/mainland versus sea
//argument area contains multipolygon of processed area, which will be used for sea generation
//Process:
//1. merge all costlines together (map_process_coastlines)
//2. area is area of multipolygons (map_process_coastlines)
//3. for each area cut it with coastlines to 1 large sea - result is again aray of multipolygons (create_sea)
//4. each sea multipolygon cut to smaller peaces - result is again array of multipolygons (save_sea)
//5. save each small sea multipolygon as plane
int map_process_coastlines(map_multipolygons *area)
{
  long cnt = 0;
  sql_result *result;
  sql_row row;

  //test of failed polygon cut
  /*map_multipolygons *mp1 = load_multipolygons ("p_polygon.poly");
  map_multipolygons *mp2 = load_multipolygons ("p_cut.poly");
  map_multipolygon *mp3 = map_polygoncut(mp1->mp[0]->polygon[0], mp2->mp[0]->polygon[0], 1, 0);
  return -1;*/  

  if (area == NULL)
  {
    map_log(LOG_INFO, "No polygon area defined, will not create coastlines\n");
    return 0;
  }
  
  //detect any coastlines
  if (sql_query("SELECT COUNT(*) as cnt FROM mcm_coastlines") != 0)
  {
    return -1;
  }
  if ((result = sql_storeresult()) == NULL)
  {
    return -1;
  }
  if (row = sql_fetch_row(result))
  {
    cnt = sql_record_long(row, 0);
  }
  sql_freeresult(result);

  if (cnt == 0)
  {
    map_log(LOG_INFO, "No coastlines found, assume all is land\n");
    return 0;
  }

  //merge coastlines
  map_log(LOG_INFO, "Merge coastlines\n");
  printf("Merge coastlines 0%%\r");
  fflush(stdout);
  int merged = 0;
  while (merged == 0)
  {
    if (sql_query("SELECT c1.id, c1.start, c1.stop, c1.polygon, c2.id, c2.start, c2.stop, c2.polygon FROM mcm_coastlines c1, mcm_coastlines c2 WHERE (c1.id!=c2.id) AND (c1.start!=c1.stop) AND (c2.start!=c2.stop) AND (c1.stop=c2.start OR c2.start=c1.stop)") != 0)
    {
      return -1;
    }

    if ((result = sql_storeresult()) == NULL)
    {
      return -1;
    }
    if (row = sql_fetch_row(result))
    {
      long id1 = sql_record_long(row, 0);
      long start1 = sql_record_long(row, 1);
      long stop1 = sql_record_long(row, 2);
      unsigned char *p1 = (unsigned char*)sql_record_str (row, 3);

      long id2 = sql_record_long(row, 4);
      long start2 = sql_record_long(row, 5);
      long stop2 = sql_record_long(row, 6);
      unsigned char *p2 = (unsigned char*)sql_record_str (row, 7);

      map_polygon *polygon1 = ascii2polygon (p1);
      if (polygon1 == NULL)
      {
        sql_freeresult(result);        
        return -1;
      }
      map_polygon *polygon2 = ascii2polygon (p2);
      if (polygon2 == NULL)
      {
        free(polygon1);
        sql_freeresult(result);        
        return -1;
      }

      long newstart, newstop;
      map_polygon *newpolygon;
      if (merge_coastlines(start1, stop1, polygon1, 
                           start2, stop2, polygon2,
                           &newstart, &newstop, &newpolygon) != 0)
      {
        free(polygon1);
        free(polygon2);
        sql_freeresult(result);  
        return -1;
      }
      free(polygon1);
      free(polygon2);

      if (update_coastline (id1, newstart, newstop, newpolygon) != 0)
      {
        free(newpolygon);
        sql_freeresult(result);  
        return -1;
      }
      free(newpolygon);
        
      if (delete_coastline (id2) != 0)
      {
        sql_freeresult(result);  
        return -1;
      }
    }
    else
    {
      merged = 1;  
    }
    sql_freeresult(result);  
  }
  printf("Merge coastlines 100%%\n");

  //create sea from processed area and cut it with coastlines, that pass through it
  //assume that all is sea except islands
  map_multipolygons **seas = (map_multipolygons **)malloc(sizeof(map_multipolygons*) * area->count);
  map_log(LOG_INFO, "Create coastlines\n");
  printf("Create coastlines 0%%\r");
  fflush(stdout);
  for (int i=0; i<area->count; i++)
  {
    seas[i] = create_sea(area->mp[i], i, area->count);
    //save_plane(PT_SEA, area->mp[i]);
  }
  printf("\n");
  
  map_log(LOG_INFO, "Split coastlines\n");
  printf("Split coastlines 0%%\r");
  fflush(stdout);
  for (int i=0; i<area->count; i++)
  {
    if (seas[i])
    {
      for (int j=0; j<seas[i]->count; j++)
      {
        //save_plane(PT_SEA, sea->mp[j]);
        save_sea(seas[i]->mp[j], i, area->count, j, seas[i]->count);
      }
      map_freemultipolygons (seas[i]);
    }
  }
  printf("\n");
  free(seas);

  return 0;
}