/* -*- Mode: C; indent-tabs-mode: s; c-basic-offset: 2; tab-width: 2 -*- */
/*
 * map_convert.c
 * Copyright (C) Mike Crash 2009 <mike@mikecrash.com>
 *
 * main.c is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * main.c is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "db_mysql.h"
#include "map.h"
#include "diacritics.h"
#include "projection.h"
#include "polygon.h"
#include "map_common.h"
#include "map_text.h"
#include "map_convert.h"
#include "map_boundary.h"
#include "map_coastlines.h"

//insert dummy nodes between nodes with long distance between
//this will not be unique between different maps
//ezch next dummy node will decrease this
long dummynodeid = 0x7fffffff;

struct tag_conv
{
  const char *val;
  maptypes maptype;
};

tag_conv conv_highway[] =
{
  {"motorway",  WT_MOTORWAY },
  {"motorway_link",  WT_MOTORWAYLINK },
  {"trunk",  WT_TRUNK },
  {"trunk_link",  WT_TRUNKLINK },
  {"primary",  WT_PRIMARY },
  {"primary_link",  WT_PRIMARYLINK },
  {"secondary",  WT_SECONDARY },
  {"secondary_link",  WT_SECONDARY },
  {"tertiary",  WT_TERTIARY },
  {"unclassified",  WT_UNCLASSIFIED },
  {"residential",  WT_RESIDENTIAL },
  {"living_street",  WT_LIVING },
  {"pedestrian",  WT_LIVING },
  {"service",  WT_SERVICE },
  {"cycleway",  WT_CYCLE },
  {"bridleway",  WT_BRIDLE },
  {"track",  WT_TRACK5 },
  {"path",  WT_PATH },
  {"footway",  WT_FOOT },
  {"steps",  WT_STEPS },
  {"bus_stop",  POI_BUS },
  {"ford",  POI_FORD }
};
#define conv_highway_size sizeof(conv_highway) / sizeof(tag_conv)

tag_conv conv_route[] =
{
  {"ferry",  WT_FERRY }
};
#define conv_route_size sizeof(conv_route) / sizeof(tag_conv)

tag_conv conv_manmade[] =
{
  {"pier",  NT_PIER }
};
#define conv_manmade_size sizeof(conv_manmade) / sizeof(tag_conv)

tag_conv conv_railway[] =
{
  {"rail",  NT_RAIL },
  {"disused",  NT_RAIL },
  {"tram",  NT_TRAM },
  {"light_rail",  NT_TRAM },
  {"station",  POI_TRAIN },
  {"halt",  POI_TRAIN },
  {"tram_stop",  POI_TRAM },
  {"halt",  POI_TRAIN }
};
#define conv_railway_size sizeof(conv_railway) / sizeof(tag_conv)

tag_conv conv_waterway[] =
{
  {"stream",  NT_STREAM },
  {"river",  NT_RIVER },
  {"riverbank",  PT_WATER },
  {"weir",  NT_WEIR }
};
#define conv_waterway_size sizeof(conv_waterway) / sizeof(tag_conv)

tag_conv conv_aeroway[] =
{
  {"aerodrome",  POI_AIRPORT },
  {"runway",  NT_RUNWAY }
};
#define conv_aeroway_size sizeof(conv_aeroway) / sizeof(tag_conv)

tag_conv conv_tracktype[] =
{
  {"grade1",  WT_TRACK1 },
  {"grade2",  WT_TRACK2 },
  {"grade3",  WT_TRACK3 },
  {"grade4",  WT_TRACK4 },
  {"grade5",  WT_TRACK5 }
};
#define conv_tracktype_size sizeof(conv_tracktype) / sizeof(tag_conv)

tag_conv conv_barrier[] =
{
  {"wall",  NT_WALL },
  {"fence",  NT_FENCE }
};
#define conv_barrier_size sizeof(conv_barrier) / sizeof(tag_conv)

tag_conv conv_amenity[] =
{
  {"fuel",  POI_FUEL },
  {"pub",  POI_PUB },
  {"restaurant",  POI_RESTAURANT },
  {"fast_food",  POI_FASTFOOD },
  {"cafe",  POI_CAFE },
  {"ice_cream",  POI_ICECREAM },
  {"tee",  POI_TEE },
  {"parking",  PT_PARKING },
  {"fire_station",  POI_FIRESTATION },
  {"place_of_worship",  POI_CHURCH },
  {"tower",  POI_TOWER },
  {"fountain",  POI_FOUNTAIN },
  {"recycling",  POI_RECYCLING },
  {"hospital",  POI_HOSPITAL },
  {"doctors",  POI_DOCTORS },
  {"toilets",  POI_TOILETS },
  {"post_office",  POI_POSTOFFICE },
  {"post_box",  POI_POSTBOX },
  {"shop",  POI_SHOP },
  {"school",  POI_SCHOOL },
  {"bank",  POI_BANK },
  {"phone",  POI_PHONE },
  {"telephone",  POI_PHONE },
  {"police",  POI_POLICE },
  {"pharmacy",  POI_PHARMACY },
  {"atm",  POI_ATM },
  {"bus_station",  POI_BUS },
  {"townhall",  POI_TOWNHALL },
  {"prison",  POI_PRISON },
  {"shelter",  POI_SHELTER }
};
#define conv_amenity_size sizeof(conv_amenity) / sizeof(tag_conv)

tag_conv conv_tourism[] =
{
  {"hotel",  POI_HOTEL },
  {"motel",  POI_HOTEL },
  {"museum",  POI_MUSEUM },
  {"camp_site",  POI_CAMP },
  {"caravan_site",  POI_CAMP }
};
#define conv_tourism_size sizeof(conv_tourism) / sizeof(tag_conv)

tag_conv conv_historic[] =
{
  {"castle",  POI_HISTORIC },
  {"memorial",  POI_HISTORIC },
  {"monument",  POI_HISTORIC },
  {"ruins",  POI_HISTORIC },
  {"wayside_cross",  POI_HISTORIC },
  {"wayside_shrine",  POI_HISTORIC },
  {"wreck",  POI_HISTORIC },
  {"archaeological_site",  POI_HISTORIC },
  {"fort",  POI_HISTORIC },
  {"battlefield",  POI_HISTORIC }
};
#define conv_historic_size sizeof(conv_historic) / sizeof(tag_conv)

tag_conv conv_leisure[] =
{
  {"playground",  PT_PLAYGROUND },
  {"pitch",  PT_PITCH },
  {"golf_course",  PT_PITCH },
  {"dog_park",  PT_PITCH },
  {"park",  PT_PARK },
  {"garden",  PT_PARK },
  {"swimming_pool",  PT_SWIMMINGPOOL },
  {"water_park",  PT_WATERPARK },
  {"stadium",  PT_STADIUM },
  {"sports_centre",  PT_STADIUM }
};
#define conv_leisure_size sizeof(conv_leisure) / sizeof(tag_conv)

tag_conv conv_landuse[] =
{
  {"forest",  PT_FOREST },
  {"orchard",  PT_ORCHARD },
  {"cemetery",  PT_CEMETERY },
  {"reservoir",  PT_WATER },
  {"basin",  PT_BASIN },
  {"parking",  PT_PARKING },
  {"village_green",  PT_PARK },
  {"farm",  PT_FARM },
  {"farmyard",  PT_FARM },
  {"residential",  PT_RESIDENTIAL },
  {"industrial",  PT_INDUSTRIAL },
  {"commercial",  PT_COMMERCIAL }
};
#define conv_landuse_size sizeof(conv_landuse) / sizeof(tag_conv)

tag_conv conv_natural[] =
{
  {"wood",  PT_FOREST },
  {"scrub",  PT_FOREST },
  {"water",  PT_WATER },
  {"marsh",  PT_MARSH },
  {"wetland",  PT_MARSH },
  {"bay",  PT_WATER },
  {"beach",  PT_SAND },
  {"sand",  PT_SAND },
  {"peak",  POI_PEAK },
  {"volcano",  POI_PEAK },
  {"stone",  POI_STONE },
  {"tree",  POI_TREE }
};
#define conv_natural_size sizeof(conv_natural) / sizeof(tag_conv)

tag_conv conv_power[] =
{
  {"tower", POI_POWERTOWER },
  {"line", NT_POWERLINE },
  {"station", PT_POWERSTATION },
  {"generator", POI_POWERGENERATOR }
};
#define conv_power_size sizeof(conv_power) / sizeof(tag_conv)

tag_conv conv_place[] =
{
  {"locality", POI_LOCALITY },
  {"city", POI_CITY },
  {"town", POI_TOWN },
  {"village", POI_VILLAGE },
  {"hamlet", POI_HAMLET },
  {"suburb", POI_SUBURB }
};
#define conv_place_size sizeof(conv_place) / sizeof(tag_conv)

/*
  {"route","foot", osmtag_route, route_foot },
  {"route","cycle", osmtag_route, route_cycle },
  {"route","bicycle", osmtag_route, route_cycle },
  {"route","mtb", osmtag_route, route_mtb },
  {"toll","yes", osmtag_toll, 1 }
*/

#define convtablesize sizeof(convtable) / sizeof(tag_conv)

int tag_parse(tag_conv *convtable, int size, char *val, maptypes *maptype)
{
  int i;
  for (i=0; i<size; i++)
  {
    if (strcmp(convtable[i].val, val) == 0)
    {
      *maptype = convtable[i].maptype;
      return 0;
    }
  }
  return -1;
}

//TODO set according to country
int map_wayspeed(maptypes waytype)
{
  switch (waytype)
  {
    case WT_MOTORWAY: return 130;
    case WT_MOTORWAYLINK:  return 80;
    case WT_TRUNK: return 120;
    case WT_TRUNKLINK: return 70;
    case WT_PRIMARY: return 90;
    case WT_PRIMARYLINK: return 60;
    case WT_SECONDARY: return 75;
    case WT_TERTIARY: return 60;
    case WT_UNCLASSIFIED: return 50;
    case WT_RESIDENTIAL: return 25;
    case WT_LIVING: return 20;
    case WT_SERVICE: return 20;
    case WT_TRACK1: return 20;
    case WT_TRACK2: return 15;
    case WT_TRACK3: return 10;
    case WT_TRACK4: return 6;
    case WT_TRACK5: return 3;
    default:
      return 0;
  }
}

int map_cityspeed(maptypes waytype)
{
  switch (waytype)
  {
    case WT_MOTORWAY: 
    case WT_MOTORWAYLINK:
    case WT_TRUNK:
    case WT_TRUNKLINK:
      return 80;
    case WT_PRIMARY:
      return 50;
    case WT_PRIMARYLINK:
      return 45;
    case WT_SECONDARY:
      return 50;
    case WT_TERTIARY:
    case WT_UNCLASSIFIED:
      return 40;
    default:
      return 0;
  }
}

long boundary2city(int boundary)
{
  char sql[100];
  sprintf(sql, "SELECT id FROM mcm_city WHERE boundary='%d'", boundary);
  if (sql_query(sql) != 0)
    return 0;

  sql_result *result;
  if ((result = sql_storeresult()) == NULL)
    return 0;
  sql_row row;
  if ((row = sql_fetch_row(result)) == NULL)
  {
    sql_freeresult(result);
    return 0;
  }
  long id = sql_record_long(row, 0);
  sql_freeresult(result);
  return id;
}

long name2city(char *name, map_position *pos)
{
  //get all cities with given name
  char *sql = (char *)malloc(200);
  long name2 = gettextid(name);
  sprintf(sql, "SELECT id, lat, lon FROM mcm_city WHERE name='%ld'", name2);
  if (sql_query(sql) != 0)
  {
    free(sql);
    return 0;
  }

  sql_result *result;
  if ((result = sql_storeresult()) == NULL)
  {
    free(sql);
    return 0;
  }

  int count = sql_numrows(result);
  if (count==0)
  {
    free(sql);
    return 0;
  }

  long *ids = (long*)malloc(sizeof(long)*count);
  map_position *poss = (map_position*)malloc(sizeof(map_position)*count);

  sql_row row;
  int i=0;
  while ((row = sql_fetch_row(result)) != NULL)
  {
    ids[i] = sql_record_long(row, 0);
    poss[i].lat = sql_record_long(row, 1);
    poss[i].lon = sql_record_long(row, 2);
    i++;
  }
  sql_freeresult(result);

  //find nearest city
  long id = ids[0];
  long dist = wgs84_distance(pos->lat, pos->lon, poss[0].lat, poss[0].lon);
  for (i=1; i<count; i++)
  {
    long d = wgs84_distance(pos->lat, pos->lon, poss[i].lat, poss[i].lon);
    if (d<dist)
    {
      id = ids[i];
      dist = d;
    }
  }

  free(sql);
  free(ids);
  free(poss);
  return id;
}

char *cityname(long id)
{
  char sql[100];

  //get text id from city
  sprintf(sql, "SELECT name FROM mcm_city WHERE id='%ld'", id);
  if (sql_query(sql) != 0)
    return NULL;

  sql_result *result;
  if ((result = sql_storeresult()) == NULL)
    return NULL;

  sql_row row;
  if ((row = sql_fetch_row(result)) == NULL)
  {
    sql_freeresult(result);
    return NULL;
  }

  long tid = sql_record_long(row, 0);
  sql_freeresult(result);

  //get text
  sprintf(sql, "SELECT name FROM mcm_text WHERE id='%ld'", tid);
  if (sql_query(sql) != 0)
    return NULL;

  if ((result = sql_storeresult()) == NULL)
    return NULL;

  if ((row = sql_fetch_row(result)) == NULL)
  {
    sql_freeresult(result);
    return NULL;
  }
  
  unsigned long *lenghts = sql_fetch_lengths(result);
  char *name = (char*)malloc(lenghts[0]+1);
  strcpy(name, sql_record_str(row, 0));
  sql_freeresult(result);
  return name;
}

long get_sector(map_position *pos)
{
  char sql[200];
  long id = 0;
  long lat = MAPSECTOR(pos->lat);
  long lon = MAPSECTOR(pos->lon);

  sprintf(sql, "SELECT id FROM mcm_sectors WHERE lat=%ld AND lon=%ld", lat, lon);
  if (sql_query(sql) != 0)
    return -1;
  sql_result *result = sql_storeresult();
  if (result == NULL)
    return -1;
  sql_row row  = sql_fetch_row(result);
  if (row)
  {
    id = sql_record_long(row, 0);
  }
  sql_freeresult(result);

  if (id)
  {
    return id;
  }

  //insert sector
  sprintf(sql, "INSERT INTO mcm_sectors (lat, lon, nodes, segments) VALUES (%ld, %ld, 0, 0)", lat, lon);
  if (sql_query(sql) != 0)
    return -1;
  id = (long)sql_insertid();
  return id;
}

long create_node(long long id, map_position *pos)
{
	char sql[500];
  //update coresponding sector
  map_position nodepos;
  if (pos == NULL)
  {
    if (map_nodeposition(id, &nodepos) != 0)
    {
      return -1;
    }
  }
  else
  {
    nodepos = *pos;
  }
  long sector = get_sector(&nodepos);
  if (sector == -1)
    return -1;

  //get node attributes
  unsigned int nodeflags = 0;
  map_tags *tags = map_getnodetags(id);
  if (tags)
  {
    char *value;
    if ((value = tag_str(tags, "barrier")) != NULL)
    {
      if (strcmp(value, "gate")==0) { nodeflags |= NF_BARRIER_GATE; }
      if (strcmp(value, "lift_gate")==0) { nodeflags |= NF_BARRIER_LIFTGATE; }
      if (strcmp(value, "block")==0) { nodeflags |= NF_BARRIER_BLOCK; }
    }
    if ((value = tag_str(tags, "highway")) != NULL)
    {
      if (strcmp(value, "traffic_signals")==0) { nodeflags |= NF_SEMAPHORE; }
    }
    free_tags(tags);
  }

  //insert node
  sprintf(sql, "INSERT INTO mcm_nodes (id, lat, lon, flags, sector) VALUES (%lld, %ld, %ld, %d, %ld)", 
  				id, nodepos.lat, nodepos.lon, nodeflags, sector);
  if (sql_query(sql) != 0)
  {
    return -1;
  }

  //update sector
  sprintf(sql, "UPDATE mcm_sectors SET nodes=nodes+1 WHERE id=%ld", sector);
  if (sql_query(sql) != 0)
  {
    return -1;
  }
	return sector;
}

int update_node(long long id, map_position *pos1, long long cross, map_position *pos2, maptypes mt, int speed, unsigned int flags, long textid)
{
  char sql[500];
  long sector = 0;

  //get node
  sprintf(sql, "SELECT sector, segments, `cross` FROM mcm_nodes WHERE id=%lld", id);
  if (sql_query(sql) != 0)
    return -1;
  sql_result *result = sql_storeresult();
  if (result == NULL)
    return -1;
  sql_row row  = sql_fetch_row(result);

	char *newcross;
  int duplicated = 0;
	if (row)
  {
    //node exists, update cross
    sector = sql_record_long(row, 0);
    int segments = sql_record_int(row, 1);
    char *oldcross = sql_record_str(row, 2);

    //check for duplicated data
    char *cend = oldcross;
    for (int i=0; i<segments; i++)
    {
      long cid = strtol(cend, &cend, 10);  //TODO implement as long long with strtoll
      cend++;
      maptypes waytype = (maptypes)strtoul(cend, &cend, 10);
      cend++;
      unsigned int speed2 = (unsigned int)strtoul(cend, &cend, 10);
      cend++;
      unsigned int flags2 = (unsigned int)strtoul(cend, &cend, 10);
      cend++;
      long name = strtol(cend, &cend, 10);
      cend++;
      if ((cid==cross) && (waytype==mt))
      {
        duplicated = 1;
        map_log(LOG_WARNING, "Duplicated way segment from node %lld to %ld\n", id, cid);
        break;
      }
    }

		if (duplicated == 0)
		{
		  newcross = (char*)malloc(strlen(oldcross)+50);
      if (oldcross[0] == 0)
		    sprintf(newcross, "%lld,%d,%d,%u,%ld", cross, (int)mt, speed, flags, textid);
      else
		    sprintf(newcross, "%s,%lld,%d,%d,%u,%ld", oldcross, cross, (int)mt, speed, flags, textid);
		}
	}
	else
	{
		sector = create_node(id, pos1);
		if (sector == -1)
		{
			sql_freeresult(result);
			return -1;
		}
		newcross = (char*)malloc(50);
    sprintf(newcross, "%lld,%d,%d,%u,%ld", cross, (int)mt, speed, flags, textid);
	}
  sql_freeresult(result);
	
  if (duplicated==0)
  {
    //append next node as cross
    char *sql2 = (char*)malloc(strlen(newcross)+200);
    sprintf(sql2, "UPDATE mcm_nodes SET segments=segments+1, `cross`='%s' WHERE id=%lld", newcross, id);
    free(newcross);
    if (sql_query(sql2) != 0)
    {
      free(sql2);
      return -1;
    }
    free(sql2);

    //update segments count in sector
    sprintf(sql, "UPDATE mcm_sectors SET segments=segments+1 WHERE id=%ld", sector);
    if (sql_query(sql) != 0)
    {
      return -1;
    }
  }

  return 0;
}

//INSERT INTO DATABASE
int save_way(maptypes mt, osm_way *way, int speed, unsigned int flags, char *name, char *note)
{
  int i;
  map_position pos1, pos2;
  int ok1 = 0, ok2 = 0;
  int cl1 = 0, cl2 = 0;

  //text
  long textid = gettextid(name);
 
  //check all nodes exists and add it to database
  for (i=0; i<way->count; i++)
  {
    if (map_nodeposition(way->nodes[i], &pos2) == 0)
    {
      ok2 = 1;
      if ((speed == 0) && (mt <= WT_TERTIARY) && (config->residential_citylimit))
      {
        int inside = 0;
        boundary* bnd = map_boundaryinside(&pos2, BT_TRAFFIC_CITYLIMIT);
        if (bnd)
        {
          cl2 = 1;
          map_freeboundary(bnd);
        }
      }
    }

    if (ok1 && ok2)
    {
      //set speed - check if one of nodes is in city limit boundary
      int wayspeed = speed;
      if ((wayspeed == 0) && (cl1 || cl2))
        wayspeed = mt<=WT_TRUNKLINK ? 70 : 45;
      if (wayspeed == 0)
        wayspeed = map_wayspeed(mt);

      //update nodes
      int len = wgs84_distance (pos1.lat, pos1.lon, pos2.lat, pos2.lon);
      //maximum nodes distance is sector size
      if (len <= MAPMAXNODEDIST)
      {
        if (update_node(way->nodes[i], &pos2, way->nodes[i-1], &pos1, mt, wayspeed, flags | WF_DIR, textid) != 0)
          return -1;
        if (update_node(way->nodes[i-1], &pos1, way->nodes[i], &pos2, mt, wayspeed, flags, textid) != 0)
          return -1;
      }
      else
      {
        //add nodes between
        int count = len / MAPMAXNODEDIST;
        map_position posx1, posx2;
        posx1.lat = pos1.lat + (pos2.lat-pos1.lat) * 1 / (count+1);
        posx1.lon = pos1.lon + (pos2.lon-pos1.lon) * 1 / (count+1);
        if (update_node(dummynodeid, &posx1, way->nodes[i-1], &pos1, mt, wayspeed, flags | WF_DIR, textid) != 0)
          return -1;
        if (update_node(way->nodes[i-1], &pos1, dummynodeid, &posx1, mt, wayspeed, flags, textid) != 0)
          return -1;
        dummynodeid--;
        for (int j=1; j<count; j++)
        {
          posx2.lat = pos1.lat + (pos2.lat-pos1.lat) * (j+1) / (count+1);
          posx2.lon = pos1.lon + (pos2.lon-pos1.lon) * (j+1) / (count+1);
          if (update_node(dummynodeid, &posx2, dummynodeid+1, &posx1, mt, wayspeed, flags | WF_DIR, textid) != 0)
            return -1;
          if (update_node(dummynodeid+1, &posx1, dummynodeid, &posx2, mt, wayspeed, flags, textid) != 0)
            return -1;
          posx1 = posx2;
          dummynodeid--;
        }
        if (update_node(way->nodes[i], &pos2, dummynodeid+1, &posx1, mt, wayspeed, flags | WF_DIR, textid) != 0)
          return -1;
        if (update_node(dummynodeid+1, &posx1, way->nodes[i], &pos2, mt, wayspeed, flags, textid) != 0)
          return -1;
      }
    }

    pos1 = pos2;
    ok1 = ok2;
    cl1 = cl2;
  }

  return 0;
}

int save_plane(maptypes mt, map_multipolygon *mp)
{
  int i, j;
  int res;
  long id = 0;
  long lat1, lon1, lat2, lon2;
  long slat, slon, zlat, zlon, zoom, zoomd, size;

  //check size
  if (mp->count>65535)
  {
    map_log(LOG_ERROR, "Plane multipolygon with too many polygons (%d)\n", mp->count);
    return -1;
  }
  for (i=0; i<mp->count; i++)
  {
    if (mp->polygon[i]->count>65535)
    {      
      map_log(LOG_ERROR, "Plane polygon with too many points (%d)\n", mp->polygon[i]->count);
      save_polygon (mp->polygon[i], "largepoly.poly");
      return -1;
    }
  }

  size = map_multipolygonsize(mp);
  char sql[300];

  //compute sector
  map_polygonarea(mp->polygon[0], &lat1, &lon1, &lat2, &lon2);
  zlat = (lat2 - lat1) / MAPSECTRECT;
  zlon = (lon2 - lon1) / MAPSECTRECT;
  zoom = zlat>zlon ? zlat : zlon;
  if (zoom==0)
  {
    zoomd = 1;
  }
  else
  {
    zoom = (long)(log((double)zoom)/log(2.0)+1);
    zoomd = (long)pow(2, (double)zoom);
  }
  slat = (lat1 + lat2) / 2;
  slon = (lon1 + lon2) / 2; 
  zlat = slat<0 ? 1 : 0;
  zlon = slon<0 ? 1 : 0;
  slat = (slat / (zoomd * MAPSECTRECT) - zlat) * (zoomd * MAPSECTRECT);
  slon = (slon / (zoomd * MAPSECTRECT) - zlon) * (zoomd * MAPSECTRECT);
  size = map_plane_size + map_multipolygon_size;
  for (i=0; i<mp->count; i++)
  {
    size += map_polygon_size + mp->polygon[i]->count * map_position_size;
  }

  sprintf(sql, "SELECT id FROM mcm_planes_sectors WHERE lat=%ld AND lon=%ld AND zoom=%ld", slat, slon, zoom);
  if (sql_query(sql) != 0)
  {
    return -1;
  }
  sql_result *result = sql_storeresult();
  if (result == NULL)
  {
    return -1;
  }
  sql_row row  = sql_fetch_row(result);
  if (row)
  {
    id = sql_record_long(row, 0);
    sprintf(sql, "UPDATE mcm_planes_sectors SET planes=planes+1, size=size+%ld WHERE id=%ld", size, id);
    if (sql_query(sql) != 0)
    {
      sql_freeresult(result);
      return -1;
    }
  }
  sql_freeresult(result);
  if (id == 0)
  {
    sprintf(sql, "INSERT INTO mcm_planes_sectors (lat, lon, zoom, planes, size) VALUES (%ld, %ld, %ld, 1, %ld)", slat, slon, zoom, size);
    if (sql_query(sql) != 0)
    {
      return -1;
    }
    id = (long)sql_insertid();
  }

  if (id==0)
  {
    return -1;
  }

  //save plane
  char *arr = multipolygon2ascii(mp);
  if (arr == NULL)
    return -1;
  char *sql2 = (char*)malloc(strlen(arr)+200);
  sprintf(sql2, "INSERT INTO mcm_planes (typ, sector, lat1, lon1, lat2, lon2, polygon) VALUES (%d, %ld, %ld, %ld, %ld, %ld, %s)",
          mt, id, lat1, lon1, lat2, lon2, arr);
  res = sql_query(sql2);
  free(sql2);
  return res;
}

//remove plane (only for development purposes)
int remove_plane(long id)
{
  char sql[200];
  sprintf(sql, "SELECT typ, sector, polygon FROM mcm_planes WHERE id=%ld", id);
  if (sql_query(sql) != 0)
    return -1;
  sql_result *result = sql_storeresult();
  if (result == NULL)
  {
    return -1;
  }

  sql_row row;
  if ((row = sql_fetch_row(result)) != NULL)
  {
    char typ = sql_record_char(row, 0);
    long sector = sql_record_long(row, 1);
    char *arr = sql_record_str(row, 2);
    map_multipolygon *mp = ascii2multipolygon(arr);
    long size = map_plane_size + map_multipolygon_size;
    for (int i=0; i<mp->count; i++)
    {
      size += map_polygon_size + mp->polygon[i]->count * map_position_size;
    }
    
    sprintf(sql, "UPDATE mcm_planes_sectors SET planes=planes-1, size=size-%ld WHERE id=%ld", size, sector);
    if (sql_query(sql) != 0)
    {
      sql_freeresult(result);
      return -1;
    }

    sprintf(sql, "DELETE FROM mcm_planes WHERE id=%ld", id);
    if (sql_query(sql) != 0)
    {
      sql_freeresult(result);
      return -1;
    }
  }
  sql_freeresult(result);
  
  return 0;
}

int save_loc(maptypes mt, long lat, long lon, char *name)
{
  int res;
  char sql[500];

  long slat = MAPSECTOR(lat);
  long slon = MAPSECTOR(lon);
  long sid = 0;
  long tid = gettextid(name);

  long size = map_loc_size;
  
  sprintf(sql, "SELECT id FROM mcm_loc_sectors WHERE lat=%ld AND lon=%ld", slat, slon);
  if (sql_query(sql) != 0)
    return -1;
  sql_result *result = sql_storeresult();
  if (result == NULL)
    return -1;
  sql_row row  = sql_fetch_row(result);
  if (row)
  {
    sid = sql_record_long(row, 0);
    sprintf(sql, "UPDATE mcm_loc_sectors SET locs=locs+1, size=size+%ld WHERE id=%ld", size, sid);
    if (sql_query(sql) != 0)
    {
      sql_freeresult(result);
      return -1;
    }
  }
  sql_freeresult(result);
  if (sid == 0)
  {
    sprintf(sql, "INSERT INTO mcm_loc_sectors (lat, lon, locs, size) VALUES (%ld, %ld, 1, %ld)", slat, slon, size);
    if (sql_query(sql) != 0)
    {
      return -1;
    }
    sid = (long)sql_insertid();
  }

  if (sid==0)
  {
    return -1;
  }

  sprintf(sql, "INSERT INTO mcm_loc (sector, typ, lat, lon, name) VALUES (%ld, %d, %ld, %ld, %ld)",
          sid, mt, lat, lon, tid);
  res = sql_query(sql);
  return res;
}

int save_poi(maptypes mt, long lat, long lon, int flags, char *name)
{
  int res;
  char sql[500];

  long slat = MAPSECTOR(lat);
  long slon = MAPSECTOR(lon);
  long sid = 0;
  long tid = gettextid(name);

  long size = map_poi_size;
  
  sprintf(sql, "SELECT id FROM mcm_poi_sectors WHERE lat=%ld AND lon=%ld", slat, slon);
  if (sql_query(sql) != 0)
    return -1;
  sql_result *result = sql_storeresult();
  if (result == NULL)
    return -1;
  sql_row row  = sql_fetch_row(result);
  if (row)
  {
    sid = sql_record_long(row, 0);
    sprintf(sql, "UPDATE mcm_poi_sectors SET pois=pois+1, size=size+%ld WHERE id=%ld", size, sid);
    if (sql_query(sql) != 0)
    {
      sql_freeresult(result);
      return -1;
    }
  }
  sql_freeresult(result);
  if (sid == 0)
  {
    sprintf(sql, "INSERT INTO mcm_poi_sectors (lat, lon, pois, size) VALUES (%ld, %ld, 1, %ld)", slat, slon, size);
    if (sql_query(sql) != 0)
    {
      return -1;
    }
    sid = (long)sql_insertid();
  }

  if (sid==0)
  {
    return -1;
  }

  sprintf(sql, "INSERT INTO mcm_poi (sector, typ, lat, lon, flags, name) VALUES (%ld, %d, %ld, %ld, %d, %ld)",
          sid, mt, lat, lon, flags, tid);
  res = sql_query(sql);
  return res;
}

//return id of city
long save_city(citytypes citytype, int country, map_position *pos, char *name)
{
  long res = -1;
  long tid = gettextid(name);
  int len = (int)strlen(name);
  char *sql = (char *)malloc(len + 300);

  sprintf(sql, "UPDATE mcm_country SET size=size+%d WHERE id=%d", map_city_size, country);
  if (sql_query(sql) != 0)
  {
    free(sql);
    return -1;
  }
  if (sql_affectedrows() == 0)
  {
    sprintf(sql, "INSERT INTO mcm_country (id, size) VALUES (%d, %d)", country, map_city_size);
    if (sql_query(sql) != 0)
    {
      free(sql);
      return -1;
    }
  }

  char *namec = (char*)malloc(len+1);
  utf8_to_ascii(name, namec, 1);
  sprintf(sql, "INSERT INTO mcm_city (country, boundary, typ, lat, lon, name, namec) VALUES (%d, 0, %d, %ld, %ld, %ld, '%s')",
    country, citytype, pos->lat, pos->lon, tid, namec);
  if (sql_query(sql) == 0)
  {
    res = (long)sql_insertid();
  }
  free(sql);
  return res;
}

//return id of street
long save_street(long city, map_position *pos, char *name)
{
  long res = 0;
  long len = (long)strlen(name);
  long tid = gettextid(name);
  char *sql = (char *)malloc(len + 300);

  //check if it exists
  sprintf(sql, "SELECT id FROM mcm_street WHERE city=%ld AND name=%ld", city, tid);
  if (sql_query(sql) != 0)
  {
    free(sql);
    return -1;
  }

  sql_result *result;
  if ((result = sql_storeresult()) == NULL)
  {
    free(sql);
    return -1;
  }

  sql_row row;
  if ((row = sql_fetch_row(result)) != NULL)
  {
    int id = sql_record_int(row, 0);
    sql_freeresult(result);
    free(sql);
    return id;
  }
  sql_freeresult(result);

  //update city size
  sprintf(sql, "UPDATE mcm_city SET size=size+%d WHERE id=%ld", map_street_size, city);
  if (sql_query(sql) != 0)
  {
    free(sql);
    return -1;
  }

  //not exists, add new street
  char *namec = (char*)malloc(len+1);
  utf8_to_ascii(name, namec, 1);
  sprintf(sql, "INSERT INTO mcm_street (city, lat, lon, name, namec) VALUES (%ld, %ld, %ld, %ld, '%s')",
          city, pos->lat, pos->lon, tid, namec);
  if (sql_query(sql) == 0)
  {
    res = (long)sql_insertid();
  }
  free(sql);
  //map_log("OK\n");

  return res;
}

int save_address(long street, char *name, map_position *pos)
{
  int res;
  long len = (long)strlen(name);
  long tid = gettextid(name);
  char *sql = (char *)malloc(len + 300);

  //update street size
  sprintf(sql, "UPDATE mcm_street SET size=size+%d WHERE id=%ld", map_address_size, street);
  if (sql_query(sql) != 0)
  {
    free(sql);
    return -1;
  }

  char *namec = (char*)malloc(len+1);
  utf8_to_ascii(name, namec, 1);
  sprintf(sql, "INSERT INTO mcm_address (street, lat, lon, name, namec) VALUES (%ld, %ld, %ld, %ld, '%s')",
          street, pos->lat, pos->lon, tid, namec);
  res = sql_query(sql);
  free(sql);

  save_poi(POI_ADDRESS, pos->lat, pos->lon, 0, name);
  return res;
}

//CONVERT
int map_convert(map_tags *tags, osm_node *node, osm_way *way, map_multipolygon *mp)
{
  char *value;
  maptypes mt;

  char *name = tag_str(tags, "name");
  char *note = tag_str(tags, "note");

  //highway
  if (way)
  {
    if (((value = tag_str(tags, "highway")) != NULL) || ((value = tag_str(tags, "route")) != NULL))
    {
      mt = (maptypes)-1;
      if (tag_parse(conv_highway, conv_highway_size, value, &mt) == 0) { }
      else if (tag_parse(conv_route, conv_route_size, value, &mt) == 0) { }
      else if ((strcmp(value, "track") == 0) && (value = tag_str(tags, "tracktype")) && (tag_parse(conv_highway, conv_highway_size, value, &mt) == 0)) {};
      if (mt != (maptypes)-1)
      {
        unsigned int flags = 0;
        int speed = 0;

        if ((value = tag_str(tags, "oneway")) && (strcmp(value, "yes")==0)) { flags |= WF_ONEWAY; }
        if ((value = tag_str(tags, "junction")) && (strcmp(value, "roundabout")==0)) { flags |= WF_ONEWAY | WF_ROUNDABOUT; }
        if ((value = tag_str(tags, "toll")) && (strcmp(value, "yes")==0)) { flags |= WF_TOLL; }
        if ((value = tag_str(tags, "access")) && (strcmp(value, "no")==0)) { flags |= WF_NOBICYCLE | WF_NOMOTORCYCLE | WF_NOMOTORCAR | WF_NOTRUCKS; }
        if ((value = tag_str(tags, "bicycle")) && (strcmp(value, "no")==0)) { flags |= WF_NOBICYCLE; }
        if ((value = tag_str(tags, "motorcar")) && (strcmp(value, "no")==0)) { flags |= WF_NOMOTORCAR; }
        if ((value = tag_str(tags, "motor_vehicle")) && (strcmp(value, "no")==0)) { flags |= WF_NOMOTORCAR | WF_NOMOTORCYCLE | WF_NOTRUCKS; }
        if ((value = tag_str(tags, "motorcycle")) && (strcmp(value, "no")==0)) { flags |= WF_NOMOTORCYCLE; }
        if ((value = tag_str(tags, "hgv")) && (strcmp(value, "no")==0)) { flags |= WF_NOTRUCKS; }
        if ((value = tag_str(tags, "layer")) != NULL) { flags |= (atoi(value) & WF_LEVEL)<< WF_LEVELSHIFT; }
        if ((value = tag_str(tags, "bridge")) && (strcmp(value, "yes")==0)) { flags |= WF_BRIDGE; }
        if ((value = tag_str(tags, "tunnel")) && (strcmp(value, "yes")==0)) { flags |= WF_TUNNEL; }
        if ((value = tag_str(tags, "ford")) && (strcmp(value, "yes")==0)) { flags |= WF_FORD; }
        if ((value = tag_str(tags, "maxspeed")) != NULL )
        {
          speed = atoi(value);
        }

        save_way(mt, way, speed, flags, name, note);

        if ((value = tag_str(tags, "area")) && (strcmp(value, "yes")==0))
        {
          //TODO
        }

        if ((value = tag_str(tags, "name")) != NULL)
        {
          map_position pos1, pos2;
          int city = 0;
          if ((map_nodeposition(way->nodes[way->count/2], &pos1) == 0) &&
             (map_nodeposition(way->nodes[way->count/2+1], &pos2) == 0))
          {
            pos1.lat = (pos1.lat+pos2.lat)/2;
            pos1.lon = (pos1.lon+pos2.lon)/2;
            boundary *bnd;
            if ((bnd = map_boundaryinside(&pos1, BT_CITY)) != NULL)
            {
              if ((city = boundary2city(bnd->id)) != 0)
              {
                save_street(city, &pos1, value);
              }
              map_freeboundary(bnd);
            }
          }
        }
      }
    }

    if ((value = tag_str(tags, "waterway")) != NULL)
    {
      mt = (maptypes)-1;
      if ((tag_parse(conv_waterway, conv_waterway_size, value, &mt) == 0) && (mt < maxwaytype))
      {
        save_way(mt, way, 0, 0, name, note);
      }
    }

    if ((value = tag_str(tags, "aeroway")) != NULL)
    {
      mt = (maptypes)-1;
      if ((tag_parse(conv_aeroway, conv_aeroway_size, value, &mt) == 0) && (mt < maxwaytype))
      {
        save_way(mt, way, 0, 0, name, note);
      }
    }

    if ((value = tag_str(tags, "barrier")) != NULL)
    {
      mt = (maptypes)-1;
      if ((tag_parse(conv_barrier, conv_barrier_size, value, &mt) == 0) && (mt < maxwaytype))
      {
        save_way(mt, way, 0, 0, name, note);
      }
    }

    if ((value = tag_str(tags, "railway")) != NULL)
    {
      mt = (maptypes)-1;
      if ((tag_parse(conv_railway, conv_railway_size, value, &mt) == 0) && (mt < maxwaytype))
      {
        save_way(mt, way, 0, 0, name, note);
      }
    }

    if ((value = tag_str(tags, "power")) != NULL)
    {
      mt = (maptypes)-1;
      if ((tag_parse(conv_power, conv_power_size, value, &mt) == 0) && (mt < maxwaytype))
      {
        save_way(mt, way, 0, 0, name, note);
      }
    }

    if ((value = tag_str(tags, "man_made")) != NULL)
    {
      mt = (maptypes)-1;
      if ((tag_parse(conv_manmade, conv_manmade_size, value, &mt) == 0) && (mt < maxwaytype))
      {
        save_way(mt, way, 0, 0, name, note);
      }
    }

    if ((value = tag_str(tags, "boundary")) != NULL)
    {
      if ((value = tag_str(tags, "admin_level")) && (atoi(value)==2))
      {
        save_way(NT_BOUNDARY, way, 0, 0, NULL, NULL);
      }
    }
    
    if ((value = tag_str(tags, "natural")) && (strcmp(value, "coastline")==0))
    {
      save_coastline(way);
      //save_way(WT_MOTORWAY, way, 0, 0, NULL, NULL);  //for testing only
    }
  }

  //planes
  map_multipolygon *mpl = NULL;
  map_multipolygon *mpw = NULL;
  if (mp)
  {
    mpl= mp;
  }
  else if (way && (way->nodes[0]==way->nodes[way->count-1]))
  {
    //convert way to plane, but allow highways only with tag area=yes
    map_polygon *poly;
    if ((tag_str(tags, "highway") == NULL) || 
        ((value = tag_str(tags, "area")) && (strcmp(value, "yes")==0)) )
    {
      if ((way->count>2) && (way2polygon(way, &poly) == 0))
      {
        mpw = (map_multipolygon*)malloc(sizeof(map_multipolygon) + sizeof(map_polygon));
        mpw->count = 1;
        mpw->polygon[0] = poly;
        mpl = mpw;
      }
    }
  }
  
  if (mpl)
  {
    mt = (maptypes)-1;
    if ((value = tag_str(tags, "building")) && (strcmp(value, "yes")==0)) { mt = PT_BUILDING; }
    else if ((value = tag_str(tags, "amenity")) && (tag_parse(conv_amenity, conv_amenity_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "leisure")) && (tag_parse(conv_leisure, conv_leisure_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "landuse")) && (tag_parse(conv_landuse, conv_landuse_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "power")) && (tag_parse(conv_power, conv_power_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "natural")) && (tag_parse(conv_natural, conv_natural_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "waterway")) && (tag_parse(conv_waterway, conv_waterway_size, value, &mt) == 0)) { };

    if ((mt >= firstplanetype) && (mt<=maxplanetype))
    {
      save_plane(mt, mpl);
    }
  }

  if (mpw)
  {
    map_freemultipolygon(mpw);
    mpw = NULL;
  }

  //NODES and POIs
  map_position *pos = NULL;
  int flags = 0;
  if (node)
  {
    pos = (map_position *)malloc(sizeof(map_position));
    *pos = node->pos;
  }
  else if (mp)
  {
    long lat1, lon1, lat2, lon2;
    map_polygonarea(mp->polygon[0], &lat1, &lon1, &lat2, &lon2);
    pos = (map_position *)malloc(sizeof(map_position));
    pos->lat = (lat1+lat2)/2;
    pos->lon = (lon1+lon2)/2;
  }
  else if (way)
  {
    if (way->nodes[0] == way->nodes[way->count-1])
    {
      long lat1, lon1, lat2, lon2;
      map_wayarea(way, &lat1, &lon1, &lat2, &lon2);
      pos = (map_position *)malloc(sizeof(map_position));
      pos->lat = (lat1+lat2)/2;
      pos->lon = (lon1+lon2)/2;
    }
  }

  if (pos)
  {
    if ((value = tag_str(tags, "addr:housenumber")) != NULL)
    {
      char *street = tag_str(tags, "addr:street");
      char *city = tag_str(tags, "addr:city");
      boundary *bnd;
      int cityid = 0;
      long streetid = 0;
      if ((bnd = map_boundaryinside(pos, BT_CITY)) != NULL)
      {
        //get city from boundary
        cityid = boundary2city(bnd->id);
        map_freeboundary(bnd);
        
      }
      if ((cityid == 0) && city)
      {
        //get city id from name
        cityid = name2city(city, pos);
      }
      if ((cityid == 0) && city)
      {
        //city not found, add new city, check country
        if ((bnd = map_boundaryinside(pos, BT_COUNTRY)) != NULL)
        {
          cityid = save_city(CT_HAMLET, bnd->country, pos, city);
          map_freeboundary(bnd);
        }
      }
      if (cityid)
      {
        if (street == NULL)
        {
          //create new street...
          if (city)
          {
            //... as city name
            streetid = save_street(cityid, pos, city);
          }
          else
          {
            //... as cityname from city id
            if ((street = cityname(cityid)) != NULL)
            {
              streetid = save_street(cityid, pos, street);
              free(street);
              street = NULL;
            }
          }
        }
        else
        {
          streetid = save_street(cityid, pos, street);
        }
        if (streetid)
        {
          save_address(streetid, value, pos);
        }
      }
    }

    mt = (maptypes)-1;
    if ((value = tag_str(tags, "amenity")) && (tag_parse(conv_amenity, conv_amenity_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "leisure")) && (tag_parse(conv_leisure, conv_leisure_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "landuse")) && (tag_parse(conv_landuse, conv_landuse_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "natural")) && (tag_parse(conv_natural, conv_natural_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "power")) && (tag_parse(conv_power, conv_power_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "highway")) && (tag_parse(conv_highway, conv_highway_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "railway")) && (tag_parse(conv_railway, conv_railway_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "aeroway")) && (tag_parse(conv_aeroway, conv_aeroway_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "ford")) && (strcmp(value, "yes")==0)) { mt = POI_FORD; }
    else if ((value = tag_str(tags, "tourism")) && (tag_parse(conv_tourism, conv_tourism_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "place")) && (tag_parse(conv_place, conv_place_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "natural")) && (tag_parse(conv_natural, conv_natural_size, value, &mt) == 0)) { }
    else if ((value = tag_str(tags, "historic")) && (tag_parse(conv_historic, conv_historic_size, value, &mt) == 0)) { };

    if ((mt >= firstpoitype) && (mt<=maxpoitype))
    {
      char *name;
      switch (mt)
      {
        case POI_FUEL:
          name = tag_str(tags, "brand");
          if ((value = tag_str(tags, "fuel:octane_98")) && (strcmp(value, "yes")==0)) flags |= PF_FUEL_GASOLINE;
          if ((value = tag_str(tags, "fuel:octane_95")) && (strcmp(value, "yes")==0)) flags |= PF_FUEL_GASOLINE;
          if ((value = tag_str(tags, "fuel:octane_91")) && (strcmp(value, "yes")==0)) flags |= PF_FUEL_GASOLINE;
          if ((value = tag_str(tags, "fuel:lpg")) && (strcmp(value, "yes")==0)) flags |= PF_FUEL_LPG;
          if ((value = tag_str(tags, "fuel:cng")) && (strcmp(value, "yes")==0)) flags |= PF_FUEL_CNG;
          break;
        default:
          name = tag_str(tags, "name");
          break;
      }
      save_poi(mt, pos->lat, pos->lon, flags, name);
    }

    if ((mt >= firstloctype) && (mt<=maxloctype))
    {
      char *name = tag_str(tags, "name");
      save_loc(mt, pos->lat, pos->lon, name);
    }

    free(pos);
  }

  return 0;
}

int map_reindex()
{
  char sql[500];
  long sectors;
  long progress = -1;

  //write index
  sprintf(sql, "SELECT id FROM mcm_sectors ORDER BY lat, lon ");
  if (sql_query(sql) != 0)
    return -1;
  sql_result *result1 = sql_storeresult();
  if (result1 == NULL)
    return -1;
  sectors = sql_numrows(result1);

  sql_row row1;
  long index = 0;
  long mid = 1;
  while ((row1 = sql_fetch_row(result1)) != NULL)
  {
    //write sector coordinates
    long id = sql_record_long(row1, 0);

    sprintf(sql, "SELECT id FROM mcm_nodes WHERE sector=%ld", id);
    if (sql_query(sql) != 0)
      return -1;
    sql_result *result2 = sql_storeresult();
    if (result2 == NULL)
    {
      sql_freeresult(result1);
      return -1;
    }

    sql_row row2;
    while ((row2 = sql_fetch_row(result2)) != NULL)
    {
      long nodeid = sql_record_long(row2, 0);
      sprintf(sql, "UPDATE mcm_nodes SET mid=%ld WHERE id=%ld", mid, nodeid);
      if (sql_query(sql) != 0)
      {
        sql_freeresult(result1);
        sql_freeresult(result2);
        return -1;
      }
      mid++;
    }
    sql_freeresult(result2);

    index++;
    long prg = index * 100 / sectors;
    if (prg!=progress)
    {
      progress = prg;
      printf("Reindex nodes %ld%%\r", prg);
      fflush(stdout);
    }
  }
  sql_freeresult(result1);
  printf("\n");
  return 0;
}

//fix duplicates in city names
//TODO
int map_fix_duplicate_text(void)
{
  map_log(LOG_INFO, "Process duplicate city names\n");

  map_log(LOG_INFO, "Process duplicate city names done\n");
  return 0;
}
